(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Monitor
imports First

begin

section \<open>Monitors\<close>

text {*
The RV monitors language is  introduced plus the algebraic properties of the monitor operators. 

*}

subsection \<open>Syntax\<close>

datatype 'a monitor =
   mFIRST_d "'a pitl" ("(FIRST _)")
 | mUPTO_d "'a monitor" "'a monitor" ("(_ UPTO _)" [84,84] 83)
 | mTHRU_d "'a monitor" "'a monitor" ("(_ THRU _)" [84,84] 83)
 | mTHEN_d "'a monitor" "'a monitor" ("(_ THEN _)" [84,84] 83)
 | mWITH_d "'a monitor" "'a pitl"    ("(_ WITH _)" [84,84] 83)

subsection \<open>Derived Monitors\<close>

definition mHALT_d ("(HALT _)" [84] 83)
where
 "HALT w \<equiv> FIRST (fin (init w))"

definition mLEN_d ("(LEN _)" [84] 83)
where
 "LEN k \<equiv> FIRST (len k)"

definition mEMPTY_d ("EMPTY")
where
 "EMPTY \<equiv> FIRST empty"

definition mSKIP_d ("SKIP")
where
 "SKIP \<equiv> FIRST skip"

definition mGUARD_d ("(GUARD _)" [84] 83)
where
 "GUARD w \<equiv> EMPTY WITH ( (init w))"  

definition mFAIL_d ("FAIL")
where 
 "FAIL \<equiv> GUARD false\<^sub>i"

primrec mTIMES_d :: "'a monitor \<Rightarrow> nat \<Rightarrow> 'a monitor" ("(_ TIMES _)" [84,84] 83)
where
  mTIMES_0  : "a TIMES 0       = EMPTY"
| mTIMES_Suc: "a TIMES (Suc k) = a THEN (a TIMES k)"

definition mALWAYS_d ("(_ ALWAYS _)" [84,84] 83)
where 
 "a ALWAYS (w) \<equiv> a WITH (bi (fin (init w)))"

definition mSOMETIME_d ("(_ SOMETIME _)" [84,84] 83)
where 
 "a SOMETIME (w) \<equiv> a WITH (di (fin (init w)))"
 
definition mlimit_d ("(Limit _)" [84] 83)
where 
 "Limit f \<equiv> (bs (\<not>\<^sub>i f))"  (*  \<equiv> (\<not>\<^sub>i di (f;skip))  *)

definition mWITHIN_d ("(_ WITHIN _)" [84,84] 83)
where 
 "a WITHIN (f) \<equiv> a WITH (Limit f)"

definition mUNTIL_d ("(_ UNTIL _)" [84,84] 83)
  where
 "w1 UNTIL w2 \<equiv> (HALT w2) WITH (bm w1)"

subsection \<open>Semantics\<close>

fun semantics_monitor :: "'a monitor \<Rightarrow> 'a pitl" ("(\<M> _)" [80] 80)
where
  "(\<M> (FIRST a))  = \<rhd> a"
| "(\<M> (a UPTO b)) = \<rhd>( (\<M> a) \<or>\<^sub>i (\<M> b))"
| "(\<M> (a THRU b)) = \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> b))"
| "(\<M> (a THEN b)) = ((\<M> a);(\<M> b))"
| "(\<M> (a WITH f)) = ( (\<M> a) \<and>\<^sub>i f )" 

definition mAND_d ("(_ AND _)" [84,84] 83)
where 
  " a AND b \<equiv> a WITH (\<M> b) "

definition mITERATE_d ("(_ ITERATE _)" [84,84] 83)
where
  " a ITERATE b \<equiv> a WITH (\<M> b)\<^sup>\<star>"

definition mSTAR_d ("(_ STAR _)" [84,84] 83)
where
 " a STAR f \<equiv> (FIRST(\<diamond> f)) ITERATE (a)"

definition mREPEAT_d ("(_ REPEATUNTIL _)" [84,84] 83)
where 
  " a REPEATUNTIL w \<equiv> ( (HALT w) ITERATE (a WITH (keep(\<not>\<^sub>i (init w))))) THEN a"


subsection \<open>Monitor Laws\<close>

lemma MFixFst:
 "\<turnstile> (\<M> a) \<equiv>\<^sub>i \<rhd> (\<M> a)"
proof
 (induct a )
 case (mFIRST_d x)
 then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> FIRST x) \<equiv>\<^sub>i \<rhd> x"  by simp
   have 2: "\<turnstile> \<rhd> x \<equiv>\<^sub>i \<rhd> (\<rhd> x)" using FstFixFst by auto
   have 3: "\<turnstile>  \<rhd> (\<rhd> x) \<equiv>\<^sub>i \<rhd>(\<M> FIRST x)" by simp
   from 1 2 3 show ?thesis by auto
  qed
 next
  case (mUPTO_d a1 a2)
  then show ?case 
  proof -
  have 1: "\<turnstile> (\<M> (a1 UPTO a2)) \<equiv>\<^sub>i \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2))" by simp
  have 2: "\<turnstile>  \<rhd>( (\<M> a1) \<or>\<^sub>i (\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<rhd>((\<M> a1) \<or>\<^sub>i (\<M> a2)))" using FstFixFst by auto
  have 3: "\<turnstile> \<rhd>(\<rhd>((\<M> a1) \<or>\<^sub>i (\<M> a2))) \<equiv>\<^sub>i \<rhd>(\<M> (a1 UPTO a2))" by simp
  from 1 2 3 show ?thesis by auto
 qed  
next
  case (mTHRU_d a1 a2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a1 THRU a2)) \<equiv>\<^sub>i \<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2))" by simp
   have 2: "\<turnstile> \<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<rhd>(di(\<M> a1) \<and>\<^sub>i di(\<M> a2)))" using FstFixFst by auto
   have 3: "\<turnstile> \<rhd>(\<rhd>( di (\<M> a1) \<and>\<^sub>i di(\<M> a2))) \<equiv>\<^sub>i \<rhd>(\<M> (a1 THRU a2))" by simp
   from 1 2 3 show ?thesis by auto
 qed
next
  case (mTHEN_d a1 a2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a1 THEN a2)) \<equiv>\<^sub>i (\<M> a1) ; (\<M> a2) " 
        by simp
   have 2: "\<turnstile> (\<M> a1) ; (\<M> a2) \<equiv>\<^sub>i \<rhd>(\<M> a1) ; \<rhd>(\<M> a2)" 
        using ChopEqvChop mTHEN_d.hyps(1) mTHEN_d.hyps(2) by blast
   have 3: "\<turnstile> \<rhd>(\<M> a1) ; \<rhd>(\<M> a2) \<equiv>\<^sub>i \<rhd>(\<rhd>(\<M> a1) ; (\<M> a2))" 
        using FstFstChopEqvFstChopFst itl_prop(30) by blast
   have 4: "\<turnstile> \<rhd>(\<rhd>(\<M> a1) ; (\<M> a2)) \<equiv>\<^sub>i \<rhd>((\<M> a1) ; (\<M> a2))" 
        using FstEqvRule LeftChopEqvChop itl_prop(30) mTHEN_d.hyps(1) by blast
   have 5: "\<turnstile> \<rhd>((\<M> a1) ; (\<M> a2)) \<equiv>\<^sub>i \<rhd>(\<M> (a1 THEN a2))" 
        by simp
   from 1 2 3 4 5 show ?thesis by auto
  qed
next
  case (mWITH_d a x2)
  then show ?case 
  proof -
   have 1: "\<turnstile> (\<M> (a WITH x2)) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i ( x2)" 
        by simp
   have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i ( x2) \<equiv>\<^sub>i \<rhd>(\<M> a) \<and>\<^sub>i ( x2)" 
        using mWITH_d.hyps by auto
   have 3: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i ( x2) \<equiv>\<^sub>i \<rhd>(\<rhd>(\<M> a) \<and>\<^sub>i ( x2))"  
        using FstFstAndEqvFstAnd itl_prop(30) by blast
   have 4: "\<turnstile> \<rhd>(\<rhd>(\<M> a) \<and>\<^sub>i ( x2)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<and>\<^sub>i ( x2))" 
        using "2" FstEqvRule itl_prop(30) by blast
   have 5: "\<turnstile> \<rhd>((\<M> a) \<and>\<^sub>i ( x2)) \<equiv>\<^sub>i \<rhd>(\<M> (a WITH x2))" 
        by simp
   from 1 2 3 4 5 show ?thesis by auto
 qed
qed

lemma MGuardFalseEqvFalse:
 "\<turnstile> \<M> (GUARD false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M>(GUARD false\<^sub>i) \<equiv>\<^sub>i \<M>(EMPTY WITH (init false\<^sub>i))" by (simp add: mGUARD_d_def)
  have 2: "\<turnstile> \<M>(EMPTY WITH (init false\<^sub>i)) \<equiv>\<^sub>i \<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i)" by simp
  have 3: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i (init false\<^sub>i)" by simp
  have 4: "\<turnstile> \<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i) \<equiv>\<^sub>i \<M>(EMPTY) \<and>\<^sub>i false\<^sub>i" using "3" by simp
  have 5: "\<turnstile> \<M>(EMPTY) \<and>\<^sub>i false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" by simp
  have 6: "\<turnstile>\<M>(EMPTY) \<and>\<^sub>i (init false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using "4" "5" by simp
  have 7: "\<turnstile>\<M>(EMPTY WITH (init false\<^sub>i)) \<equiv>\<^sub>i false\<^sub>i" using "2" "6" by simp
  have 8: "\<turnstile>\<M>(GUARD false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using "1" "7" by simp
  from 8 show ?thesis by auto
qed

lemma MFirstFalseEqvFalse:
 "\<turnstile> \<M> (FIRST false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M>(FIRST false\<^sub>i) \<equiv>\<^sub>i \<rhd> false\<^sub>i" by simp
  have 2:  "\<turnstile> \<M>(FIRST false\<^sub>i) \<equiv>\<^sub>i false\<^sub>i" using FstFalse by simp
 from 2 show ?thesis by auto
qed

lemma MFailAlt:
 "\<turnstile> \<M> FAIL \<equiv>\<^sub>i false\<^sub>i"
proof -
  have 1: "\<turnstile> \<M> FAIL \<equiv>\<^sub>i \<M> (GUARD (false\<^sub>i)) " by (simp add: mFAIL_d_def)
  have 2: "\<turnstile> \<M> (GUARD (false\<^sub>i)) \<equiv>\<^sub>i false\<^sub>i" using MGuardFalseEqvFalse by auto
 from 1 2 show ?thesis by auto
qed

lemma MFailEqvFirstFalseWithinEmpty:
 "\<turnstile> (\<M> FAIL) \<equiv>\<^sub>i \<M>((FIRST false\<^sub>i) WITHIN ( empty ))"
proof -
  have 1: "\<turnstile> \<M>( (FIRST false\<^sub>i) WITHIN ( empty )) \<equiv>\<^sub>i \<M> ((FIRST false\<^sub>i) WITH (Limit empty) )" 
       by (simp add: mWITHIN_d_def)
  have 2: "\<turnstile> \<M> ((FIRST false\<^sub>i) WITH (Limit empty) ) \<equiv>\<^sub>i \<M> (FIRST false\<^sub>i) \<and>\<^sub>i (Limit empty ) " 
       by simp
  have 3: "\<turnstile> \<M> ((FIRST false\<^sub>i) WITH (Limit empty) ) \<equiv>\<^sub>i false\<^sub>i" 
       using MFirstFalseEqvFalse by auto
  have 4: "\<turnstile> \<M>( (FIRST false\<^sub>i) WITHIN ( empty )) \<equiv>\<^sub>i false\<^sub>i" 
       using "1" "3" by auto
  have 5: "\<turnstile> (\<M> FAIL) \<equiv>\<^sub>i false\<^sub>i" 
       using MFailAlt by simp
 from 4 5 show ?thesis by auto 
 qed

lemma MEmptyAlt:
 "\<turnstile> (\<M> EMPTY) \<equiv>\<^sub>i empty"
proof -
  have 1: "\<turnstile> (\<M> EMPTY) \<equiv>\<^sub>i (\<M> (FIRST empty))" by (simp add: mEMPTY_d_def)
  have 2: "\<turnstile> (\<M> (FIRST empty)) \<equiv>\<^sub>i \<rhd> empty" by simp
  have 3: "\<turnstile> \<rhd> empty \<equiv>\<^sub>i empty" using FstEmpty by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MSkipAlt:
 "\<turnstile> \<M> SKIP \<equiv>\<^sub>i skip"
proof -
  have 1: "\<turnstile> \<M> SKIP \<equiv>\<^sub>i \<M>(FIRST skip)" by (simp add: mSKIP_d_def)
  have 2: "\<turnstile> \<M>(FIRST skip) \<equiv>\<^sub>i \<rhd> skip" by simp
  have 3: "\<turnstile> \<rhd> skip \<equiv>\<^sub>i skip" using FstSkip by simp
 from 1 2 3 show ?thesis by auto
qed

lemma MGuardAlt:
 "\<turnstile> \<M> (GUARD(w)) \<equiv>\<^sub>i empty \<and>\<^sub>i init w"
proof -
  have 1: "\<turnstile> \<M> (GUARD(w)) \<equiv>\<^sub>i \<M>(EMPTY WITH ( (init w)))" by (simp add:mGUARD_d_def)
  have 2: "\<turnstile> \<M>(EMPTY WITH ( (init w))) \<equiv>\<^sub>i (\<M> EMPTY) \<and>\<^sub>i ( init w)" by simp
  have 3: "\<turnstile> (\<M> EMPTY) \<and>\<^sub>i ( init w) \<equiv>\<^sub>i empty \<and>\<^sub>i ( init w)" using MEmptyAlt prop06 by blast
  have 4: "\<turnstile>  empty \<and>\<^sub>i ( init w) \<equiv>\<^sub>i empty \<and>\<^sub>i init w" by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma MLengthAlt:
 "\<turnstile>\<M>(LEN(k)) \<equiv>\<^sub>i len(k)"
proof -
  have 1: "\<turnstile> \<M>(LEN(k)) \<equiv>\<^sub>i \<M>(FIRST(len(k)))" by (simp add:mLEN_d_def)
  have 2: "\<turnstile> \<M>(FIRST(len(k))) \<equiv>\<^sub>i \<rhd>(len(k))" by simp
  have 3: "\<turnstile> \<rhd>(len(k)) \<equiv>\<^sub>i len(k)" using FstLenEqvLen by blast
 from 1 2 3 show ?thesis by auto
qed

lemma BoxStateEqvBiFinState:
 "\<turnstile> \<box> (init w) \<equiv>\<^sub>i bi (fin (init w))"
proof -
  have 1: "\<turnstile> \<diamond> (\<not>\<^sub>i (init w)) \<equiv>\<^sub>i true\<^sub>i ; \<not>\<^sub>i(init w)" 
       by simp
  have 2: "\<turnstile> \<diamond> (init(\<not>\<^sub>i w)) \<equiv>\<^sub>i true\<^sub>i ; init (\<not>\<^sub>i w)" 
       by simp
  have 3: "\<turnstile> di (true\<^sub>i \<and>\<^sub>i fin (init (\<not>\<^sub>i w))) \<equiv>\<^sub>i true\<^sub>i ; init (\<not>\<^sub>i w)" 
       using DiAndFinEqvChopState by blast
  have 4: "\<turnstile> \<diamond> (init(\<not>\<^sub>i w)) \<equiv>\<^sub>i di (true\<^sub>i \<and>\<^sub>i fin (init (\<not>\<^sub>i w))) " 
       using 1 2 3 by simp
  have 5: "\<turnstile> \<not>\<^sub>i (\<diamond> (init(\<not>\<^sub>i w))) \<equiv>\<^sub>i \<not>\<^sub>i (di (true\<^sub>i \<and>\<^sub>i fin (init (\<not>\<^sub>i w))))" 
       using 4 by simp
  have 6: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i \<not>\<^sub>i (di (true\<^sub>i \<and>\<^sub>i fin (init (\<not>\<^sub>i w))))" 
       using 5 by auto
  have 7: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i bi (\<not>\<^sub>i (fin (init (\<not>\<^sub>i w))))" 
       using 6 by auto
  have 8: "\<turnstile> init (\<not>\<^sub>i w) \<equiv>\<^sub>i \<not>\<^sub>i (init w)" 
       by simp
  have 9: "\<turnstile> fin (init (\<not>\<^sub>i w)) \<equiv>\<^sub>i fin (\<not>\<^sub>i (init w))" 
       using 8 FinEqvFin by blast
  have 10: "\<turnstile> fin (init (\<not>\<^sub>i w)) \<equiv>\<^sub>i \<not>\<^sub>i (fin (init w))" 
       using 8 FinNotStateEqvNotFinState FinEqvFin by blast
  have 11: "\<turnstile> \<not>\<^sub>i (fin (init (\<not>\<^sub>i w))) \<equiv>\<^sub>i (fin (init w))" 
       using 10  by simp
  have 12: "\<turnstile> bi (\<not>\<^sub>i (fin (init (\<not>\<^sub>i w)))) \<equiv>\<^sub>i bi (fin (init w))" 
       using 11  by simp
  have 13: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i bi (fin (init w))" 
       using 7 12 by simp
  from 13 show ?thesis by simp
qed

lemma MAlwaysAlt:
  "\<turnstile> \<M>(a ALWAYS w) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<box> (init w)"
proof -
  have 1: "\<turnstile> \<M>(a ALWAYS w) \<equiv>\<^sub>i \<M>(a WITH (bi (fin (init w))))" 
       by (simp add: mALWAYS_d_def)
  have 2: "\<turnstile> \<M>(a WITH (bi (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bi (fin (init w)))" 
       by simp
  have 3: "\<turnstile> \<M>(a) \<and>\<^sub>i (bi (fin (init w))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<box> (init w)" 
       using BoxStateEqvBiFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma DiamondStateEqvDiFinState:
 "\<turnstile> \<diamond> (init w) \<equiv>\<^sub>i di (fin (init w))"
proof -
  have 1: "\<turnstile> \<box> (init (\<not>\<^sub>i w)) \<equiv>\<^sub>i bi (fin (init (\<not>\<^sub>i w)))" using BoxStateEqvBiFinState by blast
  have 2: "\<turnstile> \<not>\<^sub>i (\<box> (init (\<not>\<^sub>i w))) \<equiv>\<^sub>i \<not>\<^sub>i (bi (fin (init (\<not>\<^sub>i w))))" using 1 by auto
  have 3: "\<turnstile> \<diamond> (\<not>\<^sub>i (init (\<not>\<^sub>i w))) \<equiv>\<^sub>i di (\<not>\<^sub>i (fin (init (\<not>\<^sub>i w))))" using 2 by auto
  have 4: "\<turnstile> \<diamond> (init  w) \<equiv>\<^sub>i di (\<not>\<^sub>i (fin (init (\<not>\<^sub>i w))))" using 3 by auto
  have 5: "\<turnstile> \<diamond> (init  w) \<equiv>\<^sub>i di (fin (init w))" using 4 FinNotStateEqvNotFinState by auto
  from 1 2 3 4 5 show ?thesis by simp
qed

lemma MSometimeAlt:
 "\<turnstile> \<M>(a SOMETIME w) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<diamond> (init w)"
proof -
  have 1: "\<turnstile> \<M>(a SOMETIME w) \<equiv>\<^sub>i \<M>(a WITH (di (fin (init w))))" 
       by (simp add: mSOMETIME_d_def)
  have 2: "\<turnstile>\<M>(a WITH (di (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (di (fin (init w)))" 
       by simp
  have 3: "\<turnstile>\<M>(a WITH (di (fin (init w)))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i  \<diamond> (init w)" 
       using DiamondStateEqvDiFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MWithinAlt:
 "\<turnstile> \<M>(a WITHIN f) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bs (\<not>\<^sub>i f))"
proof -
  have 1: "\<turnstile> \<M>(a WITHIN f) \<equiv>\<^sub>i \<M>(a WITH (bs (\<not>\<^sub>i f)))" 
       by (simp add: mWITHIN_d_def mlimit_d_def)
  have 2: "\<turnstile> \<M>(a WITH (bs (\<not>\<^sub>i f))) \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i (bs (\<not>\<^sub>i f))" 
       by simp
  from 1 2 show ?thesis by simp
qed

lemma MTimesAlt:
 "\<turnstile> \<M>(a TIMES k) \<equiv>\<^sub>i power (\<M> a) k"
proof 
 (induct k)
 case 0
 then show ?case 
  proof -
   have 1: "\<turnstile> \<M>  a TIMES 0 \<equiv>\<^sub>i \<M> EMPTY" by simp
   have 2: "\<turnstile> \<M> EMPTY \<equiv>\<^sub>i empty " using MEmptyAlt by simp
   have 3: "\<turnstile> empty \<equiv>\<^sub>i power (\<M> a) 0" by simp
   from 1 2 3 show ?thesis by auto
  qed
 next
 case (Suc k)
 then show ?case 
  proof -
   have 1: "\<turnstile> \<M>  a TIMES Suc k \<equiv>\<^sub>i \<M>(a THEN (a TIMES k) ) " 
        by simp
   have 2: "\<turnstile> \<M>(a THEN (a TIMES k) ) \<equiv>\<^sub>i (\<M> a);(\<M>(a TIMES k))" 
        by simp
   have 3: "\<turnstile> (\<M> a);(\<M>(a TIMES k)) \<equiv>\<^sub>i (\<M> a);(power (\<M> a) k)" 
        using RightChopEqvChop Suc.hyps by blast
   have 4: "\<turnstile> (\<M> a);(power (\<M> a) k) \<equiv>\<^sub>i power (\<M> a) (Suc k)" 
        by simp
   from 1 2 3 4 show ?thesis by auto
  qed
 qed

lemma MUptoAlt:
 "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i ((\<M> a) \<and>\<^sub>i (\<M> b)) "
proof -
  have 1: "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))" 
       by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i (\<rhd>(\<M> a) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) " 
       using FstWithOrEqv by blast
  have 3: "\<turnstile> (\<rhd>(\<M> a) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ((\<M> b) \<or>\<^sub>i \<not>\<^sub>i(\<M> b)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i 
            ((\<M> b) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i \<not>\<^sub>i(\<M> a)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a)))"  
       using MFixFst by auto
  have 4: "\<turnstile> ((\<M> a) \<and>\<^sub>i ((\<M> b) \<or>\<^sub>i \<not>\<^sub>i(\<M> b)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> b))) \<or>\<^sub>i 
            ((\<M> b) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i \<not>\<^sub>i(\<M> a)) \<and>\<^sub>i  (bs \<not>\<^sub>i(\<M> a))) \<equiv>\<^sub>i 
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) )" 
       by auto
  have 5: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( ((\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( (\<rhd>(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( (\<rhd>(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) )" 
       by (simp add: first_d_def)
  have 6: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( (\<rhd>(\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( (\<rhd>(\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a))) ) " 
       using MFixFst by auto
  have 7: "\<turnstile> (\<not>\<^sub>i(\<M> b) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> b)) \<equiv>\<^sub>i bi(\<not>\<^sub>i(\<M> b))" 
       using AndBsEqvBi by blast
  have 8: "\<turnstile> (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i bs\<not>\<^sub>i(\<M> a)) \<equiv>\<^sub>i bi(\<not>\<^sub>i(\<M> a))" 
       using AndBsEqvBi by blast
  have 9: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ((\<not>\<^sub>i(\<M> b)) \<and>\<^sub>i bs(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ((\<not>\<^sub>i(\<M> a)) \<and>\<^sub>i bs(\<not>\<^sub>i(\<M> a)))) ) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> a)))) )" 
       using "7" "8" by auto
  have 10: "\<turnstile>((\<M> a) \<and>\<^sub>i ( ((\<M> b)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> b)))) ) \<or>\<^sub>i
            ((\<M> b) \<and>\<^sub>i ( ((\<M> a)) \<or>\<^sub>i ( bi(\<not>\<^sub>i(\<M> a)))) ) \<equiv>\<^sub>i 
            (( (\<M> a) \<and>\<^sub>i (\<M> b)) \<or>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> b)))) \<or>\<^sub>i 
            (( (\<M> b) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i ( (\<M> b) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> a)))) "  
       by auto
  have 11: "\<turnstile> (( (\<M> a) \<and>\<^sub>i (\<M> b)) \<or>\<^sub>i ( (\<M> a) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> b)))) \<or>\<^sub>i 
            (( (\<M> b) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i ( (\<M> b) \<and>\<^sub>i bi(\<not>\<^sub>i(\<M> a)))) \<equiv>\<^sub>i
           ( (\<M> a) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i ((\<M> a) \<and>\<^sub>i (\<M> b)) " 
       by auto
 from 1 2 3 4 5 6 9 10 11 show ?thesis by auto
qed

lemma MThruAlt:
 "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i di(\<M> a))"
proof -
  have 1: "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by simp
  have 2: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i (\<rhd>(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i di(\<M> a))" 
       using FstDiAndDiEqv by auto
  have 3: "\<turnstile> (\<rhd>(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (\<rhd>(\<M> b) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i
            ((\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i ((\<M> b) \<and>\<^sub>i di(\<M> a))" 
       using MFixFst by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MHaltAlt:
 "\<turnstile> \<M>(HALT w) \<equiv>\<^sub>i halt(init w)"
proof -
  have 1: "\<turnstile> \<M>(HALT w) \<equiv>\<^sub>i \<M>(FIRST (fin (init  w)))" by (simp add: mHALT_d_def)
  have 2: "\<turnstile> \<M>(FIRST (fin (init  w))) \<equiv>\<^sub>i \<rhd> (fin (init  w))" by simp
  have 3: "\<turnstile> \<rhd> (fin (init  w)) \<equiv>\<^sub>i halt(init w)" using HaltStateEqvFstFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MFailUpto:
"\<turnstile> (\<M> (FAIL UPTO a)) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> (\<M> (FAIL UPTO a)) \<equiv>\<^sub>i \<rhd>((\<M> FAIL) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> \<M> FAIL \<or>\<^sub>i \<M> a \<equiv>\<^sub>i false\<^sub>i \<or>\<^sub>i \<M> a" using MFailAlt by auto
  have 3: "\<turnstile> \<rhd>((\<M> FAIL) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<rhd>(false\<^sub>i \<or>\<^sub>i (\<M> a))" using 2 FstEqvRule by blast
  have 4: "\<turnstile> false\<^sub>i \<or>\<^sub>i (\<M> a) \<equiv>\<^sub>i \<M> a" by simp
  have 5: "\<turnstile> \<rhd>(false\<^sub>i \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i  \<rhd>(\<M> a)" using 4 FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i \<M> a" using MFixFst by auto
  from 1 2 3 4 5 6 show ?thesis by auto
qed

lemma MFailThru:
 "\<turnstile> \<M>(FAIL THRU a) \<equiv>\<^sub>i \<M> FAIL"
proof -
  have 1: "\<turnstile> \<M>(FAIL THRU a) \<equiv>\<^sub>i \<rhd>(di (\<M> FAIL) \<and>\<^sub>i di(\<M> a))" 
       by simp
  have 2: "\<turnstile> \<rhd>(di (\<M> FAIL) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) " 
       using MFailAlt DiEqvDi FstEqvRule prop06 by blast
  have 3: "\<turnstile> di false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" 
       by simp
  hence 4: "\<turnstile> \<rhd>(di (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>( (false\<^sub>i) \<and>\<^sub>i di(\<M> a))" 
       using FstEqvRule prop06 by blast
  have 5: "\<turnstile> \<rhd>( (false\<^sub>i) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>false\<^sub>i" 
       using FstEqvRule itl_prop(19) by blast
  have 6: "\<turnstile> \<rhd>false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" using FstFalse 
       by auto
  have 7: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" 
       using MFailAlt by auto
 from 1 2 4 5 6 7 show ?thesis by auto
qed

lemma MFailAnd:
"\<turnstile> \<M> (FAIL AND a) \<equiv>\<^sub>i \<M> FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL AND a) \<equiv>\<^sub>i (\<M> FAIL) \<and>\<^sub>i (\<M> a)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> FAIL) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i false\<^sub>i \<and>\<^sub>i (\<M> a)" using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> \<M> (FAIL AND a) \<equiv>\<^sub>i false\<^sub>i" using 1 2 3 by auto
  have 5: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
  from 1 2 3 4 5 show ?thesis by auto
qed

lemma MThenFail:
 "\<turnstile> \<M> (a THEN FAIL) \<equiv>\<^sub>i \<M> FAIL"
proof -
  have 1: "\<turnstile> \<M> (a THEN FAIL) \<equiv>\<^sub>i (\<M> a);(\<M> FAIL)" by simp
  have 2: "\<turnstile> (\<M> a);(\<M> FAIL) \<equiv>\<^sub>i (\<M> a);false\<^sub>i " using MFailAlt by auto
  have 3: "\<turnstile> (\<M> a);false\<^sub>i \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma MFailThen:
 "\<turnstile> \<M> (FAIL THEN a) \<equiv>\<^sub>i \<M> FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL THEN a) \<equiv>\<^sub>i (\<M> FAIL);(\<M> a)" by simp
  have 2: "\<turnstile> (\<M> FAIL);(\<M> a) \<equiv>\<^sub>i false\<^sub>i;(\<M> a) " using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i;(\<M> a) \<equiv>\<^sub>i false\<^sub>i" by auto
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma MFailWith:
 "\<turnstile> \<M> (FAIL WITH f) \<equiv>\<^sub>i \<M> FAIL"
proof -
  have 1: "\<turnstile> \<M> (FAIL WITH f) \<equiv>\<^sub>i (\<M> FAIL) \<and>\<^sub>i f" by simp
  have 2: "\<turnstile> (\<M> FAIL) \<and>\<^sub>i f \<equiv>\<^sub>i false\<^sub>i \<and>\<^sub>i f" using MFailAlt by auto
  have 3: "\<turnstile> false\<^sub>i \<and>\<^sub>i f \<equiv>\<^sub>i false\<^sub>i" by simp
  have 4: "\<turnstile> false\<^sub>i \<equiv>\<^sub>i \<M> FAIL" using MFailAlt by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma MEmptyUpto:
 "\<turnstile> \<M> (EMPTY UPTO a) \<equiv>\<^sub>i \<M> EMPTY"
proof -
  have 1: "\<turnstile> \<M> (EMPTY UPTO a) \<equiv>\<^sub>i \<rhd>((\<M> EMPTY) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> (\<M> EMPTY)  \<equiv>\<^sub>i empty " using MEmptyAlt by auto
 hence 3: "\<turnstile> (\<M> EMPTY) \<or>\<^sub>i (\<M> a) \<equiv>\<^sub>i empty \<or>\<^sub>i (\<M> a)" by auto
 hence 4: "\<turnstile>  \<rhd>((\<M> EMPTY) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i  \<rhd>(empty \<or>\<^sub>i (\<M> a))" using FstEqvRule by blast
  have 5: "\<turnstile> \<rhd>(empty \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i empty" using FstEmptyOrEqvEmpty by blast
  have 6: "\<turnstile> empty \<equiv>\<^sub>i (\<M> EMPTY)" using MEmptyAlt by auto
 from 1 4 5 6 show ?thesis by auto
qed

lemma MEmptyThru:
 "\<turnstile> \<M> (EMPTY THRU a) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY THRU a) \<equiv>\<^sub>i \<rhd>(di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a))" by simp
  have 2: "\<turnstile> di(\<M> EMPTY) \<equiv>\<^sub>i di empty " using MEmptyAlt DiEqvDi by blast
 hence 3: "\<turnstile> di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di empty \<and>\<^sub>i di(\<M> a)" by auto
 hence 4: "\<turnstile> di empty \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di(\<M> a)" by auto
  have 5: "\<turnstile> di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a) \<equiv>\<^sub>i di(\<M> a)" using "3" "4" by auto
 hence 6: "\<turnstile> \<rhd>(di(\<M> EMPTY) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di(\<M> a))" using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstDiEqvFst by blast
  have 8: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 6 7 8 show ?thesis by auto
qed

lemma MThenEmpty: 
 "\<turnstile> \<M> (a THEN EMPTY) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> \<M> (a THEN EMPTY) \<equiv>\<^sub>i (\<M> a);(\<M> EMPTY)" by simp
  have 2: "\<turnstile> (\<M> a);(\<M> EMPTY) \<equiv>\<^sub>i (\<M> a); empty" using MEmptyAlt by auto
  have 3: "\<turnstile> (\<M> a); empty \<equiv>\<^sub>i (\<M> a)" using ChopEmpty by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MEmptyThen: 
 "\<turnstile> \<M> (EMPTY THEN a) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY THEN a) \<equiv>\<^sub>i (\<M> EMPTY);(\<M> a)" by simp
  have 2: "\<turnstile> (\<M> EMPTY);(\<M> a) \<equiv>\<^sub>i empty; (\<M> a)" using MEmptyAlt by auto
  have 3: "\<turnstile> empty;(\<M> a) \<equiv>\<^sub>i (\<M> a)" using ChopEmpty by auto
 from 1 2 3 show ?thesis by auto
qed

lemma MEmptyIterate: 
 "\<turnstile> \<M> (EMPTY ITERATE b) \<equiv>\<^sub>i (\<M> EMPTY)"
proof -
  have 1: "\<turnstile> \<M> (EMPTY ITERATE b) \<equiv>\<^sub>i \<M> (EMPTY WITH (\<M> b)\<^sup>\<star>)" 
       by (simp add: mITERATE_d_def)
  have 2: "\<turnstile> \<M> (EMPTY WITH (\<M> b)\<^sup>\<star>) \<equiv>\<^sub>i \<M> EMPTY \<and>\<^sub>i (\<M> b)\<^sup>\<star>" 
       by simp
  have 3: "\<turnstile> \<M> EMPTY \<and>\<^sub>i (\<M> b)\<^sup>\<star>  \<equiv>\<^sub>i empty \<and>\<^sub>i (\<M> b)\<^sup>\<star> " 
       using MEmptyAlt by auto
  have 4: "\<turnstile> empty \<and>\<^sub>i (\<M> b)\<^sup>\<star>  \<equiv>\<^sub>i empty \<and>\<^sub>i (empty \<or>\<^sub>i (((\<M> b) \<and>\<^sub>i more);(\<M> b)\<^sup>\<star>))" 
       using ChopstarEqv by auto
  have 5: "\<turnstile> empty \<and>\<^sub>i (empty \<or>\<^sub>i (((\<M> b) \<and>\<^sub>i more);(\<M> b)\<^sup>\<star>)) \<equiv>\<^sub>i empty" 
       by auto
  have 6: "\<turnstile> \<M> (EMPTY ITERATE b) \<equiv>\<^sub>i \<M> EMPTY" 
       using 1 2 3 4 5 MEmptyAlt by auto
 from 6 show ?thesis by simp
qed

lemma FstAndFstStarEqvFst:
 "\<turnstile> \<rhd>f \<and>\<^sub>i (\<rhd>f)\<^sup>\<star> \<equiv>\<^sub>i \<rhd>f "
proof -
  have 1: "\<turnstile> (\<rhd>f)\<^sup>\<star> \<equiv>\<^sub>i empty \<or>\<^sub>i (\<rhd>f);(\<rhd>f)\<^sup>\<star> " using CSEqvOrChopCS by simp
  have 2: "\<turnstile> (\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i (empty \<or>\<^sub>i (\<rhd>f);(\<rhd>f)\<^sup>\<star>) \<and>\<^sub>i \<rhd>f" using 1 prop06 by blast
  have 3: "\<turnstile> (empty \<or>\<^sub>i (\<rhd>f);(\<rhd>f)\<^sup>\<star>) \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i (empty \<and>\<^sub>i \<rhd>f) \<or>\<^sub>i ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f)" by auto
  have 4: "\<turnstile> (\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i (empty \<and>\<^sub>i \<rhd>f) \<or>\<^sub>i ((\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f)" using 2 3 by simp
  have 5: "\<turnstile> (\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i (\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f;empty" using ChopEmpty by auto
  have 6: "\<turnstile> (\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f;empty \<equiv>\<^sub>i (\<rhd>f);((\<rhd>f)\<^sup>\<star> \<and>\<^sub>i empty)" using LFstAndDistrC by blast
  have 7: "\<turnstile> (\<rhd>f)\<^sup>\<star> \<and>\<^sub>i empty \<equiv>\<^sub>i empty" using EmptyImpCS by auto
  have 8: "\<turnstile> (\<rhd>f);((\<rhd>f)\<^sup>\<star> \<and>\<^sub>i empty) \<equiv>\<^sub>i \<rhd>f" using 7 RightChopEqvChop ChopEmpty by auto
  have 9: "\<turnstile> (\<rhd>f);(\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i \<rhd>f" using 5 6 8 by simp
  have 10: "\<turnstile> (\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i (empty \<and>\<^sub>i \<rhd>f) \<or>\<^sub>i \<rhd>f" using 4 9 by simp
  have 11: "\<turnstile> (empty \<and>\<^sub>i \<rhd>f) \<or>\<^sub>i \<rhd>f \<equiv>\<^sub>i \<rhd>f" by auto
  have 12:  "\<turnstile> (\<rhd>f)\<^sup>\<star> \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i \<rhd>f" using 10 11 by simp
  from 12 show ?thesis by auto
qed

lemma MIterateIdemp: 
 "\<turnstile> \<M> (a ITERATE a) \<equiv>\<^sub>i \<M> a"
proof -
  have 1: "\<turnstile> \<M> (a ITERATE a) \<equiv>\<^sub>i \<M> (a WITH (\<M> a)\<^sup>\<star>)" by (simp add: mITERATE_d_def)
  have 2: "\<turnstile> \<M> (a WITH (\<M> a)\<^sup>\<star>) \<equiv>\<^sub>i \<M> a \<and>\<^sub>i (\<M> a)\<^sup>\<star>"  by simp
  have 3: "\<turnstile> \<M> a \<and>\<^sub>i (\<M> a)\<^sup>\<star> \<equiv>\<^sub>i \<rhd>(\<M> a) \<and>\<^sub>i (\<rhd>(\<M> a))\<^sup>\<star> " using MFixFst by auto
  have 4: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i (\<rhd>(\<M> a))\<^sup>\<star> \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstAndFstStarEqvFst by simp
  have 5: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i \<M> a" using MFixFst by auto
  from 1 2 3 4 5 show ?thesis using prop03 by blast
qed

lemma MUptoIdemp:
 "\<turnstile> \<M> (a UPTO a) \<equiv>\<^sub>i \<M> a"
proof -
  have 1: "\<turnstile> \<M> (a UPTO a) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> a))" by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstEqvRule itl_prop(27) by blast
  have 3: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 2 3 show ?thesis by simp
qed

lemma MThruIdemp:
 "\<turnstile> \<M> ( a THRU a) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> \<M> ( a THRU a) \<equiv>\<^sub>i \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> a))" by simp
  have 2: "\<turnstile> \<rhd>( di(\<M> a) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<rhd>(di (\<M> a))" using FstEqvRule itl_prop(20) by blast
  have 3: "\<turnstile> \<rhd>(di (\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" using FstDiEqvFst by blast
  have 4: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" using MFixFst by auto
 from 1 2 3 4 show ?thesis by simp
qed

lemma MAndIdemp:
 "\<turnstile> \<M>(a AND a) \<equiv>\<^sub>i (\<M> a)"
proof -
  have 1: "\<turnstile> \<M>(a AND a) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> a)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i (\<M> a)" by auto
 from 1 2 show ?thesis by auto
qed

lemma MWithIdemp:
 "\<turnstile> \<M> ( (a WITH f)  WITH f) \<equiv>\<^sub>i \<M>(a WITH f)"
proof -
  have 1: "\<turnstile> \<M> ( (a WITH f)  WITH f) \<equiv>\<^sub>i ((\<M> a) \<and>\<^sub>i ( f)) \<and>\<^sub>i ( f)" by simp
  have 2: "\<turnstile> ((\<M> a) \<and>\<^sub>i ( f)) \<and>\<^sub>i ( f) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i ( f)" by auto
  have 3: "\<turnstile> (\<M> a) \<and>\<^sub>i ( f) \<equiv>\<^sub>i \<M>(a WITH f)" by simp
 from 1 2 3 show ?thesis by auto
qed

lemma MUptoCommut:
 "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i \<M>(b UPTO a)"
proof -
  have 1: "\<turnstile> \<M>(a UPTO b) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))" by simp
  have 2: "\<turnstile> ((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> a))" by auto
 hence 3: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<equiv>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> a)) " using FstEqvRule by blast
  have 4: "\<turnstile> \<rhd>((\<M> b) \<or>\<^sub>i (\<M> a)) \<equiv>\<^sub>i \<M>(b UPTO a)" by simp
 from 1 3 4 show ?thesis by auto
qed

lemma MThruCommut:
 "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i \<M>(b THRU a)"
proof -
  have 1: "\<turnstile> \<M>(a THRU b) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))" by simp
  have 2: "\<turnstile> (di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> a)) " by auto
 hence 3: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> a))" using FstEqvRule by blast
  have 4: "\<turnstile> \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> a)) \<equiv>\<^sub>i \<M>(b THRU a)" by simp
 from 1 3 4 show ?thesis by auto
qed

lemma MAndCommut:
 "\<turnstile> \<M>(a AND b) \<equiv>\<^sub>i \<M>(b AND a)"
proof -
  have 1: "\<turnstile> \<M>(a AND b) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> b)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> b) \<equiv>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> a) " by auto
  have 3: "\<turnstile> (\<M> b) \<and>\<^sub>i (\<M> a) \<equiv>\<^sub>i \<M>(b AND a)" by (simp add: mAND_d_def)
 from 1 2 3 show ?thesis by auto
qed

lemma MWithCommut:
 "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i \<M>((a WITH g) WITH f) "
proof -
  have 1: "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g)" by simp
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g) \<equiv>\<^sub>i \<M>((a WITH g) WITH f)" by auto
 from 1 2  show ?thesis by auto
qed

lemma MWithAbsorp:
 "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i \<M>(a WITH (f \<and>\<^sub>i g)) "
proof -
  have 1: "\<turnstile> \<M>((a WITH f) WITH g) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g)" by simp
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (f) \<and>\<^sub>i (g) \<equiv>\<^sub>i  (\<M> a) \<and>\<^sub>i (f \<and>\<^sub>i g)" by auto
 from 1 2 show ?thesis by auto
qed

lemma MUptoAssoc:
 "\<turnstile> \<M>((a UPTO b) UPTO c) \<equiv>\<^sub>i \<M>(a UPTO (b UPTO c))"
proof -
  have 1: "\<turnstile> \<M>((a UPTO b) UPTO c) \<equiv>\<^sub>i \<rhd>(\<M>(a UPTO b) \<or>\<^sub>i (\<M> c))" 
       by simp
  have 2: "\<turnstile> \<rhd>(\<M>(a UPTO b) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c))" 
       by simp
  have 3: "\<turnstile> \<rhd>(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>(((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c))"  
       using FstFstOrEqvFstOrL by blast
  have 4: "\<turnstile> (((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i ((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c)))" 
       by auto
  hence 5: "\<turnstile> \<rhd>(((\<M> a) \<or>\<^sub>i (\<M> b)) \<or>\<^sub>i (\<M> c)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c)))"  
       using FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i ((\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> c)))" 
       using FstFstOrEqvFstOrR  itl_prop(30) by blast
  have 7: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>((\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<M>(b UPTO c))" 
       by simp
  have 8: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<M>(b UPTO c)) \<equiv>\<^sub>i \<M>(a UPTO (b UPTO c))" 
       by simp
 from 1 2 3 5 6 7 8 show ?thesis by auto
qed

lemma MThruAssoc:
 "\<turnstile> \<M>((a THRU b) THRU c) \<equiv>\<^sub>i \<M>(a THRU (b THRU c))"
proof -
  have 1: "\<turnstile> \<M>((a THRU b) THRU c) \<equiv>\<^sub>i \<rhd>(di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<and>\<^sub>i di(\<M> c))" 
       by simp
  have 2: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di((di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       using DiEqvDiFst itl_prop(30) by blast
  have 3: "\<turnstile> di((di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b) " 
       using DiDiAndEqvDi by blast
  have 4: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b)" 
       using "2" "3" by auto
  hence 5: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di(\<M> b)  \<and>\<^sub>i di(\<M> c)" 
       by auto
  have 6: "\<turnstile> di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di (di(\<M> b)  \<and>\<^sub>i di(\<M> c))" 
       using DiDiAndEqvDi itl_prop(30) by blast
  have 7: "\<turnstile> di (di(\<M> b)  \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       using DiEqvDiFst by blast
  have 8: "\<turnstile> di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       using "6" "7" using prop03 by blast
  hence 9: "\<turnstile> di(\<M> a) \<and>\<^sub>i di(\<M> b)  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))" 
       by auto
  have 10: "\<turnstile> di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c) \<equiv>\<^sub>i 
             di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c))) " 
       using "5" "9" by auto
  hence 11: "\<turnstile> \<rhd>(di(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)))  \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i 
             \<rhd>(di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))) "  
       using FstEqvRule by blast
  have 12: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di (\<rhd>(di(\<M> b)  \<and>\<^sub>i di(\<M> c)))) \<equiv>\<^sub>i \<M>(a THRU (b THRU c)) " 
       by simp
 from 1 11 12 show ?thesis by auto
qed

lemma MAndAssoc:
 "\<turnstile> \<M>((a AND b) AND c) \<equiv>\<^sub>i \<M>(a AND (b AND c))"
proof -
  have 1: "\<turnstile> \<M>((a AND b) AND c) \<equiv>\<^sub>i (\<M> a) \<and>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> c)" by (simp add: mAND_d_def)
  have 2: "\<turnstile> (\<M> a) \<and>\<^sub>i (\<M> b) \<and>\<^sub>i (\<M> c) \<equiv>\<^sub>i \<M>(a AND (b AND c))" by (simp add: mAND_d_def)
 from 1 2 show ?thesis by auto
qed

lemma MThenAssoc:
 "\<turnstile> \<M>((a THEN b) THEN c) \<equiv>\<^sub>i \<M>(a THEN (b THEN c))"
proof -
  have 1: "\<turnstile> \<M>((a THEN b) THEN c) \<equiv>\<^sub>i ((\<M> a);(\<M> b));(\<M> c) " by simp
  have 2: "\<turnstile> ((\<M> a);(\<M> b));(\<M> c) \<equiv>\<^sub>i (\<M> a);((\<M> b);(\<M> c))" using ChopAssocB by blast
  have 3: "\<turnstile> (\<M> a);((\<M> b);(\<M> c)) \<equiv>\<^sub>i \<M>(a THEN (b THEN c))" by simp
 from 1 2 3 show ?thesis by auto
qed

lemma OrDiEqvDi:
 "\<turnstile> f \<or>\<^sub>i di f \<equiv>\<^sub>i di f"
proof -
  have 1: "\<turnstile> f \<supset>\<^sub>i di f" using DiIntro by blast
 from 1 show ?thesis by auto
qed

lemma AndDiEqv:
 "\<turnstile> f \<and>\<^sub>i di f \<equiv>\<^sub>i f"
proof -
  have 1: "\<turnstile> f \<supset>\<^sub>i di f" using DiIntro by blast
 from 1 show ?thesis by auto
qed

lemma MUptoThruAbsorp:
 "\<turnstile> \<M>(a UPTO (a THRU b)) \<equiv>\<^sub>i \<M> a"
proof -
  have 1: "\<turnstile> \<M>(a UPTO (a THRU b)) \<equiv>\<^sub>i \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b) ))" 
       by simp
  have 2: "\<turnstile> \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
            \<rhd>((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) ))" 
       using FstFstOrEqvFstOrR by auto
  have 3: "\<turnstile>  ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
             (((\<M> a) \<or>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       by auto
  have 4: "\<turnstile> (((\<M> a) \<or>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
            ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using OrDiEqvDi by auto
  have 5: "\<turnstile> ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
            ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using "3" "4" by auto
  hence 6: "\<turnstile> \<rhd> ((\<M> a) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b) )) \<equiv>\<^sub>i
             \<rhd> ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) " 
       using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd> ((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
            (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) " 
       by (simp add: first_d_def)
  have 8: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<equiv>\<^sub>i 
            (di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by auto
  hence 9: "\<turnstile> \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
            \<not>\<^sub>i((di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using prop01 by blast
  have 10: "\<turnstile> \<not>\<^sub>i((di(\<M> a) \<and>\<^sub>i (\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             \<not>\<^sub>i(((\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using AndDiEqv by auto
  have 11: "\<turnstile> \<not>\<^sub>i(((\<M> a)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
              \<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       by auto
  have 12: "\<turnstile> \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              \<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))" 
       using "9" "10" "11" by auto
  hence 13: "\<turnstile> bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              bs (\<not>\<^sub>i(\<M> a) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using BsEqvRule by blast
  have 14: "\<turnstile> bs ((\<not>\<^sub>i(\<M> a)) \<and>\<^sub>i \<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using BsAndEqv using itl_prop(30) by blast
  have 141: "\<turnstile> bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using "13" "14" by auto
  hence 15: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
              bs \<not>\<^sub>i((di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       by auto
  have 16: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i
              bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             (bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       by auto
  have 17: "\<turnstile> (bs ((\<not>\<^sub>i(\<M> a))) \<and>\<^sub>i di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             (\<rhd>(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) " 
       using FstEqvBsNotAndDi itl_prop(30) prop06 by blast
  have 18: "\<turnstile> (\<rhd>(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
             ((\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using MFixFst itl_prop(30) prop06 by blast
  have 19: "\<turnstile> ((\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i  
              bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       by auto
  have 20: "\<turnstile>   (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (\<not>\<^sub>i(di(\<M> a)) \<or>\<^sub>i \<not>\<^sub>i(di(\<M> b)) )"  
       by auto
  have 21: "\<turnstile> (\<not>\<^sub>i(di(\<M> a)) \<or>\<^sub>i \<not>\<^sub>i(di(\<M> b)) ) \<equiv>\<^sub>i ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b)) )" 
       by auto
  have 22: "\<turnstile> (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b)))  " 
       using "20" "21" by auto
  hence 23: "\<turnstile> bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i bs ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b))) " 
       using BsEqvRule by blast
  have 24: "\<turnstile> bs ((bi \<not>\<^sub>i(\<M> a)) \<or>\<^sub>i (bi\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))" 
       using BsOrBsEqvBsBiOrBi  itl_prop(30) by blast
  have 25: "\<turnstile> bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))" 
       using "23" "24" by auto
  hence 26: "\<turnstile> (\<M> a) \<and>\<^sub>i bs (\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i 
              (\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b)))" 
       by auto
  have 27: "\<turnstile>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i 
            \<rhd>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) " 
       using MFixFst prop06 by blast 
  have 28: "\<turnstile> \<rhd>(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i
             (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b)))" 
       by (simp add: first_d_def)
  have 29: "\<turnstile> (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<and>\<^sub>i (bs (\<not>\<^sub>i(\<M> a)) \<or>\<^sub>i bs (\<not>\<^sub>i(\<M> b))) \<equiv>\<^sub>i
             (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a)" 
       by auto
  have 30: "\<turnstile> (\<M> a) \<and>\<^sub>i bs \<not>\<^sub>i(\<M> a) \<equiv>\<^sub>i \<rhd>(\<M> a)" 
       by (simp add: first_d_def)
  have 31: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" 
       using MFixFst by auto
  have 32: "\<turnstile> \<M>(a UPTO (a THRU b)) \<equiv>\<^sub>i 
            (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using "1" "2" "6" "7" by auto
  have 33: "\<turnstile> (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i 
            bs \<not>\<^sub>i( (di(\<M> a)) \<and>\<^sub>i ((\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i
             ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b)))" 
       using "15" "16" "17" "18" "19" by auto
  have 34: "\<turnstile> ((\<M> a)) \<and>\<^sub>i  bs(\<not>\<^sub>i(di(\<M> a) \<and>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (\<M> a) " 
       using "26" "27" "28" "29" "30" "31" using prop03 by blast
 from 32 33 34 show ?thesis by auto
qed

lemma MThruUptoAbsorp:
 "\<turnstile> \<M>(a THRU (a UPTO b)) \<equiv>\<^sub>i (\<M> a) "
proof -
  have 1: "\<turnstile> \<M>(a THRU (a UPTO b)) \<equiv>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))))" 
       by simp
  have 2: "\<turnstile>  \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b)))) \<equiv>\<^sub>i 
              \<rhd>(di(\<M> a) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> b))))" 
       using DiEqvDiFst using FstEqvRule itl_prop(30) prop05 by blast
  have 3: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> b)))) \<equiv>\<^sub>i 
            \<rhd>(di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b)))" 
       using DiOrEqv using FstEqvRule prop05 by blast
  have 4: "\<turnstile> (di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i (di(\<M> a))" 
        by auto
  hence 5: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> b))) \<equiv>\<^sub>i \<rhd>(di(\<M> a))" 
       using FstEqvRule by blast
  have 6: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>(\<M> a)" 
       using FstDiEqvFst by blast
  have 7: "\<turnstile> \<rhd>(\<M> a) \<equiv>\<^sub>i (\<M> a)" 
       using MFixFst by auto
 from 1 2 3 5 6 7 show ?thesis by auto
qed

lemma MUptoThruDistrib:
 "\<turnstile> \<M>(a UPTO (b THRU c)) \<equiv>\<^sub>i \<M>((a UPTO b) THRU (a UPTO c))"
proof -
  have 1: "\<turnstile> \<M>((a UPTO b) THRU (a UPTO c)) \<equiv>\<^sub>i 
            \<rhd>( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) " 
       by simp
  have 2: "\<turnstile> ( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            ( di(((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> c))) )" 
       using DiEqvDiFst by (metis itl_prop(31) prop22)
  have 3: "\<turnstile> ( di(((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            ( di(\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> c))"  
       using DiOrEqv by auto
  have 4: "\<turnstile> ( di(\<M> a) \<or>\<^sub>i di(\<M> b)) \<and>\<^sub>i (di(\<M> a) \<or>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i
           di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) " 
       by auto
  have 5: "\<turnstile> ( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c))" 
        using "2" "3" "4" by auto
  hence 6: "\<turnstile> \<rhd>( di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> b))) \<and>\<^sub>i di(\<rhd>((\<M> a) \<or>\<^sub>i (\<M> c))) ) \<equiv>\<^sub>i
            \<rhd>( di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) " 
       using FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>( di(\<M> a) \<or>\<^sub>i (di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) \<equiv>\<^sub>i
            \<rhd>( \<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) ) " 
       using FstFstOrEqvFstOr by auto
  have 8: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i \<rhd>((\<M> a))" 
       using FstDiEqvFst  by blast
  have 9: "\<turnstile> \<rhd>((\<M> a)) \<equiv>\<^sub>i (\<M> a) " 
       using MFixFst by auto
  have 10: "\<turnstile> \<rhd>(di(\<M> a)) \<equiv>\<^sub>i (\<M> a)" 
       using "8" "9" by auto
  hence 11: "\<turnstile> \<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)) \<equiv>\<^sub>i 
              (\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))" 
       by auto
  hence 12: "\<turnstile> \<rhd>(\<rhd>(di(\<M> a)) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i 
              \<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c)))" 
       using FstEqvRule by blast
  have 13: "\<turnstile>\<rhd>((\<M> a) \<or>\<^sub>i \<rhd>(di(\<M> b) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i \<M>(a UPTO (b THRU c)) " 
       by simp
 from 1 6 7 12 13 show ?thesis by auto
qed

lemma MThruUptoDistrib:
 "\<turnstile> \<M>(a THRU (b UPTO c)) \<equiv>\<^sub>i \<M>((a THRU b) UPTO (a THRU c))"
proof -
  have 1: "\<turnstile> \<M>((a THRU b) UPTO (a THRU c)) \<equiv>\<^sub>i
          \<rhd>(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> c)))" by simp
  have 2: "\<turnstile> \<rhd>(\<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            \<rhd>((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c)))" using FstFstOrEqvFstOr by auto
  have 3: "\<turnstile> ((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i (di(\<M> b) \<or>\<^sub>i di(\<M> c)))" by auto
  have 4: "\<turnstile> (di(\<M> a) \<and>\<^sub>i (di(\<M> b) \<or>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di( (\<M> b) \<or>\<^sub>i (\<M> c)))" using DiOrEqv by auto
  have 5: "\<turnstile> (di(\<M> a) \<and>\<^sub>i di( (\<M> b) \<or>\<^sub>i (\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c))))" using DiEqvDiFst prop05 by blast
  have 6: "\<turnstile> ((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            (di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c))))" using "3" "4" "5" by auto
 hence 7: "\<turnstile> \<rhd>((di(\<M> a) \<and>\<^sub>i di(\<M> b)) \<or>\<^sub>i (di(\<M> a) \<and>\<^sub>i di(\<M> c))) \<equiv>\<^sub>i
            \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c)))) " using FstEqvRule by blast
  have 8: "\<turnstile> \<rhd>(di(\<M> a) \<and>\<^sub>i di(\<rhd>( (\<M> b) \<or>\<^sub>i (\<M> c)))) \<equiv>\<^sub>i
            \<M>(a THRU (b UPTO c))" by simp
 from 1 2 7 8 show ?thesis by auto
qed

lemma MWithAndDistrib:
 "\<turnstile> \<M>((a AND b) WITH f) \<equiv>\<^sub>i \<M>((a WITH f) AND (b WITH f))"
proof -
  have 1: "\<turnstile> \<M>((a AND b) WITH f) \<equiv>\<^sub>i \<M>(a AND b) \<and>\<^sub>i f" 
       by simp
  have 2: "\<turnstile> \<M>(a AND b) \<equiv>\<^sub>i \<M>(a WITH (\<M> b))" 
       by (simp add: mAND_d_def)
  have 3: "\<turnstile> \<M>(a AND b) \<and>\<^sub>i f \<equiv>\<^sub>i \<M>(a WITH (\<M> b)) \<and>\<^sub>i f" 
       using 2 prop06 by simp
  have 4: "\<turnstile> \<M>(a WITH (\<M> b)) \<and>\<^sub>i f \<equiv>\<^sub>i \<M>(a) \<and>\<^sub>i \<M>(b) \<and>\<^sub>i f" 
       by simp
  have 5: "\<turnstile> \<M>(a) \<and>\<^sub>i \<M>(b) \<and>\<^sub>i f \<equiv>\<^sub>i (\<M>(a) \<and>\<^sub>i f) \<and>\<^sub>i (\<M>(b) \<and>\<^sub>i f)" 
       by auto
  have 6: "\<turnstile> (\<M>(a) \<and>\<^sub>i f) \<and>\<^sub>i (\<M>(b) \<and>\<^sub>i f) \<equiv>\<^sub>i \<M>(a WITH f) \<and>\<^sub>i \<M>(b WITH f)" 
       by simp
  have 7: "\<turnstile> \<M>(a WITH f) \<and>\<^sub>i \<M>(b WITH f) \<equiv>\<^sub>i \<M>((a WITH f) WITH (\<M>(b WITH f)))" 
       by simp
  have 8: "\<turnstile> \<M>((a WITH f) WITH (\<M>(b WITH f))) \<equiv>\<^sub>i \<M>((a WITH f) AND (b WITH f))" 
       by (simp add: mAND_d_def)
  from 1 2 3 4 5 6 7 8 show ?thesis by simp
qed


lemma MHaltWithAndDistrib:
 "\<turnstile> \<M>(((HALT w) WITH f) AND ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<and>\<^sub>i g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) AND ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<M>(((HALT w) WITH f) WITH (\<M>((HALT w) WITH g)))" 
       by (simp add: mAND_d_def)
  have 2: "\<turnstile> \<M>(((HALT w) WITH f) WITH (\<M>((HALT w) WITH g))) \<equiv>\<^sub>i 
             \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i \<M>(HALT w) \<and>\<^sub>i g" 
       by (simp add: mHALT_d_def)
  have 3: "\<turnstile> \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i \<M>(HALT w) \<and>\<^sub>i g \<equiv>\<^sub>i \<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g"  
       by auto
  from 1 2 3 show ?thesis by auto
qed

lemma MHaltWithUptoHaltWithEqvHaltWithOr:
 "\<turnstile> \<M>(((HALT w) WITH f) UPTO ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<or>\<^sub>i g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) UPTO ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>(\<M>((HALT w) WITH f) \<or>\<^sub>i \<M>((HALT w) WITH g))" 
       by simp
  have 2: "\<turnstile> \<rhd>(\<M>((HALT w) WITH f) \<or>\<^sub>i \<M>((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>((\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g))" 
       by simp
  have 3: "\<turnstile> (\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g))" 
       by auto 
  have 4: "\<turnstile> \<rhd>((\<M>(HALT w) \<and>\<^sub>i f) \<or>\<^sub>i (\<M>(HALT w) \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g))" 
       using 3 FstEqvRule by blast
  have 5: "\<turnstile> \<rhd>(\<M>(HALT w) \<and>\<^sub>i (f \<or>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<or>\<^sub>i g)))" 
       by simp
  have 6: "\<turnstile> (\<M>((HALT w) WITH (f \<or>\<^sub>i g))) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<or>\<^sub>i g)))" 
       using MFixFst by blast
  from 1 2 3 4 5 6 show ?thesis by simp
qed

lemma DiHaltAndDiHaltAndEqvDiHaltAndAnd:
   "\<turnstile> di(halt (init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt (init w) \<and>\<^sub>i g) \<equiv>\<^sub>i di(halt (init w) \<and>\<^sub>i f \<and>\<^sub>i g)"
proof -
  have 1: "\<turnstile> di(halt (init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt (init w) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(\<rhd>(fin (init w)) \<and>\<^sub>i f) \<and>\<^sub>i di (\<rhd>(fin (init w)) \<and>\<^sub>i g)" 
       using HaltStateEqvFstFinState by auto
  have 2: "\<turnstile> di(\<rhd>(fin (init w)) \<and>\<^sub>i f) \<and>\<^sub>i di(\<rhd>(fin (init w)) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(\<rhd>(fin (init w)) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using LFstAndDistrD by simp
  have 3: "\<turnstile> di(\<rhd>(fin (init w)) \<and>\<^sub>i f \<and>\<^sub>i g)  \<equiv>\<^sub>i di(halt (init w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using HaltStateEqvFstFinState by auto
  from 1 2 3 show ?thesis by simp
qed

lemma MHaltWithThruHaltWithEqvHaltWithAndHaltWith:
 "\<turnstile> \<M>(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>(((HALT w) WITH f) AND ((HALT w) WITH g))"
proof -
  have 1: "\<turnstile> \<M>(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<equiv>\<^sub>i 
             \<rhd>( di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) )" 
       by simp
  have 2: "\<turnstile> di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(halt(init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt(init w) \<and>\<^sub>i g)" 
       using MHaltAlt DiEqvDi by auto
  have 3: "\<turnstile> di(halt(init w) \<and>\<^sub>i f) \<and>\<^sub>i di(halt(init w) \<and>\<^sub>i g) \<equiv>\<^sub>i 
             di(halt(init w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using DiHaltAndDiHaltAndEqvDiHaltAndAnd by simp
  have 4: "\<turnstile> di( halt(init w) \<and>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using MHaltAlt by auto
  have 5: "\<turnstile> di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g) \<equiv>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using 2 3 4 by simp
  have 6: "\<turnstile> \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f) \<and>\<^sub>i di(\<M>(HALT w) \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g))" 
       using 5 FstEqvRule by blast
  have 7: "\<turnstile> \<rhd>(di(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g)" 
       using FstDiEqvFst by simp
  have 8: "\<turnstile> \<rhd>(\<M>(HALT w) \<and>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<and>\<^sub>i g)))" 
       by simp
  have 9: "\<turnstile> \<M>((HALT w) WITH (f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(\<M>((HALT w) WITH (f \<and>\<^sub>i g)))" 
       using MFixFst by blast
  have 10: "\<turnstile> \<M>(((HALT w) WITH f) THRU ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<and>\<^sub>i g))" 
       using 1 2 3 4 5 6 7 8 9 by simp
  have 11: "\<turnstile> \<M>(((HALT w) WITH f) AND ((HALT w) WITH g)) \<equiv>\<^sub>i \<M>((HALT w) WITH (f \<and>\<^sub>i g))" 
       using MHaltWithAndDistrib by simp
  have 12: "\<turnstile> \<M>((HALT w) WITH (f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<M>(((HALT w) WITH f) AND ((HALT w) WITH g))" 
       using 11 by simp
  from 10 12 show ?thesis by simp
qed


end
