(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Semantics
 imports
   Interval
   Syntax
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)

section \<open>Semantics\<close>

text {*

The semantics of the basic ITL operators is defined. Then lemmas are introduced that define 
the derived ITL operators in terms of operations on intervals. 
Furthermore we prove the soundness of the ITL axioms. 
This is followed by the key time reversal lemma.  

*}

subsection \<open>Semantics Primitive Operators\<close>

fun semantics_pitl :: "['a interval, 'a pitl] \<Rightarrow> bool" ("(_ \<Turnstile> _)" [80,10] 10)
where
  "(\<sigma> \<Turnstile> false\<^sub>i) = False"
| "(\<sigma> \<Turnstile> atom\<^sub>i p) = (p (intfirst \<sigma>))"
| "(\<sigma> \<Turnstile> f \<supset>\<^sub>i g) = ((\<sigma> \<Turnstile> f) \<longrightarrow> (\<sigma> \<Turnstile> g))"
| "(\<sigma> \<Turnstile> skip)  = (intlen \<sigma> = 1)"
| "(\<sigma> \<Turnstile> f ; g) =
   (\<exists>i. ((0 \<le> i) \<and> (i \<le> (intlen \<sigma>)) \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> ((suffix i \<sigma>) \<Turnstile> g)) )"
| "(\<sigma> \<Turnstile> f\<^sup>\<star>) = (\<exists> (l::index). index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                          )"



subsection \<open>Semantics Boolean Operators\<close>

lemma not_defs [simp]:
  " (\<sigma> \<Turnstile> \<not>\<^sub>i f ) = Not ( \<sigma> \<Turnstile> f) " 
by (simp add: not_d_def)

lemma or_defs [simp]: 
  " (\<sigma> \<Turnstile> f1 \<or>\<^sub>i f2) = ( (\<sigma> \<Turnstile> f1) \<or> (\<sigma> \<Turnstile> f2))"
by (metis not_defs or_d_def semantics_pitl.simps(3))

lemma and_defs [simp]:
  " (\<sigma> \<Turnstile> f1 \<and>\<^sub>i f2) = ( (\<sigma> \<Turnstile> f1) \<and> (\<sigma> \<Turnstile> f2))"
by (simp add: and_d_def)

lemma implies_defs [simp]:
  " (\<sigma> \<Turnstile> f1 \<supset>\<^sub>i f2) = ( (\<sigma> \<Turnstile> f1) \<longrightarrow> (\<sigma> \<Turnstile> f2))"
by auto

lemma iff_defs [simp]: 
  " (\<sigma> \<Turnstile> f1 \<equiv>\<^sub>i f2) = ( (\<sigma> \<Turnstile> f1) = (\<sigma> \<Turnstile> f2))"
by (metis and_defs iff_d_def semantics_pitl.simps(3))

lemma true_defs [simp]:
  " (\<sigma> \<Turnstile> true\<^sub>i)"
by (simp add: true_d_def)

subsection \<open>Semantics Box and Diamond Operators\<close>

lemma sometimes_defs [simp]: 
  " (\<sigma> \<Turnstile> \<diamond> f) = (\<exists> i \<le> intlen \<sigma>.  (suffix i \<sigma> \<Turnstile> f))"
by (simp add: sometimes_d_def)

lemma always_defs [simp]:
  " (\<sigma> \<Turnstile> \<box> f) = (\<forall> i \<le> intlen \<sigma>.  (suffix i \<sigma> \<Turnstile> f))"
by (simp add: always_d_def)

lemma di_defs [simp]:
  " (\<sigma> \<Turnstile> di f) = (\<exists> i \<le> intlen \<sigma>.  (prefix i \<sigma> \<Turnstile> f))"
by (simp add: di_d_def)

lemma bi_defs [simp]:
  " (\<sigma> \<Turnstile> bi f) = (\<forall> i \<le> intlen \<sigma>. (prefix i \<sigma> \<Turnstile> f))"
by (simp add: bi_d_def)

lemma da_defs [simp]:
  " (\<sigma> \<Turnstile> da f) = (\<exists> i ia . 0 \<le> i  \<and> ia+i \<le> intlen \<sigma> \<and> (sub i (ia +i) \<sigma> \<Turnstile> f))"
apply (simp add: da_d_def)
using interval_prefix_length_good interval_suffix_length_good 
by (smt add.commute add_diff_cancel_left' add_leD2 interval_sub_prefix_suffix_0 le_iff_add 
        nat_add_left_cancel_le zero_le)

lemma ba_defs [simp]:
  " (\<sigma> \<Turnstile> ba f) = (\<forall> i ia . (0 \<le> i \<and> ia+i \<le> intlen \<sigma>) \<longrightarrow> (sub i (ia+i) \<sigma> \<Turnstile> f))"
by (metis ba_d_def da_defs not_defs)

subsection \<open>Semantics Next and Previous Operators\<close>

lemma skip_defs :
  " (\<sigma>  \<Turnstile> skip) = ( (intlen \<sigma>) =1) "
by simp

lemma next_defs [simp]:
  " (\<sigma> \<Turnstile> \<circle> f) = ( (intlen \<sigma>) >0 \<and> ((suffix 1 \<sigma>) \<Turnstile> f))"  
using Suc_le_eq by (simp add: next_d_def) force

lemma wnext_defs [simp]:
  " (\<sigma> \<Turnstile> wnext f) = ((intlen \<sigma>) =0 \<or> ((suffix 1 \<sigma>) \<Turnstile> f))"
by (simp add: wnext_d_def)

lemma prev_defs [simp]:
  " (\<sigma> \<Turnstile> prev f) = ( (intlen \<sigma>) >0 \<and> ((prefix ((intlen \<sigma>)-1) \<sigma>) \<Turnstile> f))"
by (simp add: prev_d_def) 
   (metis One_nat_def Suc_leI diff_diff_cancel diff_is_0_eq' diff_le_self 
          interval_suffix_length_good neq0_conv zero_neq_one)

lemma wprev_defs [simp]:
  " (\<sigma> \<Turnstile> wprev f) = ( (intlen \<sigma>) =0 \<or> ((prefix ((intlen \<sigma>)-1) \<sigma>) \<Turnstile> f))"
by (simp add: wprev_d_def)

subsection \<open>Semantics More and Empty\<close>

lemma more_defs [simp] :
  "(\<sigma> \<Turnstile> more) = (intlen \<sigma> > 0)"
by (simp add: more_d_def)

lemma empty_defs [simp]: 
  "(\<sigma> \<Turnstile> empty) = ( (intlen \<sigma>) = 0) "
by (simp add: empty_d_def)

subsection \<open>Semantics Initial and Final Operators\<close>

lemma init_defs [simp]:
  " (\<sigma> \<Turnstile> init f) = ( (St (intfirst \<sigma>)) \<Turnstile> f)"
by (simp add: init_d_def) auto

lemma fin_defs [simp]: 
  " (\<sigma> \<Turnstile> fin f) = ( (St (intlast \<sigma> )) \<Turnstile> f) "
by (simp add: fin_d_def interval_nth_intlen_intlast)
   
subsection \<open>Semantics Programming Constructs\<close>

lemma ifthenelse_defs [simp]:
 " (\<sigma> \<Turnstile> if\<^sub>i f0 then f1 else f2) = 
  ( ((\<sigma> \<Turnstile> f0) \<and> (\<sigma> \<Turnstile> f1)) \<or> ( \<not>(\<sigma> \<Turnstile>f0) \<and> (\<sigma> \<Turnstile> f2)))"
by (simp add: ifthenelse_d_def)

lemma len_defs [simp]:
  "(\<sigma> \<Turnstile> len n) = ((intlen \<sigma>) = n)"
by (induct n arbitrary: \<sigma>, simp, simp) fastforce
 
subsection \<open>Soundness Axioms\<close>

subsubsection \<open>ChopAssoc\<close>

lemma ChopAssocSemHelp:
 "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h)) =
  (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
by (smt Nat.le_diff_conv2 add_diff_cancel_left' interval_pref_pref_3 interval_suffix_prefix_swap 
        le_add1 le_add_diff_inverse2 le_trans)

lemma ChopAssocSemHelp2:
 " (\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h)"
proof -
 have " (\<sigma> \<Turnstile>  f ; (g ; h)) =
        ((\<exists>i\<le>intlen \<sigma>. (prefix i \<sigma> \<Turnstile> f) \<and> (\<exists>ia\<le>intlen (suffix i \<sigma>). 
          (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))))" 
 by simp_all
 also have " ... = 
             (\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
         (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
 by fastforce
 also have " ... =
            (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and>
            (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
 using ChopAssocSemHelp[of \<sigma> f g h]  by blast
 also have " ... =
            (\<exists>i\<le>intlen \<sigma>. (\<exists>ia\<le>intlen (prefix i \<sigma>).  (prefix ia (prefix i \<sigma>) \<Turnstile> f) \<and>
             (suffix ia (prefix i \<sigma>) \<Turnstile> g)) \<and> (suffix i \<sigma> \<Turnstile> h)) " 
 by fastforce
 also have " ... =
             (\<sigma> \<Turnstile> (f;g);h)" by simp_all
 finally show "(\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h) " .
qed

lemma ChopAssocSem:
  " (\<sigma> \<Turnstile>  f ; (g ; h) \<equiv>\<^sub>i  (f;g);h)"
using ChopAssocSemHelp2 using iff_defs by blast

subsubsection \<open>OrChopImp\<close>

lemma OrChopImpSem:
  " (\<sigma> \<Turnstile>  ( f \<or>\<^sub>i g);h   \<supset>\<^sub>i  f;h \<or>\<^sub>i g;h ) "
by simp_all blast

subsubsection \<open>ChopOrImp\<close>

lemma ChopOrImpSem:
  " (\<sigma> \<Turnstile> f;(g \<or>\<^sub>i h) \<supset>\<^sub>i   f;g \<or>\<^sub>i  f;h ) "
by simp_all blast

subsubsection \<open>EmptyChop\<close>

lemma EmptyChopSem:
  " (\<sigma> \<Turnstile> empty ; f \<equiv>\<^sub>i f ) "
by simp_all auto

subsubsection \<open>ChopEmpty\<close>

lemma ChopEmptySem:
  " (\<sigma> \<Turnstile> f;empty \<equiv>\<^sub>i f ) "
by simp_all auto

subsubsection \<open>StateImpBi\<close>

lemma StateImpBiSem:
  " (\<sigma> \<Turnstile> init f \<supset>\<^sub>i   bi (init f) ) "
by simp_all

subsubsection \<open>NextImpNotNextNot\<close>

lemma NextImpNotNextNotSem:
  " (\<sigma> \<Turnstile> \<circle> f \<supset>\<^sub>i  \<not>\<^sub>i (\<circle> \<not>\<^sub>i f) ) "
by auto

subsubsection \<open>BiBoxChopImpChop\<close>

lemma BiBoxChopImpChopSem:
  " (\<sigma> \<Turnstile> bi ( f \<supset>\<^sub>i f1) \<and>\<^sub>i \<box>(g \<supset>\<^sub>i g1) \<supset>\<^sub>i f;g \<supset>\<^sub>i f1;g1 ) "
by fastforce

subsubsection \<open>BoxInduct\<close>

lemma box_induct_help_1 :
  "(\<sigma> \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (suffix i \<sigma> \<Turnstile> f) \<longrightarrow> (suffix (Suc i) \<sigma> \<Turnstile> f)) 
     \<Longrightarrow> (\<forall> j. j \<le> intlen \<sigma> \<longrightarrow> (suffix j \<sigma> \<Turnstile> f)) "
proof
    fix j
    show " (\<sigma> \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (suffix i \<sigma> \<Turnstile> f) \<longrightarrow> (suffix (Suc i) \<sigma> \<Turnstile> f)) 
          \<Longrightarrow>  j \<le> intlen \<sigma> \<longrightarrow> (suffix j \<sigma> \<Turnstile> f)" 
     proof
       (induct j arbitrary: \<sigma>)
       case 0
       then show ?case by simp
       next
       case (Suc j)
       then show ?case 
       by (metis Nat.le_diff_conv2 One_nat_def Suc_eq_plus1_left Suc_leD)
       qed
qed

lemma BoxInductSem:
  " (\<sigma> \<Turnstile> \<box> (f \<supset>\<^sub>i wnext f) \<and>\<^sub>i f \<supset>\<^sub>i \<box> f)"
apply (simp)
using box_induct_help_1 by (metis One_nat_def diff_self_eq_0 not_one_le_zero)
 
subsubsection \<open>ChopStarEqv\<close>

lemma chopstar_help_1:
  "( \<exists>l. l= \<langle>0\<rangle> \<and> index_sequence 0 l \<and> 
         Interval.nth l (intlen l) = (intlen \<sigma>) \<and> 
         (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )) \<longleftrightarrow> (intlen \<sigma> =0) " 
by (simp add: index_sequence_def)

lemma chopstar_help_2:
 "(\<forall>i. (0<i \<and> i< 1+(intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i-1)) (nth ls ((i-1)+1)) \<sigma>) \<Turnstile> f)
                           ) =
  (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )"
by (metis Suc_eq_plus1 Suc_pred add_diff_cancel_right' add_less_cancel_left 
          add_nonneg_pos le_add2 le_add_same_cancel2 plus_1_eq_Suc zero_less_one)

lemma chop_power_chain:
  "(\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                ) =
   (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    ))
         ) "
proof -
 have " (\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
                       (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                )
      = 
      (\<exists> x ls l.  (intlen l) = (Suc n) \<and> l=x\<odot>ls \<and> index_sequence 0 l \<and> 
                  (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
               )" 
 using interval_intlen_cons_1 by (metis zero_less_Suc)
 also have " ... =
            (\<exists> x ls l.  (intlen l) = (Suc n) \<and> l=x\<odot>ls \<and> index_sequence 0 (x\<odot>ls) \<and> 
                          (nth (x\<odot>ls) (intlen (x\<odot>ls))) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f)
                           )
                     )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> index_sequence 0 (x\<odot>ls) \<and> 
                        (nth (x\<odot>ls) (intlen (x\<odot>ls))) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f)
                           )
                  )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence 0 (x \<odot> ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                           ((\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                 )" 
 by (simp add: index_sequence_def)
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                       ((\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )" 
 using interval_idx_cons by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                        ((sub (nth (x\<odot>ls) 0) (nth (x\<odot>ls) (1)) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"
 by (metis (no_types, lifting) One_nat_def add.right_neutral add_Suc add_Suc_right 
           add_cancel_right_left interval_intlen_cons not_gr_zero zero_le zero_less_Suc)
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> (nth (x\<odot>ls) 0) = x \<and> (nth (x\<odot>ls) (1)) = (nth ls 0) \<and>
                        ((sub (nth (x\<odot>ls) 0) (nth (x\<odot>ls) (1)) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"  
 by auto
 also have " ... = 
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                          (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> (nth (x\<odot>ls) 0) = x \<and> (nth (x\<odot>ls) (1)) = (nth ls 0) \<and>
                        ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                        ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 ( x < (nth ls 0) \<and> 
                 ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                 \<and>
                  (\<forall>i. (0<i \<and> i< 1+(intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i-1)) (nth ls ((i-1)+1)) \<sigma>) \<Turnstile> f)
                           )))" 
 using interval_nth_cons by metis
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 (x < (nth ls 0) \<and> 
                 ((sub x (nth ls 0) \<sigma>) \<Turnstile> f))
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 using chopstar_help_2 by metis
 also have " ... =
            (\<exists>  ls . (intlen ls) = n \<and> index_sequence (nth ls 0) (ls) \<and>
                     (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 (0 < (nth ls 0) \<and> 
                 ((sub 0 (nth ls 0) \<sigma>) \<Turnstile> f))
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 by simp
 also have " ... =
            (\<exists>  lsk .  (intlen lsk) = n \<and> (nth lsk 0) \<le> intlen \<sigma> \<and> (nth lsk 0) > 0 \<and>
                    ((sub 0 (nth lsk 0) \<sigma>) \<Turnstile> f) \<and>
                    index_sequence (nth lsk 0) (lsk) \<and>
                    (nth (lsk) (intlen (lsk))) = (intlen \<sigma>) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub (nth lsk (i)) (nth lsk ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 by (metis Suc_eq_plus1 Suc_pred add.left_neutral eq_iff interval_idx_less_last 
           interval_intlen_gr_zero le_neq_implies_less lessI less_imp_le_nat)
 also have " ... =
            (\<exists> k lsk.  (intlen lsk) = n \<and> (nth lsk 0) \<le> intlen \<sigma> \<and> 
                       (nth lsk 0) > 0 \<and> k=(nth lsk 0) \<and>
                    (sub 0 (nth lsk 0) \<sigma> \<Turnstile> f) \<and> 
                    index_sequence (nth lsk 0) (lsk) \<and> 
                    (nth (lsk) (intlen (lsk))) = (intlen (\<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    )
         )"
 by auto
 also have " ... =
            (\<exists> k lsk.  (intlen lsk) = n \<and> 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and> k=(nth lsk 0) \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (index_sequence k (lsk) \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )" 
 apply (simp add: interval_prefix_suffix_intlen interval_suffix_length interval_prefix_length)
 by auto
 also have " ... =
            (\<exists> k lsk. (intlen lsk) = n \<and>  0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (index_sequence k (lsk) \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         ) "
 using index_sequence_def by auto
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n \<and> index_sequence k (lsk) \<and>
                           ls = map (shiftm k) lsk \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 by blast
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n \<and> index_sequence k (lsk) \<and>
                           ls = map (shiftm k) lsk \<and> 
                           index_sequence 0 (ls) \<and> (intlen ls) = n \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )" 
 using interval_idx_link_shiftm by blast
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n  \<and> index_sequence k (lsk) \<and>
                          lsk = map (shift k) ls \<and> 
                          index_sequence 0 (ls) \<and> (intlen ls) = n  \<and>
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 using interval_lsk_ls by blast
 also have " ... =
            (\<exists> k ls lsk .   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               ( (intlen lsk) = n  \<and> lsk = map (shift k) ls \<and> 
                           index_sequence 0 (ls) \<and>
                           index_sequence k (lsk) \<and>
                          (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 apply (simp add: Interval.shift_def interval_intlen_map interval_nth_map) by blast
 also have " ... =
            (\<exists> k ls lsk.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               ( (intlen lsk) = n  \<and> lsk = map (shift k) ls \<and>
                          (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                          (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 using interval_idx_link by blast
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         ) "
 by (simp add: interval_intlen_map)
 also have " ... =
            (\<exists> k . 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i\<le>intlen ls. Interval.nth ls i \<le> intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    )
               )
         ) " 
 using interval_idx_bound_1 by blast
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                   (\<forall>i\<le>intlen ls. Interval.nth ls i \<le> intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    )
               )
         ) " 
 by (smt add.commute index_sequence_def interval_idx_expand interval_sub_suffix 
         interval_suffix_length_good plus_1_eq_Suc)
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    ))
         ) " 
 using interval_idx_bound_1 by blast
 finally show "(\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
                  (nth l (intlen l)) = (intlen \<sigma>) \<and>
                  (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                        ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                  )
                ) =
                (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and> (sub 0 k \<sigma> \<Turnstile> f) \<and> 
                  (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                     (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and> 
                     (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                        ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    )
                  )
                )
               " .
qed

lemma chop_power_eqv_sem:
  "(\<exists> n. (\<sigma> \<Turnstile> (power_chop_d f n))) = 
   ((\<sigma> \<Turnstile> empty) \<or>  (\<exists> n. (\<sigma> \<Turnstile> (f \<and>\<^sub>i more); (power_chop_d f n))))"
by (metis not0_implies_Suc power_chop_d.power_0 power_chop_d.power_Suc)

lemma chopstar_eqv_power_chop_help:
  " ( \<sigma> \<Turnstile> power_chop_d f n) = 
    (\<exists> (l::index). intlen(l) = n \<and>  index_sequence 0 l  \<and> 
      (nth l (intlen l)) = (intlen ( \<sigma>)) \<and>
      (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
            ((sub (nth l i) (nth l (i+1)) (\<sigma>)) \<Turnstile> f)
      )
    ) "
proof 
 (induct n arbitrary: \<sigma>)
 case 0
 then show ?case using index_sequence_def chopstar_help_1 by fastforce
 next
 case (Suc n)
 then show ?case 
  proof -
   have 1: "(\<sigma> \<Turnstile> power_chop_d f (Suc n)) = (\<sigma> \<Turnstile> ((f \<and>\<^sub>i more);(power_chop_d f n)))  " 
   by simp
   have 2: "(\<sigma> \<Turnstile> ((f \<and>\<^sub>i more);(power_chop_d f n))) =
             (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (prefix k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power_chop_d f n)
             )
           " 
  by auto
  have 3: "(\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (prefix k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power_chop_d f n)
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power_chop_d f n)
           )" 
  by (simp add: interval_sub_zero_prefix)
  have 4: "(\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power_chop_d f n)
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (\<exists> (l::index). intlen(l) = n \<and>  index_sequence 0 l  \<and> 
                  (nth l (intlen l)) = (intlen (suffix k \<sigma>)) \<and>
                  (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                       ((sub (nth l i) (nth l (i+1)) (suffix k \<sigma>)) \<Turnstile> f)
                  )
                )
             )" 
   using Suc.hyps by auto
   have 5: "
           (\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
              (nth l (intlen l)) = (intlen \<sigma>) \<and>
              (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                   ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
              )
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                  (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and> 
                  (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                        ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                  )
               )
           )" 
    using chop_power_chain by simp
    from 1 2 3 4 5 show ?thesis by auto
  qed
qed

lemma chopstar_eqv_power_chop:
  "(\<sigma> \<Turnstile> f\<^sup>\<star>) = (\<exists> k. (\<sigma> \<Turnstile> power_chop_d f k)) "
by (simp add: chopstar_eqv_power_chop_help)

lemma ChopstarEqvSem:
  " (\<sigma> \<Turnstile> f\<^sup>\<star> \<equiv>\<^sub>i  empty \<or>\<^sub>i  (f \<and>\<^sub>i more); f\<^sup>\<star> )"
using chopstar_eqv_power_chop 
by (smt chop_power_eqv_sem iff_defs or_defs semantics_pitl.simps(5))

subsection \<open>Time Reversal\<close>

lemma time_reverse_help_1:
 assumes "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and> 
          ls = map (\<lambda> x. (intlen \<sigma>) -x) (intrev l) \<and> index_sequence 0 ls "
 shows   "(\<forall> i.  0\<le> i \<and> i<intlen ls \<longrightarrow> 
          (sub (nth ls ((intlen ls)-(i+1))) ((nth ls ((intlen ls) -i)) ) (intrev \<sigma>) \<Turnstile> f\<^sup>r))
          =
          (\<forall> i. 0\<le>i \<and> i<intlen ls \<longrightarrow> (sub (nth ls (i)) ((nth ls (i+1)) ) (intrev \<sigma>) \<Turnstile> f\<^sup>r)) " 
by (smt Suc_diff_Suc Suc_eq_plus1 add_gr_0 diff_diff_cancel diff_less le_add_diff_inverse2 
        le_simps(1) zero_less_one zero_order(1))
         
lemma TimeReverseSem:
 " (\<sigma> \<Turnstile> f) \<longleftrightarrow> ((intrev \<sigma>) \<Turnstile> f\<^sup>r)"
proof
 (induct f arbitrary: \<sigma>)
 case false_d
 then show ?case by auto
 next
 case (atom_d x)
 then show ?case by (simp add: interval_intlast_intrev)
 next
 case (implies_d f1 f2)
 then show ?case by simp
 next
 case skip_d
 then show ?case by simp
 next
 case (chop_d f1 f2)
 then show ?case using interval_intrev_prefix interval_intrev_suffix 
 by (smt interval_intlen_gr_zero interval_intrev_intlen interval_prefix_length 
         interval_rev_rev_ident interval_suffix_length_good order_refl rev_d.simps(5) 
         semantics_pitl.simps(5))
 next
 case (chopstar_d f)
 then show ?case 
  proof -
   have "(\<sigma> \<Turnstile>  f \<^sup>\<star>) = 
            (\<exists> l. index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                          ) " 
   by simp
   also have " ... =
              (\<exists> l. index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               (intrev (sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f\<^sup>r)
                           )
                          ) " 
   using chopstar_d.hyps by simp
   also have " ... =
             (\<exists> l. index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                      ( (sub ((intlen \<sigma>)- (nth l (i+1))) ((intlen \<sigma>) - (nth l i)) (intrev \<sigma>)) \<Turnstile> f\<^sup>r)
                           )
                          )" 
   using interval_intrev_idx_2 by metis
   also have " ... =
            (\<exists> l ls. ls = map (\<lambda> x. (intlen \<sigma>) -x) (intrev l) \<and> index_sequence 0 ls \<and>
                     index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                      ( (sub ((intlen \<sigma>)- (nth l (i+1))) ((intlen \<sigma>) - (nth l i)) (intrev \<sigma>)) \<Turnstile> f\<^sup>r)
                           )
                          )  " 
   using interval_intrev_idx_7 by auto
   also have " ... =
            (\<exists> l ls. ls = map (\<lambda> x. (intlen \<sigma>) -x) (intrev l) \<and> 
                     index_sequence 0 ls \<and> index_sequence 0 l  
               \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and> (nth ls (intlen ls)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                      ( (sub ((intlen \<sigma>)- (nth l (i+1))) ((intlen \<sigma>) - (nth l i)) (intrev \<sigma>)) \<Turnstile> f\<^sup>r)
                           )
                          )" 
   by (metis interval_intrev_idx_3)
   also have " ... =
             (\<exists> l ls. ls = map (\<lambda> x. (intlen \<sigma>) -x) (intrev l) \<and> 
                      index_sequence 0 ls \<and> index_sequence 0 l  
               \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and> (nth ls (intlen ls)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                    ( sub (nth ls ((intlen ls)-(i+1))) ((nth ls ((intlen ls) -i)) ) (intrev \<sigma>) \<Turnstile> f\<^sup>r)
                           )
                          )
            " 
   using interval_intrev_idx_9 by metis
   also have "... =
             (\<exists> l ls. ls = map (\<lambda> x. (intlen \<sigma>) -x) (intrev l) \<and> 
                      index_sequence 0 ls \<and> index_sequence 0 l  
               \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and> (nth ls (intlen ls)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ( (sub (nth ls (i)) ((nth ls (i+1)) ) (intrev \<sigma>)) \<Turnstile> f\<^sup>r)
                           )
                          )
            " 
   using time_reverse_help_1 by metis
   also have " ... =
             (\<exists> ls.  index_sequence 0 ls \<and>  
                (nth ls (intlen ls)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ( (sub (nth ls (i)) ((nth ls (i+1)) ) (intrev \<sigma>)) \<Turnstile> f\<^sup>r)
                           )
                          )" 
   using interval_intrev_idx_12 by (smt interval_intrev_idx_3 interval_intrev_idx_7)
   also have " ... = (intrev \<sigma> \<Turnstile> ( f\<^sup>r \<^sup>\<star>) ) " by simp
   also have " ... = (intrev \<sigma> \<Turnstile> ( f \<^sup>\<star>) \<^sup>r) " by simp
   finally show "(\<sigma> \<Turnstile>  f \<^sup>\<star>) =  (intrev \<sigma> \<Turnstile> ( f \<^sup>\<star>) \<^sup>r)" .
 qed
qed

end
