(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Syntax
 imports
   Main
   abbrevs "&_i" = "\<and>\<^sub>i" and
           "|_i" = "\<or>\<^sub>i" and
           ")_i" = "\<supset>\<^sub>i" and
           "-_i" = "\<not>\<^sub>i" and
           "=_i" = "\<equiv>\<^sub>i" and
           "false_i" = "false\<^sub>i" and
           "true_i"  = "true\<^sub>i" and
           "if_i" = "if\<^sub>i" and
           "_*"  = "\<^sup>\<star>"
begin

section \<open>Syntax\<close>

text {*
The basic ITL syntax is introduced first followed by the derived operators including  
the time reversal operator (\cite{moszkowski14:_compos}). 

*}


subsection \<open>Primitive formulae\<close>

datatype 'a pitl =
    false_d                          ("false\<^sub>i")
  | atom_d "'a \<Rightarrow> bool"              ("(atom\<^sub>i _ )") 
  | implies_d "'a pitl" "'a pitl"    ("( _ \<supset>\<^sub>i _ )" [26,25] 25)
  | skip_d                           ("skip")
  | chop_d "'a pitl" "'a pitl"       ("( _ ; _ )" [84,84] 83)
  | chopstar_d "'a pitl"             ("( _ \<^sup>\<star>)" [85] 85)

subsection \<open>Derived Boolean Operators\<close>

definition not_d ("(\<not>\<^sub>i _)" [90] 90) 
where 
  "\<not>\<^sub>i f \<equiv> f \<supset>\<^sub>i false\<^sub>i"

definition true_d ("true\<^sub>i") 
where 
  "true\<^sub>i \<equiv> \<not>\<^sub>i false\<^sub>i"

definition or_d ("(_ \<or>\<^sub>i _)" [31,30] 30) 
where 
  "f \<or>\<^sub>i g \<equiv> \<not>\<^sub>i f \<supset>\<^sub>i g"

definition and_d ("(_ \<and>\<^sub>i _)" [36,35] 35)
where 
  "f \<and>\<^sub>i g \<equiv> \<not>\<^sub>i (\<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i g)"

definition iff_d ( "(_ \<equiv>\<^sub>i _) " [21,20] 20)
where 
  "f \<equiv>\<^sub>i g \<equiv> ( ( f \<supset>\<^sub>i g) \<and>\<^sub>i (g \<supset>\<^sub>i f))" 
 

subsection \<open>Next and Previous Operators\<close>

definition next_d ("(\<circle> _)" [88] 87)
where
  "\<circle> f \<equiv> skip ; f"

definition wnext_d ("(wnext _)" [88] 87)
where
 "wnext f \<equiv> \<not>\<^sub>i (\<circle> (\<not>\<^sub>i f))" 

definition prev_d ("(prev _)" [88] 87)
where
 "prev f \<equiv> f;skip"

definition wprev_d ("(wprev _)" [88] 87)
where
 "wprev f \<equiv> \<not>\<^sub>i(prev(\<not>\<^sub>i f))" 


subsection \<open>More and Empty\<close>

definition more_d ("more") 
where 
  "more \<equiv> \<circle> true\<^sub>i"

definition empty_d ("empty")
where
 "empty \<equiv> \<not>\<^sub>i more"

subsection \<open>Box and Diamond Operators\<close>

definition sometimes_d ("(\<diamond> _)" [88] 87)
where 
  "\<diamond> f \<equiv> true\<^sub>i ; f"

definition always_d ("(\<box> _)" [88] 87)
where 
  "\<box> f \<equiv> \<not>\<^sub>i (\<diamond> (\<not>\<^sub>i f))"

definition di_d ("(di _)" [88] 87)
where 
  "di f \<equiv> f ; true\<^sub>i"

definition bi_d ("(bi _)" [88] 87)
where 
  "bi f \<equiv> \<not>\<^sub>i (di (\<not>\<^sub>i f))"

definition da_d ("(da _)" [88] 87)
where 
  "da f \<equiv> true\<^sub>i ; (f ; true\<^sub>i)"

definition ba_d ("(ba _)" [88] 87)
where 
  "ba f \<equiv> \<not>\<^sub>i (da (\<not>\<^sub>i f))"

definition dm_d ("(dm _)" [88] 87)
where 
  "dm f \<equiv> \<diamond> (more \<and>\<^sub>i f)"

definition bm_d ("(bm _)" [88] 87)
where 
  "bm f \<equiv> \<not>\<^sub>i (dm (\<not>\<^sub>i f))"

subsection \<open>Initial and Final Operators\<close>

definition init_d ("(init _)" [88] 87)
where
 " init f \<equiv> ((empty \<and>\<^sub>i f); true\<^sub>i)"

definition fin_d ("(fin _)" [88] 87)
where
  "fin f \<equiv> \<box> ( empty \<supset>\<^sub>i f)"

subsection \<open>Programming Constructs\<close>

definition halt_d ("(halt _)" [88] 87)
where
  "halt f \<equiv> \<box> ( empty \<equiv>\<^sub>i f)"

definition keep_d ("(keep _)" [88] 87)
where
  "keep f \<equiv> ba ( skip \<supset>\<^sub>i f)"

definition yields_d ("( _ yields _)" [88,88] 87)
where
 " f yields g \<equiv> \<not>\<^sub>i ( f; \<not>\<^sub>i g)"

definition ifthenelse_d  ("(if\<^sub>i _ then _ else _)" [88,88,88] 87)
where
  "if\<^sub>i f then g else h \<equiv> (f \<and>\<^sub>i g) \<or>\<^sub>i ( (\<not>\<^sub>i f) \<and>\<^sub>i h)"

definition ifthen_d  ("(if\<^sub>i _ then _)" [88,88] 87)
where
  "if\<^sub>i f then g  \<equiv> if\<^sub>i f then g else true\<^sub>i"
 
definition while_d ("(while _ do _)" [88,88] 87)
where
  "while f do g \<equiv> (f \<and>\<^sub>i g)\<^sup>\<star> \<and>\<^sub>i fin ( \<not>\<^sub>i f)"

definition repeat_d ("(repeat _ until _)" [88,88] 87)
where
  "repeat f until g \<equiv> f ; (while (\<not>\<^sub>i g) do f)"

primrec len\<^sub>_d :: "nat \<Rightarrow> 'a pitl"  ("(len _)" [88] 87) where
   "(len 0) = empty" 
 | "(len (Suc n)) = (skip ; (len n))" 
  
primrec power_chop_d :: " 'a pitl \<Rightarrow> nat \<Rightarrow> 'a pitl" where
    power_0  : " power_chop_d f 0       = empty"
  | power_Suc: " power_chop_d f (Suc n) = ((f \<and>\<^sub>i more);(power_chop_d f n)) "

primrec power_d :: " 'a pitl \<Rightarrow> nat \<Rightarrow> 'a pitl" ("(power _ _)" [88,88] 87) where
    pow_0  : " power_d f 0       = empty"
  | pow_Suc: " power_d f (Suc n) = ((f);(power_d f n)) "

subsection \<open>Time reversal\<close>

primrec rev_d :: "'a pitl \<Rightarrow> 'a pitl " ("(_ \<^sup>r)" [88] 87)
where
   " false\<^sub>i\<^sup>r = false\<^sub>i"
|  " (atom\<^sub>i p)\<^sup>r = fin (atom\<^sub>i p)"
|  " (f1 \<supset>\<^sub>i f2)\<^sup>r = ( f1\<^sup>r \<supset>\<^sub>i f2\<^sup>r )"
|  " skip\<^sup>r = skip"
|  " (f1 ; f2)\<^sup>r = ( f2\<^sup>r ; f1\<^sup>r )"
|  " (f\<^sup>\<star>)\<^sup>r = (f\<^sup>r)\<^sup>\<star>" 
   
end
