(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory First
 imports
   Theorems
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
declare [[smt_timeout=120]]
*)

section \<open>The First Occurrence Operator in ITL\<close>

text {*
Runtime verification (RV) has gained significant interest in recent years. 
The behaviour of a program can be verified in real time by analysing its evolving trace. 
This approach has two significant benefits over static verification techniques such as model 
checking. Firstly, it is only necessary to verify actual execution paths rather than all 
possible paths. Secondly, it is possible to react at runtime should the program diverge from 
its specified behaviour. RV does not replace traditional verification techniques but it 
does provide an extra layer of security.

Linear Temporal Logic (LTL) is a popular formalism for writing specifications from which RV 
monitors can be derived automatically. By contrast, Interval Temporal Logic (ITL) has not 
been as widely represented in this field despite being more expressive and 
compositional. The principal issue is efficiency. ITL uses non-deterministic operators to 
construct sequential and iterative specifications (chop and chop-star, respectively) and 
these introduce combinatorial complexity. Approaches to mitigate this include using a 
deterministic subset of ITL or adapting the semantics to include a deterministic chop operator.
 This thesis proposes an alternative approach, wholly within existing ITL, and based upon a new,
 derived operator called “first occurrence”.

A theory of first occurrence is developed and used to derive an algebra of RV monitors.

*}

subsection \<open>Definitions\<close>

subsubsection \<open>Definitions Strict Initial and Final\<close>

definition bs_d ("(bs _)" [88] 87)
where 
  "bs f \<equiv> (empty \<or>\<^sub>i ((bi f) ; skip))" 

definition bt_d ("(bt _)" [88] 87)
where 
  "bt f \<equiv> (empty \<or>\<^sub>i (skip;(\<box> f)))"

definition ds_d ("(ds _)" [88] 87)
where 
  "ds f \<equiv> \<not>\<^sub>i (bs (\<not>\<^sub>i f))"  

definition dt_d ("(dt _)" [88] 87)
where
  "dt f \<equiv> \<not>\<^sub>i (bt (\<not>\<^sub>i f))"

subsubsection \<open>Definition First and Last Operators\<close>

definition first_d ("(\<rhd> _)" [88] 87)
where   
  "\<rhd> f \<equiv>  (f \<and>\<^sub>i  (bs (\<not>\<^sub>i f)))"

definition last_d ("(\<lhd> _)" [88] 87)
where   
  "\<lhd> f \<equiv>  (f \<and>\<^sub>i  (bt (\<not>\<^sub>i f)))"


subsection \<open>First and Time Reversal\<close>

lemma BsEqvRule:
 assumes "\<turnstile> f \<equiv>\<^sub>i g"
 shows   "\<turnstile> bs f \<equiv>\<^sub>i bs g"
proof -
 have 1: "\<turnstile> f \<equiv>\<^sub>i g"  using assms by auto
 hence 2: "\<turnstile> bi(f) \<equiv>\<^sub>i bi(g)" by simp
 hence 3: "\<turnstile> bi(f);skip  \<equiv>\<^sub>i bi(g);skip" by auto
 hence 4: "\<turnstile> empty \<or>\<^sub>i bi(f);skip  \<equiv>\<^sub>i empty \<or>\<^sub>i bi(g);skip" by auto
 hence 5: "\<turnstile> bs(f) \<equiv>\<^sub>i bs(g)" by (simp add: bs_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma BtEqvRule:
 assumes "\<turnstile> f \<equiv>\<^sub>i g"
 shows   "\<turnstile> bt f \<equiv>\<^sub>i bt g"
proof -
 have 1: "\<turnstile> f \<equiv>\<^sub>i g"  using assms by auto
 hence 2: "\<turnstile> \<box>(f) \<equiv>\<^sub>i \<box>(g)" by simp
 hence 3: "\<turnstile> skip;\<box>(f)  \<equiv>\<^sub>i skip;\<box>(g)" using RightChopEqvChop by blast
 hence 4: "\<turnstile> empty \<or>\<^sub>i skip;\<box>(f)  \<equiv>\<^sub>i empty \<or>\<^sub>i skip;\<box>(g)" by auto
 hence 5: "\<turnstile> bt(f) \<equiv>\<^sub>i bt(g)" by (simp add: bt_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma FstEqvRule:
 assumes "\<turnstile> f \<equiv>\<^sub>i g"
 shows   "\<turnstile> \<rhd>f \<equiv>\<^sub>i \<rhd>g "
proof -
 have 1: "\<turnstile> f \<equiv>\<^sub>i g" using assms by auto
 hence 2: "\<turnstile> \<not>\<^sub>i f \<equiv>\<^sub>i \<not>\<^sub>i g" by auto
 hence 3: "\<turnstile> bs(\<not>\<^sub>i f) \<equiv>\<^sub>i bs(\<not>\<^sub>i g)" by (simp add: bs_d_def)
 hence 4: "\<turnstile> f \<and>\<^sub>i bs(\<not>\<^sub>i f) \<equiv>\<^sub>i g \<and>\<^sub>i bs(\<not>\<^sub>i g) " using "1" by auto
 from 4 show ?thesis by (simp add:first_d_def)
qed

lemma LstEqvRule:
 assumes "\<turnstile> f \<equiv>\<^sub>i g"
 shows   "\<turnstile> \<lhd>f \<equiv>\<^sub>i \<lhd>g "
proof -
 have 1: "\<turnstile> f \<equiv>\<^sub>i g" using assms by auto
 hence 2: "\<turnstile> \<not>\<^sub>i f \<equiv>\<^sub>i \<not>\<^sub>i g" by auto
 hence 3: "\<turnstile> bt(\<not>\<^sub>i f) \<equiv>\<^sub>i bt(\<not>\<^sub>i g)" by (simp add: bt_d_def)
 hence 4: "\<turnstile> f \<and>\<^sub>i bt(\<not>\<^sub>i f) \<equiv>\<^sub>i g \<and>\<^sub>i bt(\<not>\<^sub>i g) " using "1" by auto
 from 4 show ?thesis by (simp add:last_d_def)
qed

lemma RBsEqvBt:
 "\<turnstile> (bs f)\<^sup>r \<equiv>\<^sub>i (bt (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (bs f)\<^sup>r \<equiv>\<^sub>i (empty \<or>\<^sub>i ((bi f) ; skip))\<^sup>r" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> (empty \<or>\<^sub>i ((bi f) ; skip))\<^sup>r \<equiv>\<^sub>i (empty\<^sup>r \<or>\<^sub>i ((bi f) ; skip)\<^sup>r)" 
      using ROr by blast
 have 3: "\<turnstile> (empty\<^sup>r \<or>\<^sub>i ((bi f) ; skip)\<^sup>r) \<equiv>\<^sub>i (empty \<or>\<^sub>i (skip\<^sup>r;(bi f)\<^sup>r))" 
      using REmptyEqvEmpty by auto
 have 4: "\<turnstile> (empty \<or>\<^sub>i (skip\<^sup>r;(bi f)\<^sup>r)) \<equiv>\<^sub>i (empty \<or>\<^sub>i (skip;\<box> (f\<^sup>r)))" 
      by (metis (no_types, lifting) NextEqvNext RBiEqvBox REmptyEqvEmpty next_d_def 
                   or_d_def prop01 prop21 prop39 rev_d.simps(4))
 have 5: "\<turnstile> (empty \<or>\<^sub>i (skip;\<box> (f\<^sup>r))) \<equiv>\<^sub>i (bt (f\<^sup>r))" 
      by (simp add: bt_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma RRBsEqvBt:
 "\<turnstile> (bs (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i (bt (f))"
proof -
 have 1: "\<turnstile> (bs (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i bt ((f\<^sup>r)\<^sup>r)" using RBsEqvBt by blast
 have 2: "\<turnstile> bt ((f\<^sup>r)\<^sup>r) \<equiv>\<^sub>i bt f" using EqvReverseReverse using BtEqvRule by blast
 from 1 2 show ?thesis by auto
qed

lemma RBtEqvBs:
 "\<turnstile> (bt f)\<^sup>r \<equiv>\<^sub>i (bs (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (bt f)\<^sup>r \<equiv>\<^sub>i (empty \<or>\<^sub>i (skip;\<box> f))\<^sup>r " 
      by (simp add: bt_d_def)
 have 2: "\<turnstile> (empty \<or>\<^sub>i (skip;\<box> f))\<^sup>r \<equiv>\<^sub>i (empty\<^sup>r \<or>\<^sub>i (skip;\<box> f)\<^sup>r)"
      using ROr by blast
 have 3: "\<turnstile> (empty\<^sup>r \<or>\<^sub>i (skip;\<box> f)\<^sup>r) \<equiv>\<^sub>i (empty \<or>\<^sub>i (\<box> f)\<^sup>r;skip\<^sup>r)" 
      using REmptyEqvEmpty by auto
 have 4: "\<turnstile> (empty \<or>\<^sub>i (\<box> f)\<^sup>r;skip\<^sup>r) \<equiv>\<^sub>i (empty \<or>\<^sub>i (bi (f\<^sup>r));skip)" 
      by (metis (no_types, lifting) LeftChopEqvChop RBoxEqvBi REmptyEqvEmpty 
                   or_d_def prop01 prop21 prop39 rev_d.simps(4))
 have 5: "\<turnstile> (empty \<or>\<^sub>i (bi (f\<^sup>r));skip) \<equiv>\<^sub>i (bs (f\<^sup>r))" 
      by (simp add: bs_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma RRBtEqvBs:
 "\<turnstile> (bt (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i (bs (f))"
proof -
 have 1: "\<turnstile> (bt (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i bs ((f\<^sup>r)\<^sup>r)" using RBtEqvBs by blast
 have 2: "\<turnstile> bs ((f\<^sup>r)\<^sup>r) \<equiv>\<^sub>i bs f" using EqvReverseReverse using BsEqvRule by blast
 from 1 2 show ?thesis by auto
qed

lemma RFirstEqvLast:
 "\<turnstile> (\<rhd> f)\<^sup>r \<equiv>\<^sub>i (\<lhd> (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (\<rhd> f)\<^sup>r \<equiv>\<^sub>i (f \<and>\<^sub>i bs(\<not>\<^sub>i f))\<^sup>r" by (simp add: first_d_def)
 have 2: "\<turnstile> (f \<and>\<^sub>i bs(\<not>\<^sub>i f))\<^sup>r \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i (bs (\<not>\<^sub>i f))\<^sup>r)" using RAnd by blast
 have 3: "\<turnstile> (f\<^sup>r \<and>\<^sub>i (bs (\<not>\<^sub>i f))\<^sup>r) \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i bt ((\<not>\<^sub>i f)\<^sup>r))" using RBsEqvBt prop05 by blast
 have 4: "\<turnstile> (f\<^sup>r \<and>\<^sub>i bt ((\<not>\<^sub>i f)\<^sup>r)) \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i bt (\<not>\<^sub>i(f\<^sup>r)))" by (simp add: not_d_def)
 have 5: "\<turnstile> (f\<^sup>r \<and>\<^sub>i bt (\<not>\<^sub>i(f\<^sup>r))) \<equiv>\<^sub>i (\<lhd> (f\<^sup>r))" by (simp add: last_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma RRFirstEqvLast:
 "\<turnstile> (\<rhd> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i (\<lhd> (f))"
proof -
 have 1: "\<turnstile> (\<rhd> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i \<lhd> ((f\<^sup>r)\<^sup>r)" using RFirstEqvLast by blast
 have 2: "\<turnstile> \<lhd> ((f\<^sup>r)\<^sup>r) \<equiv>\<^sub>i \<lhd> f" using EqvReverseReverse using LstEqvRule by blast
 from 1 2 show ?thesis by auto
qed

lemma RLastEqvFirst:
 "\<turnstile> (\<lhd> f)\<^sup>r \<equiv>\<^sub>i (\<rhd> (f\<^sup>r))"
proof -
 have 1: "\<turnstile> (\<lhd> f)\<^sup>r \<equiv>\<^sub>i (f \<and>\<^sub>i bt(\<not>\<^sub>i f))\<^sup>r" by (simp add: last_d_def)
 have 2: "\<turnstile> (f \<and>\<^sub>i bt(\<not>\<^sub>i f))\<^sup>r \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i (bt (\<not>\<^sub>if))\<^sup>r)" using RAnd by blast
 have 3: "\<turnstile> (f\<^sup>r \<and>\<^sub>i (bt (\<not>\<^sub>if))\<^sup>r) \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i bs ( (\<not>\<^sub>if)\<^sup>r))" using RBtEqvBs prop05 by blast
 have 4: "\<turnstile> (f\<^sup>r \<and>\<^sub>i bs ( (\<not>\<^sub>if)\<^sup>r)) \<equiv>\<^sub>i (f\<^sup>r \<and>\<^sub>i bs(\<not>\<^sub>i(f\<^sup>r)))" by (simp add: not_d_def)
 have 5: "\<turnstile> (f\<^sup>r \<and>\<^sub>i bs(\<not>\<^sub>i(f\<^sup>r))) \<equiv>\<^sub>i (\<rhd> (f\<^sup>r))" by (simp add: first_d_def)
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma RRLastEqvFirst:
 "\<turnstile> (\<lhd> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i (\<rhd> (f))"
proof -
 have 1: "\<turnstile> (\<lhd> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i \<rhd> ((f\<^sup>r)\<^sup>r)" using RLastEqvFirst by blast
 have 2: "\<turnstile> \<rhd> ((f\<^sup>r)\<^sup>r) \<equiv>\<^sub>i \<rhd> f" using EqvReverseReverse using FstEqvRule by blast
 from 1 2 show ?thesis by auto
qed


subsection \<open>Semantic Theorems\<close>

subsubsection \<open>Semantics First and Last Operators\<close>

lemma FstAndBisem:
 "(intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi \<not>\<^sub>if;skip)) =
  (intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<intlen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>\<^sub>if)) ) "
apply simp_all
apply (simp add: interval_prefix_length interval_suffix_length)
proof -
 have 1: "(0 < intlen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<exists>i. (i \<le> intlen \<sigma> \<longrightarrow> (\<forall>ia\<le>i. \<not> (prefix ia (prefix i \<sigma>) \<Turnstile> f)) \<and>
           intlen \<sigma> - i = Suc 0) \<and> i \<le> intlen \<sigma>)
          ) = 
          (0 < intlen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<exists>i. (i \<le> intlen \<sigma> \<longrightarrow> (\<forall>ia\<le>i. \<not> (prefix ia (prefix i \<sigma>) \<Turnstile> f)) \<and>
           i = intlen \<sigma> - Suc 0) \<and> i \<le> intlen \<sigma>)
          )" 
      by auto
 also have "... =
          (0 < intlen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<forall>ia\<le>(intlen \<sigma> - Suc 0). \<not> (prefix ia (prefix (intlen \<sigma> - Suc 0) \<sigma>) \<Turnstile> f) )
          )" 
      using diff_le_self by blast
 also have "... =
          (intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and>
            (\<forall>ia<intlen ( \<sigma>). \<not> (prefix ia (prefix (intlen \<sigma> - Suc 0) \<sigma>) \<Turnstile> f))
          ) " by (metis Suc_pred less_Suc_eq_le)
 also have "... =
          (intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and>
           (\<forall>ia<intlen ( \<sigma>). (prefix ia (prefix (intlen \<sigma> - Suc 0) \<sigma>) \<Turnstile> \<not>\<^sub>if))
          ) " 
      using not_defs by blast
 also have "... = 
          (intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<intlen ( \<sigma>). \<not> (prefix ia  \<sigma> \<Turnstile> f))) " 
          by (simp add: interval_pref_pref_help) 
 finally show "(0 < intlen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and>
                (\<exists>i. (i \<le> intlen \<sigma> \<longrightarrow> (\<forall>ia\<le>i. \<not> (prefix ia (prefix i \<sigma>) \<Turnstile> f)) \<and> 
                                 intlen \<sigma> - i = Suc 0) \<and> i \<le> intlen \<sigma>)
               ) =
               (0 < intlen \<sigma> \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<intlen \<sigma>. \<not> (prefix ia \<sigma> \<Turnstile> f))) " .
qed
     
lemma Fstsem_0:
 "(\<sigma> \<Turnstile> \<rhd>f) = 
  ( 
   ( \<sigma> \<Turnstile> f \<and>\<^sub>i empty) \<or> (intlen \<sigma> >0  \<and> (\<sigma> \<Turnstile> f) \<and> ( \<sigma> \<Turnstile>  bi \<not>\<^sub>if;skip)) 
  )  "
apply (simp add: first_d_def bs_d_def) using neq0_conv by blast

lemma Emptysem:
 "(\<sigma> \<Turnstile> f \<and>\<^sub>i empty) = ((\<sigma> \<Turnstile> f) \<and> intlen \<sigma> = 0) "
by simp

lemma Fstsem:
 "(\<sigma> \<Turnstile> \<rhd>f) =
  (  
   ( (\<sigma> \<Turnstile> f) \<and> intlen \<sigma> = 0) \<or>
   ( intlen \<sigma>>0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia<intlen (\<sigma>). (prefix ia  \<sigma> \<Turnstile> \<not>\<^sub>if)))
  ) "
using Fstsem_0 Emptysem FstAndBisem by blast

lemma Lstsem:
 "(\<sigma> \<Turnstile> \<lhd> f) =
  ( ( (\<sigma> \<Turnstile> f) \<and> intlen \<sigma> = 0) \<or>
    ( intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> (\<forall>ia < intlen \<sigma>. (suffix ((intlen \<sigma>) -ia)  \<sigma> \<Turnstile> \<not>\<^sub>if)) )
  ) 
 "
proof -
 have "(\<sigma> \<Turnstile> \<lhd> f) = (\<sigma> \<Turnstile> (\<rhd> (f\<^sup>r))\<^sup>r)"  
      using RRFirstEqvLast iff_defs itl_valid by blast  
 also have " ... = (intrev \<sigma> \<Turnstile> \<rhd> (f\<^sup>r))" 
      by (metis TimeReverseSem interval_rev_rev_ident)
 also have " ... =
  (  
   ( (intrev \<sigma> \<Turnstile> f\<^sup>r) \<and> intlen (intrev \<sigma>) = 0) \<or>
   ( intlen (intrev \<sigma>)>0 \<and> (intrev \<sigma> \<Turnstile> f\<^sup>r) \<and> 
     (\<forall>ia<intlen (intrev \<sigma>). (prefix ia  (intrev \<sigma>) \<Turnstile> \<not>\<^sub>i(f\<^sup>r))))
  )" 
      using Fstsem by blast
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> intlen (\<sigma>) = 0) \<or>
   ( intlen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<intlen (intrev \<sigma>). (prefix ia  (intrev \<sigma>) \<Turnstile> (\<not>\<^sub>i(f))\<^sup>r)))
  )" 
      by (metis TimeReverseSem interval_intrev_intlen not_d_def rev_d.simps(1) rev_d.simps(3))
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> intlen (\<sigma>) = 0) \<or>
   ( intlen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<intlen (intrev \<sigma>). (intrev (prefix ia  (intrev \<sigma>)) \<Turnstile> (\<not>\<^sub>i(f)))))
  )" 
      using TimeReverseSem by (metis interval_rev_rev_ident)
 also have " ... =
  (  
   ( ( \<sigma> \<Turnstile> f) \<and> intlen (\<sigma>) = 0) \<or>
   ( intlen (\<sigma>)>0 \<and> ( \<sigma> \<Turnstile> f) \<and> 
     (\<forall>ia<intlen (\<sigma>). ( (suffix ((intlen \<sigma>) - ia)  (\<sigma>)) \<Turnstile> (\<not>\<^sub>i(f)))))
  )" 
      by (simp add: interval_intrev_prefix)
 finally show " 
  (\<sigma> \<Turnstile> \<lhd> f) =
  ( ( (\<sigma> \<Turnstile> f) \<and> intlen \<sigma> = 0) \<or>
    ( intlen \<sigma> >0 \<and> (\<sigma> \<Turnstile> f) \<and> 
      (\<forall>ia < intlen \<sigma>. (suffix ((intlen \<sigma>) -ia)  \<sigma> \<Turnstile> \<not>\<^sub>if)) )
  )  " .
qed

subsubsection \<open>Various Semantic Lemmas\<close>

lemma DiLensem:
 "(\<sigma> \<Turnstile> di (f \<and>\<^sub>i len(i))) =
   ( (prefix i \<sigma> \<Turnstile> f) \<and> i\<le>intlen \<sigma>) "
apply simp_all 
using interval_prefix_length_good by auto

lemma PrefixFstsem:
 "( (prefix i \<sigma> \<Turnstile> \<rhd>f) \<and> i\<le>intlen \<sigma>) =
  ( i\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 ) "
proof -
 have 1: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f)) =
  (  
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> intlen (prefix i \<sigma>) = 0) \<or>
   ( intlen (prefix i \<sigma>)>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> 
     (\<forall>ia<intlen (prefix i \<sigma>). (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>\<^sub>if)) ) 
  ) " 
       using Fstsem by blast
 hence 2: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>intlen \<sigma>) =
  (  i\<le>intlen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> intlen (prefix i \<sigma>) = 0) \<or>
   ( intlen (prefix i \<sigma>)>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> 
     (\<forall>ia<intlen (prefix i \<sigma>). (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>\<^sub>if)) )
    )
  ) " 
       by auto
 hence 3: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>intlen \<sigma>) =
  (  i\<le>intlen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> i = 0) \<or>
   ( i>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia  (prefix i \<sigma>) \<Turnstile> \<not>\<^sub>if)))
    )
  ) " 
       by auto
 hence 4: "( ((prefix i \<sigma>) \<Turnstile> \<rhd>f) \<and> i\<le>intlen \<sigma>) =
  (  i\<le>intlen \<sigma> \<and> (
   ( ((prefix i \<sigma>) \<Turnstile> f) \<and> i = 0) \<or>
   ( i>0 \<and> ((prefix i \<sigma>) \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
    )
  ) " 
      using interval_pref_pref_3 using less_imp_add_positive by fastforce
 from 4 show ?thesis by auto
qed

lemma PrefixFstAndsem:
 "( (prefix i \<sigma> \<Turnstile> \<rhd>f \<and>\<^sub>i g) \<and> i\<le>intlen \<sigma>) =
  ( i\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f \<and>\<^sub>i g ) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f \<and>\<^sub>i g) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 ) "
using PrefixFstsem by (metis and_defs)

lemma DiLenFstsem:
 "(\<sigma> \<Turnstile> di (\<rhd>f \<and>\<^sub>i len(i))) =
   ( i\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 ) "
using DiLensem PrefixFstsem by blast

lemma DiLenFstAndsem:
 "(\<sigma> \<Turnstile> di (\<rhd>f \<and>\<^sub>i g \<and>\<^sub>i len(i))) =
   ( i\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f \<and>\<^sub>i g) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f \<and>\<^sub>i g) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 ) "
proof -
 have 1: "(\<sigma> \<Turnstile> di ((\<rhd>f \<and>\<^sub>i g) \<and>\<^sub>i len(i))) =
          ( (prefix i \<sigma> \<Turnstile> (\<rhd>f \<and>\<^sub>i g)) \<and> i\<le>intlen \<sigma>) " 
      using DiLensem by blast 
 have 2: "((prefix i \<sigma> \<Turnstile> (\<rhd>f \<and>\<^sub>i g)) \<and> i\<le>intlen \<sigma>)=
          ( i\<le>intlen \<sigma> \<and>
           ( 
            ( (prefix i \<sigma> \<Turnstile> f \<and>\<^sub>i g ) \<and> i = 0) \<or>
            ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f\<and>\<^sub>i g) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
           ) 
          ) " 
      using PrefixFstAndsem by blast
 from 1 2 show ?thesis by auto
qed   

lemma FstLenSamesem:
 "( ( i\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix i \<sigma> \<Turnstile> f) \<and> i = 0) \<or>
    ( i>0 \<and> (prefix i \<sigma> \<Turnstile> f) \<and> (\<forall>ia<i. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 ) \<and>
  ( j\<le>intlen \<sigma> \<and>
   ( 
    ( (prefix j \<sigma> \<Turnstile> f) \<and> j = 0) \<or>
    ( j>0 \<and> (prefix j \<sigma> \<Turnstile> f) \<and> (\<forall>ia<j. (prefix ia \<sigma> \<Turnstile> \<not>\<^sub>if)))
   ) 
 )
) \<longrightarrow> (i=j)
"
using linorder_neqE_nat by (meson not_defs)


subsection \<open>Theorems\<close>

subsubsection \<open>Fixed length intervals\<close>

lemma LenZeroEqvEmpty:
 "\<turnstile> len(0) \<equiv>\<^sub>i empty"
by simp

lemma LenOneEqvSkip:
 "\<turnstile> len(1) \<equiv>\<^sub>i skip "
by (metis ChopEmpty One_nat_def len\<^sub>_d.simps(1) len\<^sub>_d.simps(2))

lemma LenNPlusOneA:
 "\<turnstile>  len(n+1) \<equiv>\<^sub>i skip;len(n)"
by simp

lemma LenEqvLenChopLen:
 "\<turnstile> len(i+j) \<equiv>\<^sub>i len(i);len(j)"
proof 
 (induct i)
 case 0
 then show ?case using add.left_neutral by auto
 next
 case (Suc i)
 then show ?case 
 by (metis ChopAssoc NextEqvNext add_Suc len\<^sub>_d.simps(2) next_d_def prop03)
qed 

lemma LenNPlusOneB:
 "\<turnstile> len(n+1) \<equiv>\<^sub>i len(n);skip"
proof -
 have  1: "\<turnstile> len(n+1) \<equiv>\<^sub>i len(n);len(1)" by (rule LenEqvLenChopLen)
 have  2: "\<turnstile> len(1) \<equiv>\<^sub>i skip" by (rule LenOneEqvSkip)
 hence 3: "\<turnstile> len(n);len(1) \<equiv>\<^sub>i len(n);skip" using RightChopEqvChop by blast
 from 1 3 show ?thesis using prop03 by blast
qed

lemma RLenEqvLen:
 "\<turnstile> (len k)\<^sup>r \<equiv>\<^sub>i (len k) "
apply simp_all using TimeReverseSem 
   by (metis interval_intrev_intlen interval_rev_rev_ident len_defs)

lemma ExistsLen:
 "(\<forall>\<sigma>. \<exists>k. (\<sigma> \<Turnstile> len(k)))"
using len_defs by simp

lemma AndExistsLen:
 "(\<forall>\<sigma>. (\<sigma> \<Turnstile> f) = ( (\<sigma>\<Turnstile>f) \<and> (\<exists>k. (\<sigma> \<Turnstile> len(k)))))"
using ExistsLen by simp

lemma AndExistsLenChop:
 "(\<forall>\<sigma>. (\<sigma> \<Turnstile> f;g) = (\<exists>k.  (\<sigma>\<Turnstile> (f \<and>\<^sub>i len(k));g ))) "
using  AndExistsLen by auto

lemma AndExistsLenChopR:
 "(\<forall>\<sigma>. (\<sigma> \<Turnstile> f;g) = (\<exists>k.  (\<sigma>\<Turnstile> f;(g \<and>\<^sub>i len(k))))) "
using  AndExistsLen by auto

lemma LFixedAndDistr:
 "\<turnstile> (f0 \<and>\<^sub>i len(k));g0 \<and>\<^sub>i (f1 \<and>\<^sub>i len(k));g1 \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g1) "
apply simp_all
by (metis interval_prefix_length_good)

lemma RFixedAndDistr:
 "\<turnstile> f0;(g0 \<and>\<^sub>i len(k)) \<and>\<^sub>i f1;(g1 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);(g0 \<and>\<^sub>i g1 \<and>\<^sub>i len(k))"
apply simp_all
apply (simp add: interval_prefix_length interval_suffix_length)
by (metis diff_diff_cancel)

lemma LFixedAndDistrA:
 "\<turnstile> (f0 \<and>\<^sub>i len(k));g0 \<and>\<^sub>i (f1 \<and>\<^sub>i len(k));g0 \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1 \<and>\<^sub>i len(k));g0 "
proof -
 have 1: "\<turnstile> (f0 \<and>\<^sub>i len(k));g0 \<and>\<^sub>i (f1 \<and>\<^sub>i len(k));g0 \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g0) " 
      by (rule LFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and>\<^sub>i f1 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g0) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1 \<and>\<^sub>i len(k));g0" 
      by auto
 from 1 2 show ?thesis using prop03 by blast
qed

lemma LFixedAndDistrB:
 "\<turnstile> (f0 \<and>\<^sub>i len(k));g0 \<and>\<^sub>i (f0 \<and>\<^sub>i len(k));g1 \<equiv>\<^sub>i (f0 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g1)"
proof -
 have 1: "\<turnstile> (f0 \<and>\<^sub>i len(k));g0 \<and>\<^sub>i (f0 \<and>\<^sub>i len(k));g1 \<equiv>\<^sub>i (f0 \<and>\<^sub>i f0 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g1) " 
      by (rule LFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and>\<^sub>i f0 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g1) \<equiv>\<^sub>i (f0 \<and>\<^sub>i len(k));(g0 \<and>\<^sub>i g1)" 
      by auto
 from 1 2 show ?thesis using prop03 by blast
qed

lemma RFixedAndDistrA:
 "\<turnstile> f0;(g0 \<and>\<^sub>i len(k)) \<and>\<^sub>i f0;(g1 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i f0;(g0 \<and>\<^sub>i g1 \<and>\<^sub>i len(k)) "
proof -
 have 1: "\<turnstile> f0;(g0 \<and>\<^sub>i len(k)) \<and>\<^sub>i f0;(g1 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f0);(g0 \<and>\<^sub>i g1 \<and>\<^sub>i len(k)) " 
      by (rule RFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and>\<^sub>i f0);(g0 \<and>\<^sub>i g1 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i f0;(g0 \<and>\<^sub>i g1 \<and>\<^sub>i len(k)) " 
      by auto
from 1 2 show ?thesis using prop03 by blast
qed

lemma RFixedAndDistrB:
 "\<turnstile> f0;(g0 \<and>\<^sub>i len(k)) \<and>\<^sub>i f1;(g0 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);(g0 \<and>\<^sub>i len(k)) "
proof -
 have 1: "\<turnstile> f0;(g0 \<and>\<^sub>i len(k)) \<and>\<^sub>i f1;(g0 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);(g0 \<and>\<^sub>i g0 \<and>\<^sub>i len(k)) " 
      by (rule RFixedAndDistr)
 have 2: "\<turnstile> (f0 \<and>\<^sub>i f1);(g0 \<and>\<^sub>i g0 \<and>\<^sub>i len(k)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);(g0 \<and>\<^sub>i len(k)) " 
      by auto
from 1 2 show ?thesis using prop03 by blast
qed 

lemma ChopSkipAndChopSkip:
 "\<turnstile> f0;skip \<and>\<^sub>i f1;skip \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);skip"
proof -
 have 1: "\<turnstile> f0;(true\<^sub>i \<and>\<^sub>i len(1)) \<and>\<^sub>i f1;(true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);(true\<^sub>i \<and>\<^sub>i len(1)) " 
      by (rule RFixedAndDistrB)
 have 2: "\<turnstile> (true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i skip " 
      using LenOneEqvSkip itl_prop(17) prop03 by blast
 hence 3: "\<turnstile> f0;(true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i f0;skip " 
      using RightChopEqvChop by blast
 have 4: "\<turnstile> f1;(true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i f1;skip " 
      using "2" RightChopEqvChop by blast
 have 5: "\<turnstile> f0;(true\<^sub>i \<and>\<^sub>i len(1)) \<and>\<^sub>i f1;(true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i f0;skip \<and>\<^sub>i f1;skip " 
      using "3" "4" by auto
 have 6: "\<turnstile> (f0 \<and>\<^sub>i f1);(true\<^sub>i \<and>\<^sub>i len(1)) \<equiv>\<^sub>i (f0 \<and>\<^sub>i f1);skip" 
      using "2" RightChopEqvChop by blast
 from 1 5 6 show ?thesis by auto
qed

lemma BiAndChopSkipEqv:
 "\<turnstile> (bi (f \<and>\<^sub>i g));skip \<equiv>\<^sub>i (bi f);skip \<and>\<^sub>i (bi g);skip"
proof -
 have 1: "\<turnstile> bi (f \<and>\<^sub>i g) \<equiv>\<^sub>i (bi f) \<and>\<^sub>i (bi g) " 
      by auto
 hence 2: "\<turnstile> (bi (f \<and>\<^sub>i g));skip \<equiv>\<^sub>i (bi f \<and>\<^sub>i bi g);skip" 
      by (rule LeftChopEqvChop)
 have 3: "\<turnstile> (bi f \<and>\<^sub>i bi g);skip \<equiv>\<^sub>i (bi f);skip \<and>\<^sub>i (bi g);skip" 
      using ChopSkipAndChopSkip itl_prop(30) by blast
 from 2 3 show ?thesis by auto
qed

lemma DiAndChopSkipEqv:
 "\<turnstile> (di (f \<and>\<^sub>i g));skip \<supset>\<^sub>i (di f);skip \<and>\<^sub>i (di g);skip"
proof -
 have 1: "\<turnstile> di (f \<and>\<^sub>i g) \<supset>\<^sub>i (di f) \<and>\<^sub>i (di g) " 
      by auto
 hence 2: "\<turnstile> (di (f \<and>\<^sub>i g));skip \<supset>\<^sub>i (di f \<and>\<^sub>i di g);skip" 
      by (rule LeftChopImpChop)
 have 3: "\<turnstile> (di f \<and>\<^sub>i di g);skip \<equiv>\<^sub>i (di f);skip \<and>\<^sub>i (di g);skip" 
      using ChopSkipAndChopSkip itl_prop(30) by blast
 from 2 3 show ?thesis by auto
qed

lemma NotNotChopSkip:
 "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f ;skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (f;skip) "
by (metis WprevEqvEmptyOrPrev prev_d_def wprev_d_def)

lemma NotChopFixed:
 "\<turnstile> \<not>\<^sub>i(f;(g \<and>\<^sub>i len(k))) \<equiv>\<^sub>i \<not>\<^sub>i(\<diamond>(g \<and>\<^sub>i len(k))) \<or>\<^sub>i (\<not>\<^sub>if;(g \<and>\<^sub>i len(k)))"
apply simp by (smt diff_diff_cancel interval_suffix_length_good)


lemma NotFixedChop:
 "\<turnstile> \<not>\<^sub>i((g \<and>\<^sub>i len(k));f) \<equiv>\<^sub>i \<not>\<^sub>i(di(g \<and>\<^sub>i len(k))) \<or>\<^sub>i ((g \<and>\<^sub>i len(k));\<not>\<^sub>if)"
apply simp by auto

subsubsection \<open>Strict initial intervals\<close>

lemma DsMoreDi:
 "\<turnstile> ds f \<equiv>\<^sub>i more \<and>\<^sub>i (di f);skip"
proof -
 have 1: "\<turnstile> ds f \<equiv>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f) " 
      by (simp add: ds_d_def)
 have 2: "\<turnstile> \<not>\<^sub>i(bs \<not>\<^sub>i f) \<equiv>\<^sub>i \<not>\<^sub>i(empty \<or>\<^sub>i (bi \<not>\<^sub>i f);skip) " 
      by (simp add: bs_d_def)
 have 3: "\<turnstile>\<not>\<^sub>i(empty \<or>\<^sub>i (bi \<not>\<^sub>i f);skip) \<equiv>\<^sub>i \<not>\<^sub>iempty \<and>\<^sub>i \<not>\<^sub>i((bi \<not>\<^sub>if);skip) " 
      by auto
 have 4: "\<turnstile> \<not>\<^sub>iempty \<and>\<^sub>i \<not>\<^sub>i((bi \<not>\<^sub>if);skip) \<equiv>\<^sub>i more \<and>\<^sub>i \<not>\<^sub>i((bi \<not>\<^sub>if);skip) " 
      by auto
 have 5: "\<turnstile> more \<and>\<^sub>i \<not>\<^sub>i((bi \<not>\<^sub>if);skip) \<equiv>\<^sub>i more \<and>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(di f);skip) " 
      by auto
 have 6: "\<turnstile> more \<and>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(di f);skip) \<equiv>\<^sub>i more \<and>\<^sub>i (empty \<or>\<^sub>i (di f);skip)" 
      using NotNotChopSkip using prop05 by blast
 have 7: "\<turnstile> more \<and>\<^sub>i (empty \<or>\<^sub>i (di f);skip) \<equiv>\<^sub>i more \<and>\<^sub>i   (di f);skip" 
      by auto
 from 1 2 3 4 5 6 7 show ?thesis by auto
qed

lemma DsDi:
 "\<turnstile> ds f \<equiv>\<^sub>i (di f);skip"
proof -
 have 1: "\<turnstile> ds f \<equiv>\<^sub>i more \<and>\<^sub>i (di f);skip" by (rule DsMoreDi)
 have 2: "\<turnstile> (di f);skip \<supset>\<^sub>i more " by auto
 hence 3: "\<turnstile> more \<and>\<^sub>i (di f);skip \<equiv>\<^sub>i (di f);skip" by auto
 from 1 2 show ?thesis by auto
qed

lemma BsEqvNotDsNot:
 "\<turnstile> bs f \<equiv>\<^sub>i \<not>\<^sub>i(ds \<not>\<^sub>i f)"
proof -
 have 1: "\<turnstile> ds \<not>\<^sub>i f  \<equiv>\<^sub>i more \<and>\<^sub>i (di \<not>\<^sub>i f);skip" by (rule DsMoreDi)
 hence 2: "\<turnstile> \<not>\<^sub>i(ds \<not>\<^sub>i f)  \<equiv>\<^sub>i \<not>\<^sub>i(more \<and>\<^sub>i (di \<not>\<^sub>i f);skip)" by auto
 have 3: "\<turnstile> \<not>\<^sub>i(more \<and>\<^sub>i (di \<not>\<^sub>i f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i((di \<not>\<^sub>i f);skip)" by auto
 have 4: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i((di \<not>\<^sub>i f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip)" by auto
 have 5: "\<turnstile>  \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f);skip " by (rule NotNotChopSkip)
 hence 6: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f);skip " by auto
 from 2 3 4 6 show ?thesis by (simp add: bs_d_def)
qed

lemma NotBsEqvDsNot:
 "\<turnstile> \<not>\<^sub>i(bs f) \<equiv>\<^sub>i ds \<not>\<^sub>i f"
proof -
 have 1: "\<turnstile> bs f \<equiv>\<^sub>i \<not>\<^sub>i(ds \<not>\<^sub>i f)" by (rule BsEqvNotDsNot)
 hence 2: "\<turnstile> \<not>\<^sub>i( bs f) \<equiv>\<^sub>i \<not>\<^sub>i\<not>\<^sub>i(ds \<not>\<^sub>i f) " by auto
 from 2 show ?thesis by auto
qed

lemma NotDsEqvBsNot:
 "\<turnstile> \<not>\<^sub>i (ds f) \<equiv>\<^sub>i bs \<not>\<^sub>i f"
proof -
 have 1: "\<turnstile> \<not>\<^sub>i(ds  f) \<equiv>\<^sub>i \<not>\<^sub>i\<not>\<^sub>i(bs \<not>\<^sub>if)" by (simp add: ds_d_def)
 from 1 show ?thesis by auto
qed

lemma NotDsAndEmpty:
 "\<turnstile> \<not>\<^sub>i(ds f \<and>\<^sub>i empty)"
proof -
 have 1: "\<turnstile> ds f \<equiv>\<^sub>i more \<and>\<^sub>i (di f);skip" by (rule DsMoreDi)
 have 2: "\<turnstile> more \<and>\<^sub>i (di f);skip \<and>\<^sub>i empty \<supset>\<^sub>i false\<^sub>i" by auto
 from 1 2 show ?thesis by auto
qed

lemma BsMoreEqvEmpty:
 "\<turnstile> bs more \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> bs more \<equiv>\<^sub>i empty \<or>\<^sub>i (bi more);skip" by (simp add: bs_d_def)
 have 2: "\<turnstile> bi more \<supset>\<^sub>i false\<^sub>i" using DiEmpty by auto
 hence 3: "\<turnstile> (bi more);skip \<supset>\<^sub>i false\<^sub>i " by auto
 hence 4: "\<turnstile> empty \<or>\<^sub>i ((bi more);skip) \<equiv>\<^sub>i empty" using prop25 by blast
 from 1 4 show ?thesis by auto
qed

lemma  BsAndEqv:
 "\<turnstile> bs f \<and>\<^sub>i bs g \<equiv>\<^sub>i bs(f \<and>\<^sub>i g)"
proof -
 have 1: "\<turnstile> bs f \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f) ;skip" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> bs g \<equiv>\<^sub>i empty \<or>\<^sub>i (bi g) ;skip" 
      by (simp add: bs_d_def)
 have 3: "\<turnstile> bs f \<and>\<^sub>i bs g \<equiv>\<^sub>i (empty \<or>\<^sub>i (bi f) ;skip) \<and>\<^sub>i (empty \<or>\<^sub>i (bi g) ;skip) " 
      using "1" "2" by auto
 have 4: "\<turnstile> (empty \<or>\<^sub>i (bi f) ;skip) \<and>\<^sub>i (empty \<or>\<^sub>i (bi g) ;skip) \<equiv>\<^sub>i 
             empty \<or>\<^sub>i ((bi f) ;skip \<and>\<^sub>i (bi g) ;skip) " 
      by auto
 have 5: "\<turnstile> ((bi f) ;skip \<and>\<^sub>i (bi g) ;skip) \<equiv>\<^sub>i bi(f \<and>\<^sub>i g);skip " 
      using BiAndChopSkipEqv itl_prop(30) by blast
 hence 6: "\<turnstile> empty \<or>\<^sub>i ((bi f) ;skip \<and>\<^sub>i (bi g) ;skip) \<equiv>\<^sub>i empty \<or>\<^sub>i bi(f \<and>\<^sub>i g);skip " 
      by auto
 from 3 4 6 show ?thesis by (simp add: bs_d_def)
qed

lemma DsEqvRule:
 assumes "\<turnstile> f \<equiv>\<^sub>i g"
 shows "\<turnstile> ds f \<equiv>\<^sub>i ds g"
by (meson DiEqvDi DsDi LeftChopEqvChop assms itl_prop(30) prop03)

lemma DsOrEqv:
 "\<turnstile> ds f \<or>\<^sub>i ds g \<equiv>\<^sub>i ds (f \<or>\<^sub>i g)"
proof -
 have 1: "\<turnstile> ds f \<equiv>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f)" by (simp add: ds_d_def)
 have 2: "\<turnstile> ds g \<equiv>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i g)" by (simp add: ds_d_def)
 have 3: "\<turnstile> ds f \<or>\<^sub>i ds g \<equiv>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f) \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i g)" using "1" "2" by auto
 have 4: "\<turnstile> \<not>\<^sub>i(bs \<not>\<^sub>i f) \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i g) \<equiv>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g)" by auto
 have 5: "\<turnstile> bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i bs( \<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig)" by (rule BsAndEqv)
 hence 6: "\<turnstile> \<not>\<^sub>i(bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<equiv>\<^sub>i \<not>\<^sub>i(bs( \<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig)) " by auto
 have 7: "\<turnstile> \<not>\<^sub>i(bs( \<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig)) \<equiv>\<^sub>i ds (\<not>\<^sub>i(\<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig)) " by (rule NotBsEqvDsNot)
 have 8: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig) \<equiv>\<^sub>i (f \<or>\<^sub>i g)" by auto
 hence 9: "\<turnstile> ds(\<not>\<^sub>i(\<not>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>ig)) \<equiv>\<^sub>i ds (f \<or>\<^sub>i g) " by (rule DsEqvRule)
 from 3 4 6 7 9 show ?thesis by auto
qed

lemma BiOrBiImpBiOr:
 "\<turnstile> bi f \<or>\<^sub>i bi g \<supset>\<^sub>i bi(f \<or>\<^sub>i g) "
proof -
 have 1: "\<turnstile> f \<supset>\<^sub>i f \<or>\<^sub>i g" by auto
 hence 2: "\<turnstile> bi f \<supset>\<^sub>i bi(f \<or>\<^sub>i g)" by (rule BiImpBiRule)
 have 3: "\<turnstile> g \<supset>\<^sub>i f \<or>\<^sub>i g" by auto
 hence 4: "\<turnstile> bi g \<supset>\<^sub>i bi(f \<or>\<^sub>i g)" by (rule BiImpBiRule)
 from 2 4 show ?thesis by auto
qed

lemma BsOrImp:
 "\<turnstile> bs f \<or>\<^sub>i bs g \<supset>\<^sub>i bs(f \<or>\<^sub>i g)"
proof -
 have 1: "\<turnstile> bi f \<or>\<^sub>i bi g \<supset>\<^sub>i bi(f \<or>\<^sub>i g)" 
      by (rule BiOrBiImpBiOr)
 hence 2: "\<turnstile> (bi f \<or>\<^sub>i bi g);skip \<supset>\<^sub>i (bi(f \<or>\<^sub>i g));skip" 
      by (rule LeftChopImpChop)
 have 3: "\<turnstile> (bi f);skip \<or>\<^sub>i (bi g);skip \<supset>\<^sub>i (bi(f \<or>\<^sub>i g));skip" 
      using "1" OrChopEqv "2" itl_prop(31) prop02 by blast
 hence 4: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i (bi g);skip \<supset>\<^sub>i  empty \<or>\<^sub>i (bi(f \<or>\<^sub>i g));skip" 
      by auto
 hence 5: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i empty \<or>\<^sub>i (bi g);skip \<supset>\<^sub>i  empty \<or>\<^sub>i (bi(f \<or>\<^sub>i g));skip " 
      by auto
 from 5 show ?thesis by (simp add: bs_d_def)
qed

lemma DsAndImp:
 "\<turnstile> ds (f \<and>\<^sub>i g) \<supset>\<^sub>i ds f \<and>\<^sub>i ds g "
proof -
 have 1: "\<turnstile> bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig \<supset>\<^sub>i bs(\<not>\<^sub>if \<or>\<^sub>i \<not>\<^sub>ig)" by (rule BsOrImp)
 have 2: "\<turnstile> \<not>\<^sub>if \<or>\<^sub>i \<not>\<^sub>ig \<equiv>\<^sub>i \<not>\<^sub>i(f \<and>\<^sub>i g)" by auto
 hence 3: "\<turnstile> bs(\<not>\<^sub>if \<or>\<^sub>i \<not>\<^sub>ig) \<equiv>\<^sub>i bs \<not>\<^sub>i(f \<and>\<^sub>i g)" by (rule BsEqvRule)
 have 4: "\<turnstile> bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig \<supset>\<^sub>i bs \<not>\<^sub>i(f \<and>\<^sub>i g)" using "1" "3" by auto
 have 5: "\<turnstile> bs \<not>\<^sub>if \<equiv>\<^sub>i \<not>\<^sub>i(ds f)" using NotDsEqvBsNot by auto
 have 6: "\<turnstile> bs \<not>\<^sub>ig \<equiv>\<^sub>i \<not>\<^sub>i(ds g)" using NotDsEqvBsNot by auto
 have 7: "\<turnstile> bs \<not>\<^sub>i(f \<and>\<^sub>i g) \<equiv>\<^sub>i  \<not>\<^sub>i(ds (f \<and>\<^sub>i g))" using NotDsEqvBsNot by auto
 have 8: "\<turnstile>  \<not>\<^sub>i(ds f) \<or>\<^sub>i  \<not>\<^sub>i(ds g) \<supset>\<^sub>i  \<not>\<^sub>i(ds (f \<and>\<^sub>i g))" using "4" "5" "6" "7" by auto
 hence 9: "\<turnstile> \<not>\<^sub>i(ds f \<and>\<^sub>i ds g)  \<supset>\<^sub>i  \<not>\<^sub>i(ds (f \<and>\<^sub>i g))" by auto
 from 9 show ?thesis by auto
qed

lemma DsAndImpElimL:
 "\<turnstile> ds (f \<and>\<^sub>i g) \<supset>\<^sub>i ds f "
using DsAndImp by auto

lemma DsAndImpElimR:
 "\<turnstile> ds (f \<and>\<^sub>i g) \<supset>\<^sub>i ds g "
using DsAndImp by auto

lemma MoreAndBiImpBiChopSkip:
 "\<turnstile> more \<and>\<^sub>i bi f \<supset>\<^sub>i (bi f);skip "
proof -
 have 1: "\<turnstile>  (bi f);skip \<equiv>\<^sub>i \<not>\<^sub>i(di \<not>\<^sub>i f);skip " by auto
 have 2: "\<turnstile> \<not>\<^sub>i( \<not>\<^sub>i(di \<not>\<^sub>i f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (di \<not>\<^sub>i f);skip" by (rule NotNotChopSkip)
 have 3: "\<turnstile> empty \<supset>\<^sub>i empty \<or>\<^sub>i di \<not>\<^sub>i f " by auto
 have 4: "\<turnstile> (di \<not>\<^sub>i f);skip \<supset>\<^sub>i di \<not>\<^sub>i f" using ChopImpDi DiEqvDiDi itl_prop(31) prop02 by blast 
 hence 5: "\<turnstile> (di \<not>\<^sub>i f);skip \<supset>\<^sub>i empty \<or>\<^sub>i di \<not>\<^sub>i f" by (rule prop26)
 have 6: "\<turnstile> \<not>\<^sub>i( \<not>\<^sub>i(di \<not>\<^sub>i f);skip) \<supset>\<^sub>i empty \<or>\<^sub>i di \<not>\<^sub>i f" using "2" "3" "5" by auto
 hence 7: "\<turnstile> \<not>\<^sub>i(empty \<or>\<^sub>i di \<not>\<^sub>i f) \<supset>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i( \<not>\<^sub>i(di \<not>\<^sub>i f);skip)) "  by (rule prop27)
 have 8: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i( \<not>\<^sub>i(di \<not>\<^sub>i f);skip)) \<equiv>\<^sub>i \<not>\<^sub>i(di \<not>\<^sub>i f);skip" by auto
 have 9: "\<turnstile> \<not>\<^sub>i(empty \<or>\<^sub>i di \<not>\<^sub>i f) \<equiv>\<^sub>i more \<and>\<^sub>i \<not>\<^sub>i( di \<not>\<^sub>i f) " by auto
 have 10: "\<turnstile> more \<and>\<^sub>i \<not>\<^sub>i( di \<not>\<^sub>i f) \<equiv>\<^sub>i more \<and>\<^sub>i bi f" by auto
 from 1 6 7 8 9 10 show ?thesis by auto
qed

lemma BiImpBs:
 "\<turnstile> bi f \<supset>\<^sub>i bs f  "
proof -
 have 1: "\<turnstile> empty \<supset>\<^sub>i empty \<or>\<^sub>i (bi f);skip" by auto
 hence 2: "\<turnstile> empty \<and>\<^sub>i bi f \<supset>\<^sub>i empty \<or>\<^sub>i (bi f);skip" by auto
 have 2: "\<turnstile> more \<and>\<^sub>i bi f \<supset>\<^sub>i (bi f);skip" by (rule MoreAndBiImpBiChopSkip)
 hence 3: "\<turnstile> more \<and>\<^sub>i bi f \<supset>\<^sub>i empty \<or>\<^sub>i (bi f);skip " by auto
 have 4: "\<turnstile> bi f \<equiv>\<^sub>i (bi f \<and>\<^sub>i empty) \<or>\<^sub>i (bi f \<and>\<^sub>i more)" by auto
 have 5: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<equiv>\<^sub>i bs f " by (simp add: bs_d_def)
 from 2 3 4 5 show ?thesis by auto
qed

lemma BsImpBsBs:
 "\<turnstile> bs f \<supset>\<^sub>i bs ( bs f)"
proof -
 have 1: "\<turnstile> bi f \<supset>\<^sub>i bs f" by (rule BiImpBs)
 hence 2: "\<turnstile> bi (bi f) \<supset>\<^sub>i bi(bs f)" by (rule BiImpBiRule)
 hence 3: "\<turnstile> (bi f) \<supset>\<^sub>i bi(bs f)" using BiEqvBiBi itl_prop(31) prop02 by blast
 hence 4: "\<turnstile> (bi f);skip \<supset>\<^sub>i (bi(bs f));skip" by (rule LeftChopImpChop)
 hence 5: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<supset>\<^sub>i empty \<or>\<^sub>i (bi(bs f));skip" by auto
 from 5 show ?thesis by (simp add: bs_d_def)
qed

lemma DsImpDi:
 "\<turnstile> ds f \<supset>\<^sub>i di f"
proof -
 have 1: "\<turnstile> bi \<not>\<^sub>i f \<supset>\<^sub>i bs \<not>\<^sub>i f" by (rule BiImpBs)
 hence 2: "\<turnstile> \<not>\<^sub>i(bs \<not>\<^sub>i f) \<supset>\<^sub>i \<not>\<^sub>i(bi \<not>\<^sub>i f) " by (rule prop27)
 from 2 show ?thesis using NotBsEqvDsNot DiEqvNotBiNot by (simp add: ds_d_def)
qed

lemma BsImpBsRule:
 assumes "\<turnstile> f \<supset>\<^sub>i g"
 shows "\<turnstile> bs f  \<supset>\<^sub>i bs g"
proof -
 have 1: "\<turnstile> f \<supset>\<^sub>i g" using assms by auto
 hence 2: "\<turnstile> bi f \<supset>\<^sub>i bi g " by (rule BiImpBiRule)
 hence 3: "\<turnstile> (bi f);skip \<supset>\<^sub>i (bi g);skip " by (rule LeftChopImpChop)
 hence 4: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<supset>\<^sub>i empty \<or>\<^sub>i (bi g);skip" by auto
 from 4 show ?thesis by (simp add: bs_d_def)
qed

lemma DiChopImpDiB:
 "\<turnstile> di(f;g) \<supset>\<^sub>i di f "
proof -
 have 1: "\<turnstile> f ; (g;true\<^sub>i) \<supset>\<^sub>i di f" by (rule ChopImpDi)
 have 2: "\<turnstile> f ; (g;true\<^sub>i) \<equiv>\<^sub>i (f;g);true\<^sub>i " by (rule ChopAssoc)
 from 1 2 show ?thesis by (simp add: di_d_def)
qed
 
lemma DsChopImpDsB:
 "\<turnstile> ds (f;g) \<supset>\<^sub>i ds f"
proof -
 have 1: "\<turnstile> di(f;g) \<supset>\<^sub>i di f" by (rule DiChopImpDiB)
 hence 2: "\<turnstile> (di(f;g));skip \<supset>\<^sub>i (di f);skip " by (rule LeftChopImpChop)
 from 2 show ?thesis using DsDi by (metis itl_prop(31) prop02)
qed

lemma NotBsImpBsNotChop:
 "\<turnstile> bs \<not>\<^sub>i f \<supset>\<^sub>i bs ( \<not>\<^sub>i(f;g))"
proof -
 have 1: "\<turnstile> ds (f;g) \<supset>\<^sub>i ds f" by (rule DsChopImpDsB)
 hence 2: "\<turnstile> \<not>\<^sub>i(ds f) \<supset>\<^sub>i \<not>\<^sub>i(ds (f;g))" by (rule prop27)
 from 2 show ?thesis using NotDsEqvBsNot by auto
qed

lemma BiBiOrImpBi:
 "\<turnstile> bi ( bi f \<or>\<^sub>i bi g) \<supset>\<^sub>i bi f \<or>\<^sub>i bi g"
using BiElim by auto

lemma BiImpBiBiOr:
 "\<turnstile> bi f  \<supset>\<^sub>i bi ( bi f \<or>\<^sub>i bi g)"
proof -
 have 1: "\<turnstile> bi f \<supset>\<^sub>i bi f \<or>\<^sub>i bi g" by auto
 hence 2: "\<turnstile> bi (bi f) \<supset>\<^sub>i bi(bi f \<or>\<^sub>i bi g)  " using BiImpBiRule by blast
 have 3: "\<turnstile> bi (bi f) \<equiv>\<^sub>i bi f" using BiEqvBiBi itl_prop(30) by blast
 from 2 3 show ?thesis by auto
qed

lemma BiImpBiBiOrB:
 "\<turnstile> bi g  \<supset>\<^sub>i bi ( bi f \<or>\<^sub>i bi g)"
proof -
 have 1: "\<turnstile> bi g \<supset>\<^sub>i bi f \<or>\<^sub>i bi g" by auto
 hence 2: "\<turnstile> bi (bi g) \<supset>\<^sub>i bi(bi f \<or>\<^sub>i bi g)  " using BiImpBiRule by blast
 have 3: "\<turnstile> bi (bi g) \<equiv>\<^sub>i bi g" using BiEqvBiBi itl_prop(30) by blast
 from 2 3 show ?thesis by auto
qed

lemma BiBiOrEqvBi:
 "\<turnstile> bi ( bi f \<or>\<^sub>i bi g) \<equiv>\<^sub>i bi f \<or>\<^sub>i bi g "
proof -
 have 1: "\<turnstile> bi ( bi f \<or>\<^sub>i bi g) \<supset>\<^sub>i bi f \<or>\<^sub>i bi g " by (rule BiBiOrImpBi)
 have 2: "\<turnstile> bi f  \<supset>\<^sub>i bi ( bi f \<or>\<^sub>i bi g) " by (rule BiImpBiBiOr)
 have 3: "\<turnstile> bi g  \<supset>\<^sub>i bi ( bi f \<or>\<^sub>i bi g) " by (rule BiImpBiBiOrB)
 have 4: "\<turnstile> bi f  \<or>\<^sub>i bi g \<supset>\<^sub>i bi ( bi f \<or>\<^sub>i bi g) " using "2" "3" by auto
 from 1 4 show ?thesis using itl_prop(31) by blast
qed

(* proof has changed notice the crucial lemma BiBiOrEqvBi that is needed!*)
lemma BsOrBsEqvBsBiOrBi:
 "\<turnstile> bs f \<or>\<^sub>i bs g \<equiv>\<^sub>i bs(bi f \<or>\<^sub>i bi g) "
proof -
 have 1: "\<turnstile> bs f \<or>\<^sub>i bs g \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i empty \<or>\<^sub>i (bi g);skip" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i empty \<or>\<^sub>i (bi g);skip \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i (bi g);skip" 
      by auto
 have 3: "\<turnstile> (bi f);skip \<or>\<^sub>i (bi g);skip \<equiv>\<^sub>i (bi f \<or>\<^sub>i bi g);skip" 
      using OrChopEqv using itl_prop(30) by blast
 hence 4: "\<turnstile> empty \<or>\<^sub>i (bi f);skip \<or>\<^sub>i (bi g);skip \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f \<or>\<^sub>i bi g);skip" 
      by auto
 have 5: "\<turnstile> bi ( bi f \<or>\<^sub>i bi g) \<equiv>\<^sub>i bi f \<or>\<^sub>i bi g" 
      by (rule  BiBiOrEqvBi)
 hence 6: "\<turnstile> bi ( bi f \<or>\<^sub>i bi g);skip \<equiv>\<^sub>i (bi f \<or>\<^sub>i bi g);skip" 
      using LeftChopEqvChop by blast 
 hence 7: "\<turnstile> empty \<or>\<^sub>i bi ( bi f \<or>\<^sub>i bi g);skip \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f \<or>\<^sub>i bi g);skip" 
      by auto
 from 1 2 4 7 show ?thesis by (simp add: bs_d_def)
qed

lemma DiOrDsEqvDi:
 "\<turnstile> di f \<or>\<^sub>i ds f \<equiv>\<^sub>i di f"
proof -
 have 1: "\<turnstile> di f \<supset>\<^sub>i di f \<or>\<^sub>i ds f" by auto
 have 2: "\<turnstile> di f \<supset>\<^sub>i di f" by auto
 have 3: "\<turnstile> ds f \<supset>\<^sub>i di f" by (rule DsImpDi)
 have 4: "\<turnstile> di f \<or>\<^sub>i ds f \<supset>\<^sub>i di f" using "2" "3" by auto
 from 1 4 show ?thesis by auto
qed

lemma DiAndDsEqvDs:
 "\<turnstile> di f \<and>\<^sub>i ds f \<equiv>\<^sub>i ds f "
proof -
 have 1: "\<turnstile> di f \<and>\<^sub>i ds f \<supset>\<^sub>i ds f" by auto
 have 2: "\<turnstile> ds f \<supset>\<^sub>i ds f " by auto
 have 3: "\<turnstile> ds f \<supset>\<^sub>i di f" by (rule DsImpDi)
 have 4: "\<turnstile> ds f \<supset>\<^sub>i di f \<and>\<^sub>i ds f" using "2" "3" by auto
 from 1 4 show ?thesis by auto
qed

lemma DiEqvOrDiChopSkipA:
 "\<turnstile> di f \<equiv>\<^sub>i f \<or>\<^sub>i di(f;skip)"
proof -
 have 1: "\<turnstile> di f \<equiv>\<^sub>i f ;true\<^sub>i " by (simp add: di_d_def)
 hence 2: "\<turnstile> di f \<equiv>\<^sub>i f; ( empty \<or>\<^sub>i more)" by auto
 hence 3: "\<turnstile>f; ( empty \<or>\<^sub>i more) \<equiv>\<^sub>i f;empty \<or>\<^sub>i f;more " using ChopOrEqv by blast
 have 4: "\<turnstile> f;empty \<equiv>\<^sub>i f" by (rule ChopEmpty)
 have 5: "\<turnstile> more \<equiv>\<^sub>i skip;true\<^sub>i" using MoreEqvSkipChopTrue by blast
 hence 6: "\<turnstile> f;more \<equiv>\<^sub>i f;(skip;true\<^sub>i)" using RightChopEqvChop by blast
 have 7: "\<turnstile> f;(skip;true\<^sub>i) \<equiv>\<^sub>i (f;skip);true\<^sub>i" by (rule ChopAssoc)
 from 2 3 4 6 7 show ?thesis by auto
qed   

lemma SkipTrueEqvTrueSkip: 
 "\<turnstile> skip;true\<^sub>i \<equiv>\<^sub>i true\<^sub>i;skip "
using TrueChopSkipEqvSkipChopTrue itl_prop(30) by blast

lemma DiEqvOrDiChopSkipB:
 "\<turnstile> di f \<equiv>\<^sub>i f \<or>\<^sub>i (di f);skip"
proof -
 have 1: "\<turnstile> (di f) \<equiv>\<^sub>i f \<or>\<^sub>i di(f;skip)" by (rule DiEqvOrDiChopSkipA)
 have 2: "\<turnstile>  di(f;skip) \<equiv>\<^sub>i (f;skip);true\<^sub>i " by (simp add: di_d_def)
 have 3: "\<turnstile> (f;skip);true\<^sub>i \<equiv>\<^sub>i f;(skip;true\<^sub>i)" by (rule ChopAssocB)
 have 4: "\<turnstile> di(f;skip) \<equiv>\<^sub>i f;(skip;true\<^sub>i) " using "2" "3" by auto
 have 5: "\<turnstile> skip;true\<^sub>i \<equiv>\<^sub>i true\<^sub>i;skip" by (rule SkipTrueEqvTrueSkip)
 hence 6: "\<turnstile> f;(skip;true\<^sub>i) \<equiv>\<^sub>i f;(true\<^sub>i;skip)" using RightChopEqvChop by blast
 have 7: "\<turnstile> di(f;skip) \<equiv>\<^sub>i f;(true\<^sub>i;skip)" using "4" "6" by auto
 have 8: "\<turnstile> f;(true\<^sub>i;skip) \<equiv>\<^sub>i (f;true\<^sub>i);skip" by (rule ChopAssoc)
 have 9: "\<turnstile> (f;true\<^sub>i);skip \<equiv>\<^sub>i (di f);skip" by (simp add: di_d_def)
 have 10 : "\<turnstile> di(f;skip) \<equiv>\<^sub>i(di f);skip " using "7" "8" "9" by auto
 hence 11: "\<turnstile> f \<or>\<^sub>i di(f;skip) \<equiv>\<^sub>i f \<or>\<^sub>i (di f);skip " by auto
 from 1 11 show ?thesis using prop03 by blast
qed
  
lemma OrDsEqvDi:
 "\<turnstile> f \<or>\<^sub>i ds f \<equiv>\<^sub>i di f"
proof -
 have 1: "\<turnstile> ds f \<equiv>\<^sub>i  (di f);skip" by (rule DsDi)
 hence 2: "\<turnstile> f \<or>\<^sub>i ds f \<equiv>\<^sub>i f \<or>\<^sub>i (di f);skip" by auto
 from 2 show ?thesis using DiEqvOrDiChopSkipB itl_prop(30) prop03 by blast
qed

lemma BiEqvAndEmptyOrBiChopSkip:
 "\<turnstile> bi f \<equiv>\<^sub>i f \<and>\<^sub>i (empty \<or>\<^sub>i (bi f);skip) "
proof -
 have 1: "\<turnstile> di \<not>\<^sub>i f \<equiv>\<^sub>i \<not>\<^sub>i f \<or>\<^sub>i (di \<not>\<^sub>i f;skip)" by (rule DiEqvOrDiChopSkipB)
 have 2: "\<turnstile> di \<not>\<^sub>i f \<equiv>\<^sub>i \<not>\<^sub>i( bi f )" by (rule DiNotEqvNotBi)
 have 3: "\<turnstile> \<not>\<^sub>i( bi f ) \<equiv>\<^sub>i  \<not>\<^sub>i f \<or>\<^sub>i (di \<not>\<^sub>i f;skip) " using "1" "2" using itl_prop(30) prop03 by blast
 hence 4: "\<turnstile> bi f \<equiv>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i f \<or>\<^sub>i (di \<not>\<^sub>i f;skip)) " by (metis "1" bi_d_def prop01)
 have 5: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f \<or>\<^sub>i (di \<not>\<^sub>i f;skip)) \<equiv>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>i(di \<not>\<^sub>i f;skip) " by auto
 have 6: "\<turnstile> di \<not>\<^sub>i f;skip \<equiv>\<^sub>i \<not>\<^sub>i(bi f);skip " by auto
 hence 7: "\<turnstile> \<not>\<^sub>i(di \<not>\<^sub>i f;skip) \<equiv>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) " by auto
 have 8: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) \<equiv>\<^sub>i (empty \<or>\<^sub>i (bi f);skip) " using NotNotChopSkip by blast
 hence 9: "\<turnstile> f \<and>\<^sub>i \<not>\<^sub>i(di \<not>\<^sub>i f;skip) \<equiv>\<^sub>i f \<and>\<^sub>i (empty \<or>\<^sub>i (bi f);skip) " using "7" "8" by auto
 from 4 5 9 show ?thesis using prop03 by blast
qed

lemma AndBsEqvBi:
 "\<turnstile> f \<and>\<^sub>i bs f \<equiv>\<^sub>i bi f "
proof -
 have 1: "\<turnstile> f \<and>\<^sub>i bs f \<equiv>\<^sub>i f \<and>\<^sub>i (empty \<or>\<^sub>i (bi f);skip) " by (simp add: bs_d_def)
 from 1 show ?thesis using BiEqvAndEmptyOrBiChopSkip by (metis bs_d_def itl_prop(30))
qed

lemma BsEqvBsBi:
 "\<turnstile> bs f \<equiv>\<^sub>i bs (bi f) "
proof -
 have 1: "\<turnstile> bs f \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f);skip" by (simp add: bs_d_def)
 have 2: "\<turnstile> bi f \<equiv>\<^sub>i bi ( bi f) " by (rule BiEqvBiBi)
 hence 3: "\<turnstile> (bi f);skip \<equiv>\<^sub>i bi (bi f);skip " using LeftChopEqvChop by blast
 hence 4: "\<turnstile>  empty \<or>\<^sub>i (bi f);skip \<equiv>\<^sub>i  empty \<or>\<^sub>i bi (bi f);skip " by auto
 from 1 4 show ?thesis by (simp add: bs_d_def)
qed

lemma StateImpBs:
 "\<turnstile> init w \<supset>\<^sub>i bs (init w)"
proof -
 have 1: "\<turnstile> init w \<equiv>\<^sub>i bi (init w) "  by (rule StateEqvBi)
 have 2: "\<turnstile> bi (init w) \<supset>\<^sub>i bs (init w) " by (rule BiImpBs)
 from 1 2 show ?thesis using StateImpBi prop02 by blast
qed

lemma DiDiAndEqvDi:
 "\<turnstile> di ( di f \<and>\<^sub>i di g) \<equiv>\<^sub>i di f \<and>\<^sub>i di g "
proof -
 have 1: "\<turnstile> bi ( bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g) \<equiv>\<^sub>i bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>i g " 
      by (rule BiBiOrEqvBi)
 have 2: "\<turnstile> bi \<not>\<^sub>if \<equiv>\<^sub>i \<not>\<^sub>i (di f) " 
      by auto
 have 3: "\<turnstile> bi \<not>\<^sub>ig \<equiv>\<^sub>i \<not>\<^sub>i (di g) " 
      by auto
 have 4: "\<turnstile> bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>i g \<equiv>\<^sub>i \<not>\<^sub>i (di f) \<or>\<^sub>i \<not>\<^sub>i (di g) " 
      using "2" "3" by auto 
 have 5: "\<turnstile> \<not>\<^sub>i (di f) \<or>\<^sub>i \<not>\<^sub>i (di g) \<equiv>\<^sub>i \<not>\<^sub>i(di f \<and>\<^sub>i di g) " 
      by auto
 have 6: "\<turnstile> bi ( bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g) \<equiv>\<^sub>i \<not>\<^sub>i(di f \<and>\<^sub>i di g) " 
      using "1" "5" by auto
 hence 7: "\<turnstile> \<not>\<^sub>i(bi ( bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g))\<equiv>\<^sub>i (di f \<and>\<^sub>i di g)  " 
      by simp
 have 8 : "\<turnstile> \<not>\<^sub>i(bi ( bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g)) \<equiv>\<^sub>i di ( \<not>\<^sub>i(bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g))  " 
      using DiNotEqvNotBi itl_prop(30) by blast
 have 9 : "\<turnstile> \<not>\<^sub>i(bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g) \<equiv>\<^sub>i di f \<and>\<^sub>i di g " 
      by auto
 hence 10: "\<turnstile> di ( \<not>\<^sub>i(bi \<not>\<^sub>i f \<or>\<^sub>i bi \<not>\<^sub>i g)) \<equiv>\<^sub>i di ( di f \<and>\<^sub>i di g) " 
      using DiEqvDi by blast
 from 7 8 10 show ?thesis using itl_prop(30) prop03 by blast 
qed

lemma BiInduct:
 "\<turnstile> bi(f \<supset>\<^sub>i wprev f) \<and>\<^sub>i f \<supset>\<^sub>i bi f "
proof -
 have 1: "\<turnstile> \<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)) \<and>\<^sub>i f\<^sup>r \<supset>\<^sub>i \<box>(f\<^sup>r)" using BoxInduct by blast
 hence 2: "\<turnstile> (\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)) \<and>\<^sub>i f\<^sup>r \<supset>\<^sub>i \<box>(f\<^sup>r))\<^sup>r" using ReverseEqv by blast
 have 3: "\<turnstile> ((f\<^sup>r)\<^sup>r ) \<equiv>\<^sub>i f" using EqvReverseReverse  itl_prop(30) by blast
 have 4: "\<turnstile> (\<box> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i bi (f)  " using RRBoxEqvBi by blast
 have 5: "\<turnstile> ((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r))\<^sup>r \<equiv>\<^sub>i ( (f\<^sup>r)\<^sup>r \<supset>\<^sub>i (wnext(f\<^sup>r))\<^sup>r )" by simp
 have 6: "\<turnstile> (wnext(f\<^sup>r))\<^sup>r \<equiv>\<^sub>i wprev(f)" using RRWNextEqvWPrev by blast 
 have 7: "\<turnstile> ( (f\<^sup>r)\<^sup>r \<supset>\<^sub>i (wnext(f\<^sup>r))\<^sup>r ) \<equiv>\<^sub>i ( f \<supset>\<^sub>i wprev(f) )" using "6" "3" prop39 by auto
 have 8: "\<turnstile> bi( (f\<^sup>r)\<^sup>r \<supset>\<^sub>i (wnext(f\<^sup>r))\<^sup>r ) \<equiv>\<^sub>i bi( f \<supset>\<^sub>i wprev(f) )" using "7" "3" BiEqvBi by blast
 have 9: "\<turnstile> (\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)))\<^sup>r \<equiv>\<^sub>i bi ( ((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r))\<^sup>r ) " using RBoxEqvBi by blast
 have 10: "\<turnstile> (\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)))\<^sup>r \<equiv>\<^sub>i bi( f \<supset>\<^sub>i wprev(f) )" using "8" "9" by auto
 have 11: "\<turnstile> (\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)) \<and>\<^sub>i f\<^sup>r \<supset>\<^sub>i \<box>(f\<^sup>r))\<^sup>r \<equiv>\<^sub>i 
             ((\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)))\<^sup>r \<and>\<^sub>i (f\<^sup>r)\<^sup>r \<supset>\<^sub>i (\<box>(f\<^sup>r))\<^sup>r)" using RAnd by auto
 have 12: "\<turnstile> ((\<box>((f\<^sup>r) \<supset>\<^sub>i wnext(f\<^sup>r)))\<^sup>r \<and>\<^sub>i (f\<^sup>r)\<^sup>r \<supset>\<^sub>i (\<box>(f\<^sup>r))\<^sup>r) \<equiv>\<^sub>i
            (bi( f \<supset>\<^sub>i wprev(f) ) \<and>\<^sub>i f \<supset>\<^sub>i bi f ) " using "8" "3" "4" "10" by simp
 from 2 11 12 show ?thesis using MP itl_prop(31) by blast
qed
 
lemma PrevLoop:
 assumes "\<turnstile> f \<supset>\<^sub>i prev f"
 shows   "\<turnstile> \<not>\<^sub>i f"
proof -
 have  1: "\<turnstile> f \<supset>\<^sub>i prev f " using assms by auto
 hence 2: "\<turnstile> f \<supset>\<^sub>i ( more \<and>\<^sub>i wprev f) " by auto
 hence 3: "\<turnstile> f \<supset>\<^sub>i wprev f " by auto
 hence 4: "\<turnstile> bi(f \<supset>\<^sub>i wprev f)" by (rule BiGen)
 have  5: "\<turnstile> bi(f \<supset>\<^sub>i wprev f) \<and>\<^sub>i f \<supset>\<^sub>i bi f " by (rule BiInduct)
 hence 6: "\<turnstile> bi(f \<supset>\<^sub>i wprev f) \<supset>\<^sub>i (f \<supset>\<^sub>i bi f)" using prop36 by blast 
 have  7: "\<turnstile> (f \<supset>\<^sub>i bi f)" using "4" "6" MP by blast
 have  8: "\<turnstile> bi f \<supset>\<^sub>i f" by (rule BiElim)
 have  9: "\<turnstile> f \<equiv>\<^sub>i bi f" using "7" "8"  itl_prop(31) by blast
 have 10: "\<turnstile> f \<supset>\<^sub>i more" using "2" by auto
 hence 11: "\<turnstile> bi f \<supset>\<^sub>i  bi more" using BiImpBiRule by blast
 have  12: "\<turnstile> \<not>\<^sub>i(bi more)" using DiEmpty by auto
 from 7 9 11 12 show ?thesis using MP prop27 by blast  
qed

lemma PrevImpNotPrevNot:
  "\<turnstile> prev f \<supset>\<^sub>i \<not>\<^sub>i(prev \<not>\<^sub>i f) "
by auto

lemma BiEqvAndWprevBi:
 "\<turnstile> bi f \<equiv>\<^sub>i f \<and>\<^sub>i wprev(bi f)"
proof -
 have 1: "\<turnstile> \<box> (f\<^sup>r) \<equiv>\<^sub>i f\<^sup>r \<and>\<^sub>i wnext(\<box> (f\<^sup>r))" 
      using BoxEqvAndWnextBox by blast 
 hence 2: "\<turnstile> (\<box> (f\<^sup>r) \<equiv>\<^sub>i f\<^sup>r \<and>\<^sub>i wnext(\<box> (f\<^sup>r)))\<^sup>r" 
      using ReverseEqv by blast
 have 3: "\<turnstile> (\<box> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i bi (f)  " 
      using RRBoxEqvBi by blast 
 have 4: "\<turnstile> (f\<^sup>r)\<^sup>r \<equiv>\<^sub>i f" 
      using EqvReverseReverse  itl_prop(30) by blast
 have 5: "\<turnstile> (wnext(\<box> (f\<^sup>r)))\<^sup>r \<equiv>\<^sub>i wprev( (\<box>(f\<^sup>r))\<^sup>r )" 
      using RWNextEqvWPrev by blast
 have 6: "\<turnstile> wprev( (\<box>( ((f\<^sup>r))))\<^sup>r) \<equiv>\<^sub>i wprev (bi(f))" 
      using "3" "5" by auto 
 have 7: "\<turnstile> (wnext(\<box> (f\<^sup>r)))\<^sup>r \<equiv>\<^sub>i  wprev (bi (f))" 
      using "5" "6" by auto
 have 8: "\<turnstile> (\<box> (f\<^sup>r) \<equiv>\<^sub>i f\<^sup>r \<and>\<^sub>i wnext(\<box> (f\<^sup>r)))\<^sup>r \<equiv>\<^sub>i 
            ( (\<box> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i ((f\<^sup>r)\<^sup>r) \<and>\<^sub>i (wnext(\<box> (f\<^sup>r)))\<^sup>r) " 
      by (meson "1" "2" RAnd REqvRule iff_defs prop03 valid_def)
 have 9: "\<turnstile> ( (\<box> (f\<^sup>r))\<^sup>r \<equiv>\<^sub>i ((f\<^sup>r)\<^sup>r) \<and>\<^sub>i (wnext(\<box> (f\<^sup>r)))\<^sup>r) \<equiv>\<^sub>i
            ( (bi f) \<equiv>\<^sub>i f \<and>\<^sub>i wprev(bi f)) " 
      using "7" "3" "4" prop40 by auto
 from 9 8 2 show ?thesis by auto
qed
 
lemma DiIntroLoop:
 assumes "\<turnstile> (f \<and>\<^sub>i \<not>\<^sub>i g) \<supset>\<^sub>i prev f"
 shows "\<turnstile> f \<supset>\<^sub>i di g"
proof -
 have 1: "\<turnstile> f \<and>\<^sub>i \<not>\<^sub>i  g \<supset>\<^sub>i prev f" 
      using assms by auto
 hence 2: "\<turnstile> f \<and>\<^sub>i \<not>\<^sub>i  g \<and>\<^sub>i (bi \<not>\<^sub>i g)\<supset>\<^sub>i (prev f) \<and>\<^sub>i (bi \<not>\<^sub>i g) " 
      by auto
 have 3: "\<turnstile> (bi \<not>\<^sub>i g) \<supset>\<^sub>i \<not>\<^sub>i  g" 
      by (rule BiElim)
 hence 4: "\<turnstile> bi \<not>\<^sub>i g \<equiv>\<^sub>i (bi \<not>\<^sub>i g) \<and>\<^sub>i \<not>\<^sub>i  g" 
      using prop38 by blast
 have  5: "\<turnstile> f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i prev f \<and>\<^sub>i bi \<not>\<^sub>i g" 
      using "2" "4" by auto
 have  6: "\<turnstile> bi \<not>\<^sub>i g \<equiv>\<^sub>i (\<not>\<^sub>i g) \<and>\<^sub>i wprev(bi \<not>\<^sub>i g) " 
      by (rule BiEqvAndWprevBi)
 have  7: "\<turnstile> prev f \<and>\<^sub>i bi \<not>\<^sub>i g \<supset>\<^sub>i prev f \<and>\<^sub>i wprev(bi \<not>\<^sub>i g)" 
      using "6" using itl_prop(31) itl_prop(32) prop12 by blast
 have  8: "\<turnstile> f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i prev f \<and>\<^sub>i wprev(bi \<not>\<^sub>i g) " 
      using "5" "7" by auto
 hence 9: "\<turnstile> f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i more \<and>\<^sub>i wprev f \<and>\<^sub>i wprev(bi \<not>\<^sub>i g) " 
      by auto
 hence 10: "\<turnstile> f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i wprev f \<and>\<^sub>i wprev(bi \<not>\<^sub>i g) " 
      by auto
 hence 11: "\<turnstile> f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i wprev (f \<and>\<^sub>i bi \<not>\<^sub>i g)  " 
      by auto
 hence 12: "\<turnstile> bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i wprev (f \<and>\<^sub>i bi \<not>\<^sub>i g) ) " 
      by (rule BiGen)
 have  13: "\<turnstile> bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i wprev (f \<and>\<^sub>i bi \<not>\<^sub>i g) ) \<and>\<^sub>i f \<and>\<^sub>i (bi \<not>\<^sub>i g) 
              \<supset>\<^sub>i bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g)) " 
      by (rule BiInduct)
 hence 14: "\<turnstile> bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g) \<supset>\<^sub>i wprev (f \<and>\<^sub>i bi \<not>\<^sub>i g) ) \<supset>\<^sub>i 
              ( (f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<supset>\<^sub>i bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g))) " 
      using prop36 by blast 
 have  15: "\<turnstile> ( (f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<supset>\<^sub>i bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g))) " 
      using "12" "14" MP by blast
 have  16: "\<turnstile> bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<supset>\<^sub>i f \<and>\<^sub>i (bi \<not>\<^sub>i g) " 
      by (rule BiElim)
 have  17: "\<turnstile> bi(f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<equiv>\<^sub>i (f \<and>\<^sub>i (bi \<not>\<^sub>i g)) " 
      using "16" "15" itl_prop(31) by blast
 have  18: "\<turnstile> (f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<supset>\<^sub>i more " 
      using "9" by auto
 hence 19: "\<turnstile> bi (f \<and>\<^sub>i (bi \<not>\<^sub>i g)) \<supset>\<^sub>i bi more" 
      using BiImpBiRule by blast
 have  20: "\<turnstile> \<not>\<^sub>i(bi more)" 
      using DiEmpty by auto
 have  21: "\<turnstile> \<not>\<^sub>i(f \<and>\<^sub>i (bi \<not>\<^sub>i g)) " 
      using "17" "19" "20" by fastforce
 hence 22: "\<turnstile> \<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i (bi \<not>\<^sub>i g) " 
      by auto
 have  23: "\<turnstile> \<not>\<^sub>i (bi \<not>\<^sub>i g) \<equiv>\<^sub>i di g" 
      by auto
 from 22 23 show ?thesis by auto
qed

lemma DiEqvOrChopMore:
 "\<turnstile> di f \<equiv>\<^sub>i (f \<or>\<^sub>i f;more)"
proof -
 have 1: "\<turnstile> di f \<equiv>\<^sub>i f;true\<^sub>i" by auto
 hence 2: "\<turnstile> di f \<equiv>\<^sub>i f; (empty \<or>\<^sub>i more) " by auto
 have  3: "\<turnstile> f; (empty \<or>\<^sub>i more) \<equiv>\<^sub>i f;empty \<or>\<^sub>i f;more " by auto
 have  4: "\<turnstile> f;empty \<equiv>\<^sub>i f " by (rule ChopEmpty)
 from 2 3 4 show ?thesis by auto
qed

lemma DiAndDiEqvDiAndDiOrDiAndDi:
 "\<turnstile> di f \<and>\<^sub>i di g \<equiv>\<^sub>i di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f) "
proof - 
 have 1: "\<turnstile> di f  \<equiv>\<^sub>i (f \<or>\<^sub>i f;more) " 
      using DiEqvOrChopMore by blast
 have 2: "\<turnstile> di g  \<equiv>\<^sub>i (g \<or>\<^sub>i g;more) " 
      using DiEqvOrChopMore by blast
 have 3: "\<turnstile> di f \<and>\<^sub>i di g \<equiv>\<^sub>i (f \<or>\<^sub>i f;more) \<and>\<^sub>i (g \<or>\<^sub>i g;more)  "  
      using "1" "2" by auto
 have 4: "\<turnstile> (f \<or>\<^sub>i f;more) \<and>\<^sub>i (g \<or>\<^sub>i g;more) \<equiv>\<^sub>i
            (f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more) \<or>\<^sub>i (f;more \<and>\<^sub>i g;more ) " 
      by auto
 have 5: "\<turnstile> more \<equiv>\<^sub>i true\<^sub>i;skip" 
      using MoreEqvSkipChopTrue SkipTrueEqvTrueSkip prop03 by blast
 hence 6: "\<turnstile> f;more \<equiv>\<^sub>i f;(true\<^sub>i;skip)" 
      using RightChopEqvChop by blast
 have  7: "\<turnstile> f;(true\<^sub>i;skip) \<equiv>\<^sub>i (f;true\<^sub>i);skip " 
      by (rule ChopAssoc)
 have  8: "\<turnstile> f;more \<equiv>\<^sub>i prev (di f) " 
      using "6" "7" by (simp add: prev_d_def)
 have  9: "\<turnstile> g;more \<equiv>\<^sub>i g;(true\<^sub>i;skip) " 
      using "5"  RightChopEqvChop by blast
 have 10: "\<turnstile> g;(true\<^sub>i;skip) \<equiv>\<^sub>i (g;true\<^sub>i);skip " 
      by (rule ChopAssoc)
 have 11: "\<turnstile> g;more \<equiv>\<^sub>i prev (di g) " 
      using "9" "10" by (simp add: prev_d_def)
 have 12: "\<turnstile> f;more \<and>\<^sub>i g;more \<equiv>\<^sub>i prev (di f) \<and>\<^sub>i prev (di g) " 
      using "8" "11" by auto
 hence 13: "\<turnstile> f;more \<and>\<^sub>i g;more \<equiv>\<^sub>i prev (di f \<and>\<^sub>i di g) " 
      by auto
 have 14: "\<turnstile> (di f \<and>\<^sub>i di g) \<equiv>\<^sub>i 
            ((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<or>\<^sub>i (f;more \<and>\<^sub>i g;more) " 
      using "3" "4" by auto
 have 15: "\<turnstile> (di f \<and>\<^sub>i di g) \<equiv>\<^sub>i 
            ((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<or>\<^sub>i prev (di f \<and>\<^sub>i di g) " 
      using "13" "14" prop28 by blast
 hence 16: "\<turnstile> (di f \<and>\<^sub>i di g) \<supset>\<^sub>i 
            ((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<or>\<^sub>i prev (di f \<and>\<^sub>i di g) " 
      using itl_prop(31) by blast
 hence 17: "\<turnstile> (di f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<supset>\<^sub>i  
               prev (di f \<and>\<^sub>i di g)  " 
      using prop29 by blast
 hence 18: "\<turnstile> (di f \<and>\<^sub>i di g) \<supset>\<^sub>i di((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more))" 
      using DiIntroLoop by blast
 have  19: "\<turnstile> di((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<equiv>\<^sub>i 
              di(f \<and>\<^sub>i g) \<or>\<^sub>i di( f \<and>\<^sub>i g;more) \<or>\<^sub>i di(g \<and>\<^sub>i f;more)" 
      by auto
 have  20: "\<turnstile> f   \<supset>\<^sub>i di f  " 
      using DiIntro by blast 
 hence 21: "\<turnstile> f \<and>\<^sub>i g \<supset>\<^sub>i g \<and>\<^sub>i di f" 
      by auto
 hence 22: "\<turnstile> di(f \<and>\<^sub>i g) \<supset>\<^sub>i di(g \<and>\<^sub>i di f) " 
      using DiImpDi by blast
 hence 23: "\<turnstile> di(f \<and>\<^sub>i g) \<supset>\<^sub>i di(g \<and>\<^sub>i di f) \<or>\<^sub>i di(f \<and>\<^sub>i di g) " 
      by auto
 have  24: "\<turnstile> g;more \<supset>\<^sub>i di g" 
      by auto
 hence 25: "\<turnstile> f \<and>\<^sub>i g;more \<supset>\<^sub>i f \<and>\<^sub>i di g" 
      by auto
 hence 26: "\<turnstile> di(f \<and>\<^sub>i g;more) \<supset>\<^sub>i di(f \<and>\<^sub>i di g)" 
      using DiImpDi by blast
 hence 27: "\<turnstile> di(f \<and>\<^sub>i g;more) \<supset>\<^sub>i    di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f) " 
      by auto
 have  28: "\<turnstile> f;more \<supset>\<^sub>i di f" 
      by auto
 hence 29: "\<turnstile> g \<and>\<^sub>i f;more \<supset>\<^sub>i g \<and>\<^sub>i di f" 
      by auto
 hence 30: "\<turnstile> di(g \<and>\<^sub>i f;more) \<supset>\<^sub>i di(g \<and>\<^sub>i di f)" 
      using DiImpDi by blast
 hence 31: "\<turnstile> di(g \<and>\<^sub>i f;more) \<supset>\<^sub>i  di(f \<and>\<^sub>i di g) \<or>\<^sub>i  di(g \<and>\<^sub>i di f)  " 
      by auto
 have  32: "\<turnstile>  di(f \<and>\<^sub>i g) \<or>\<^sub>i di( f \<and>\<^sub>i g;more) \<or>\<^sub>i di(g \<and>\<^sub>i f;more) \<supset>\<^sub>i 
               di(f \<and>\<^sub>i di g) \<or>\<^sub>i  di(g \<and>\<^sub>i di f)  " 
      using "23" "27" "31" by auto
 have  33: "\<turnstile> di((f \<and>\<^sub>i g) \<or>\<^sub>i ( f \<and>\<^sub>i g;more) \<or>\<^sub>i (g \<and>\<^sub>i f;more)) \<supset>\<^sub>i 
              di(f \<and>\<^sub>i di g) \<or>\<^sub>i  di(g \<and>\<^sub>i di f)  " 
      using "19" "32" by auto
 have  34: "\<turnstile> (di f \<and>\<^sub>i di g) \<supset>\<^sub>i di(f \<and>\<^sub>i di g) \<or>\<^sub>i  di(g \<and>\<^sub>i di f)" 
      using "18" "33" by auto
 have  35: "\<turnstile>  f  \<supset>\<^sub>i di f "  
      using DiIntro by blast 
 hence 36: "\<turnstile> f \<and>\<^sub>i di g \<supset>\<^sub>i di f \<and>\<^sub>i di g" 
      by auto
 hence 37: "\<turnstile> di (f \<and>\<^sub>i di g) \<supset>\<^sub>i di (di f \<and>\<^sub>i di g) "  
      using DiImpDi by blast
 have  38: "\<turnstile> di (di f \<and>\<^sub>i di g) \<equiv>\<^sub>i di f \<and>\<^sub>i di g" 
      using DiDiAndEqvDi by blast
 have  39: "\<turnstile> di (f \<and>\<^sub>i di g) \<supset>\<^sub>i di f \<and>\<^sub>i di g " 
      using "37" "38" using itl_prop(31) prop02 by blast
 have  40: "\<turnstile>  g  \<supset>\<^sub>i di g "  
      using DiIntro by blast 
 hence 41: "\<turnstile>   g \<and>\<^sub>i di f \<supset>\<^sub>i di f \<and>\<^sub>i di g" 
      by auto
 hence 42: "\<turnstile> di (g \<and>\<^sub>i di f) \<supset>\<^sub>i di (di f \<and>\<^sub>i di g) "  
      using DiImpDi by blast
 have  43: "\<turnstile> di (di f \<and>\<^sub>i di g) \<equiv>\<^sub>i di f \<and>\<^sub>i di g" 
      using DiDiAndEqvDi by blast
 have  44: "\<turnstile> di (g \<and>\<^sub>i di f) \<supset>\<^sub>i di f \<and>\<^sub>i di g " 
      using "42" "43" using itl_prop(31) prop02 by blast
 have  45: "\<turnstile> di (f \<and>\<^sub>i di g) \<or>\<^sub>i di (g \<and>\<^sub>i di f) \<supset>\<^sub>i di f \<and>\<^sub>i di g" 
      using "39" "44"  prop30 by blast
 from 34 45 show ?thesis using itl_prop(31) by blast
qed 

lemma DsAndDsEqvDsAndDiOrDsAndDi:
 "\<turnstile> ds f \<and>\<^sub>i ds g \<equiv>\<^sub>i ds (f \<and>\<^sub>i di g) \<or>\<^sub>i ds(g \<and>\<^sub>i di f)"
proof -
 have 1: "\<turnstile> di f \<and>\<^sub>i di g \<equiv>\<^sub>i di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f)" 
      by (rule DiAndDiEqvDiAndDiOrDiAndDi)
 hence 2: "\<turnstile> (di f \<and>\<^sub>i di g);skip \<equiv>\<^sub>i (di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f));skip" 
      by (rule LeftChopEqvChop)
 have 3: "\<turnstile> (di f \<and>\<^sub>i di g);skip \<equiv>\<^sub>i (di f);skip \<and>\<^sub>i (di g);skip " 
      using ChopSkipAndChopSkip using itl_prop(30) by blast
 have 4: "\<turnstile> (di f);skip \<and>\<^sub>i (di g);skip \<equiv>\<^sub>i (di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f));skip " 
      using "2" "3" by auto
 have 5: "\<turnstile> (di(f \<and>\<^sub>i di g) \<or>\<^sub>i di(g \<and>\<^sub>i di f));skip \<equiv>\<^sub>i di(f \<and>\<^sub>i di g);skip \<or>\<^sub>i di(g \<and>\<^sub>i di f);skip" 
      using OrChopEqv by blast
 have 6: "\<turnstile> ds f \<equiv>\<^sub>i (di f);skip  " 
      using DsDi by blast
 have 7: "\<turnstile> ds g \<equiv>\<^sub>i (di g);skip " 
      using DsDi by blast
 have 8: "\<turnstile>  (di f);skip \<and>\<^sub>i (di g);skip \<equiv>\<^sub>i ds f \<and>\<^sub>i  ds g" 
      using "6" "7" by auto
 have 9: "\<turnstile> ds(f \<and>\<^sub>i di g) \<equiv>\<^sub>i  di(f \<and>\<^sub>i di g);skip" 
      using DsDi by blast
 have 10: "\<turnstile> ds(g \<and>\<^sub>i di f) \<equiv>\<^sub>i di(g \<and>\<^sub>i di f);skip  " 
      using DsDi by blast
 have 11: "\<turnstile> di(f \<and>\<^sub>i di g);skip \<or>\<^sub>i di(g \<and>\<^sub>i di f);skip \<equiv>\<^sub>i ds(f \<and>\<^sub>i di g) \<or>\<^sub>i ds(g \<and>\<^sub>i di f)" 
      using "9" "10" by auto
 from 4 5 8 11 show ?thesis by simp
qed

lemma BsEqvBiMoreImpChop:
 "\<turnstile> bs f \<equiv>\<^sub>i bi(more \<supset>\<^sub>i f;skip) "
proof -
 have 1: "\<turnstile> bs f \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f;skip)" 
      by (simp add: bs_d_def)
 have 2: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (bi f;skip)" 
      using NotNotChopSkip by blast
 have 3: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(bi f);skip) \<equiv>\<^sub>i \<not>\<^sub>i(di \<not>\<^sub>i f;skip)" 
      by auto
 have 4: "\<turnstile> \<not>\<^sub>i(di \<not>\<^sub>i f;skip) \<equiv>\<^sub>i \<not>\<^sub>i((\<not>\<^sub>i f ;true\<^sub>i);skip) " 
      by (simp add:di_d_def)
 have 5: "\<turnstile> \<not>\<^sub>i((\<not>\<^sub>i f ;true\<^sub>i);skip) \<equiv>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i f ;(true\<^sub>i;skip))" 
      using ChopAssocB prop01 by blast
 have 6: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f ;(true\<^sub>i;skip)) \<equiv>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i f ;(skip;true\<^sub>i))" 
      using SkipTrueEqvTrueSkip using TrueChopSkipEqvSkipChopTrue RightChopEqvChop prop01 by blast
 have 7: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f ;(skip;true\<^sub>i)) \<equiv>\<^sub>i \<not>\<^sub>i((\<not>\<^sub>i f ;skip);true\<^sub>i) " 
      using ChopAssoc prop01 by blast
 have 8: "\<turnstile> \<not>\<^sub>i((\<not>\<^sub>i f ;skip);true\<^sub>i) \<equiv>\<^sub>i \<not>\<^sub>i(di (\<not>\<^sub>i f;skip))  "  
      by (simp add: di_d_def)
 have 9: "\<turnstile> \<not>\<^sub>i(di (\<not>\<^sub>i f;skip)) \<equiv>\<^sub>i bi (\<not>\<^sub>i(\<not>\<^sub>i f ;skip))  " 
      using NotDiEqvBiNot by blast
 have 10: "\<turnstile> bi (\<not>\<^sub>i(\<not>\<^sub>i f ;skip)) \<equiv>\<^sub>i bi( empty \<or>\<^sub>i (f;skip))" 
      using NotNotChopSkip using BiEqvBi by blast
 have 11: "\<turnstile> bi( empty \<or>\<^sub>i (f;skip)) \<equiv>\<^sub>i bi( \<not>\<^sub>i more \<or>\<^sub>i (f;skip))" 
 by auto
 from 1 2 3 4 5 6 7 8 9 10 11 show ?thesis by auto
qed


subsubsection \<open>First occurrence\<close>

lemma FstWithAndImp:
 "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i \<rhd> ( f \<and>\<^sub>i g)"
proof -
 have 1: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<equiv>\<^sub>i f \<and>\<^sub>i (bs \<not>\<^sub>if) \<and>\<^sub>i g " 
      by (simp add: first_d_def) 
 have 2: "\<turnstile> f \<and>\<^sub>i (bs \<not>\<^sub>if) \<and>\<^sub>i g \<equiv>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>i(ds f) \<and>\<^sub>i g" 
      using NotDsEqvBsNot using itl_prop(30) prop05 prop06 by blast
 have 3: "\<turnstile> \<not>\<^sub>i(ds f) \<supset>\<^sub>i \<not>\<^sub>i(ds(f \<and>\<^sub>i g))" 
      using DsAndImpElimL using prop27 by blast
 hence 4: "\<turnstile> f \<and>\<^sub>i \<not>\<^sub>i(ds f) \<and>\<^sub>i g \<supset>\<^sub>i f \<and>\<^sub>i g \<and>\<^sub>i \<not>\<^sub>i(ds(f \<and>\<^sub>i g)) " 
      by auto
 have  5: "\<turnstile> f \<and>\<^sub>i g \<and>\<^sub>i \<not>\<^sub>i(ds(f \<and>\<^sub>i g)) \<equiv>\<^sub>i f \<and>\<^sub>i g \<and>\<^sub>i (bs \<not>\<^sub>i(f \<and>\<^sub>i g))  " 
      using NotDsEqvBsNot using prop05 by blast
 have  6: "\<turnstile> f \<and>\<^sub>i g \<and>\<^sub>i (bs \<not>\<^sub>i(f \<and>\<^sub>i g)) \<equiv>\<^sub>i \<rhd>(f \<and>\<^sub>i g) " 
      by (simp add: first_d_def)
 from 1 2 4 5 6 show ?thesis by auto
qed
 
lemma FstWithOrEqv:
 "\<turnstile> \<rhd>(f \<or>\<^sub>i g) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i bs \<not>\<^sub>ig) \<or>\<^sub>i (\<rhd>g \<and>\<^sub>i bs \<not>\<^sub>if)" 
proof -
 have 1: "\<turnstile> \<rhd>(f \<or>\<^sub>i g) \<equiv>\<^sub>i (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i(f \<or>\<^sub>i g)" 
      by (simp add: first_d_def)
 have 2: "\<turnstile>  \<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i  (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>ig)" 
      by auto
 hence 3: "\<turnstile> bs \<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i  bs (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>ig)" 
      using BsEqvRule by blast
 have  4: "\<turnstile>  bs (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>ig) \<equiv>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g" 
      using BsAndEqv  itl_prop(30) by blast
 have  5: "\<turnstile> (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g" 
      using "3" "4" by auto
 have  6: "\<turnstile> (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i 
             (f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) " 
      by auto
 have  7: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i  bs \<not>\<^sub>i g  " 
      by (simp add: first_d_def)
 have  8: "\<turnstile> (g \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<equiv>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i g \<and>\<^sub>i bs \<not>\<^sub>i f)" 
      by auto
 have  9: "\<turnstile> (g \<and>\<^sub>i bs \<not>\<^sub>i g \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i \<rhd>g \<and>\<^sub>i  bs \<not>\<^sub>i f" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g) \<equiv>\<^sub>i 
            (\<rhd>f \<and>\<^sub>i  bs \<not>\<^sub>i g) \<or>\<^sub>i (\<rhd>g \<and>\<^sub>i  bs \<not>\<^sub>i f) " 
      using "7" "8" "9" by auto
 from 1 5 6 10 show ?thesis by auto
qed

lemma FstFstAndEqvFstAnd:
 "\<turnstile> \<rhd>(\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i g"
proof -
  have 1: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<equiv>\<^sub>i f \<and>\<^sub>i (bs \<not>\<^sub>if) \<and>\<^sub>i g" by (simp add: first_d_def)
  hence 2: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i (bs \<not>\<^sub>if)" by auto
  hence 3: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i \<rhd>f \<and>\<^sub>i g \<and>\<^sub>i (bs \<not>\<^sub>if) " by auto
  have 4: "\<turnstile> \<not>\<^sub>i f \<supset>\<^sub>i \<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>if) \<or>\<^sub>i \<not>\<^sub>i g" by auto
  hence 5: "\<turnstile> bs (\<not>\<^sub>i f) \<supset>\<^sub>i bs(\<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>if) \<or>\<^sub>i \<not>\<^sub>i g)" using BsImpBsRule by blast
  have 6: "\<turnstile> \<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>if) \<or>\<^sub>i \<not>\<^sub>i g \<equiv>\<^sub>i \<not>\<^sub>i(f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i g)" by auto
  hence 7: "\<turnstile> bs(\<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>if) \<or>\<^sub>i \<not>\<^sub>i g) \<equiv>\<^sub>i bs(\<not>\<^sub>i(f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i g))" using BsEqvRule by blast
  have 8: "\<turnstile> f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i g\<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i g  "  by (simp add: first_d_def)
  hence 9: "\<turnstile> \<not>\<^sub>i(f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i \<not>\<^sub>i(\<rhd>f \<and>\<^sub>i g)" by auto
  hence 10: "\<turnstile> bs \<not>\<^sub>i(f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i g) \<equiv>\<^sub>i bs \<not>\<^sub>i(\<rhd>f \<and>\<^sub>i g)" using BsEqvRule by blast
  have 11: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i \<rhd>f \<and>\<^sub>i g \<and>\<^sub>i   bs \<not>\<^sub>i(\<rhd>f \<and>\<^sub>i g)" using "3" "5" "7" "10" by auto
  hence 12: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i \<rhd>(\<rhd>f \<and>\<^sub>i g)" by (simp add: first_d_def)
  have 13: "\<turnstile> \<rhd>(\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i g \<and>\<^sub>i   bs \<not>\<^sub>i(\<rhd>f \<and>\<^sub>i g) " by (simp add: first_d_def)
  hence 14: "\<turnstile> \<rhd>(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i \<rhd>f \<and>\<^sub>i g" by auto
 from 12 14 show ?thesis using itl_prop(31) by blast
qed

lemma FstTrue:
 "\<turnstile> \<rhd> true\<^sub>i \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd> true\<^sub>i \<equiv>\<^sub>i true\<^sub>i \<and>\<^sub>i bs \<not>\<^sub>i true\<^sub>i" by (simp add: first_d_def)
 have 2: "\<turnstile> bs \<not>\<^sub>i true\<^sub>i \<equiv>\<^sub>i empty \<or>\<^sub>i (bi \<not>\<^sub>i true\<^sub>i);skip " by (simp add: bs_d_def)
 have 3: "\<turnstile> \<not>\<^sub>i(bi \<not>\<^sub>i true\<^sub>i) " by auto
 hence 4: "\<turnstile> \<not>\<^sub>i((bi \<not>\<^sub>i true\<^sub>i);skip)" by auto
 have 5: "\<turnstile> bs \<not>\<^sub>i true\<^sub>i \<equiv>\<^sub>i empty " using "2" "4" by auto
 from 1 5 show ?thesis by auto
qed

lemma FstFalse:
  "\<turnstile> \<not>\<^sub>i(\<rhd> false\<^sub>i) "
proof -
 have 1: "\<turnstile> \<rhd> false\<^sub>i \<equiv>\<^sub>i false\<^sub>i \<and>\<^sub>i bs true\<^sub>i" by (simp add: first_d_def)
 from 1 show ?thesis by auto
qed

lemma FstChopFalseEqvFalse:
 "\<turnstile> \<not>\<^sub>i(\<rhd>f ; false\<^sub>i)"
by auto

lemma FstEmpty:
 "\<turnstile> \<rhd> empty \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd> empty \<equiv>\<^sub>i empty \<and>\<^sub>i bs \<not>\<^sub>i empty" by (simp add: first_d_def)
 have 2: "\<turnstile> bs \<not>\<^sub>i empty \<equiv>\<^sub>i empty \<or>\<^sub>i bi \<not>\<^sub>i empty;skip" by (simp add: bs_d_def)
 from 1 2 show ?thesis by auto
qed

lemma FstAndEmptyEqvAndEmpty:
 "\<turnstile> \<rhd>f \<and>\<^sub>i empty \<equiv>\<^sub>i f \<and>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd>f \<and>\<^sub>i empty \<equiv>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i empty" by (simp add: first_d_def)
 have 2: "\<turnstile> bs \<not>\<^sub>i f \<equiv>\<^sub>i empty \<or>\<^sub>i bi \<not>\<^sub>if;skip" by (simp add: bs_d_def)
 from 1 2 show ?thesis by auto
qed

lemma FstEmptyOrEqvEmpty:
 "\<turnstile> \<rhd>(empty \<or>\<^sub>i f) \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd>(empty \<or>\<^sub>i f) \<equiv>\<^sub>i (\<rhd>empty \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i bs \<not>\<^sub>i empty)" using FstWithOrEqv by blast
 have 2: "\<turnstile>  \<not>\<^sub>i empty \<equiv>\<^sub>i more"  by auto
 hence 3: "\<turnstile> bs \<not>\<^sub>i empty \<equiv>\<^sub>i bs more" using BsEqvRule by blast
 have 4: "\<turnstile> bs more \<equiv>\<^sub>i empty" using BsMoreEqvEmpty by blast
 have 5: "\<turnstile> \<rhd>f \<and>\<^sub>i bs \<not>\<^sub>i empty \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i empty)" using "3" "4" by auto
 have  6: "\<turnstile> \<rhd>empty \<equiv>\<^sub>i empty" using FstEmpty by blast
 hence 7: "\<turnstile> (\<rhd>empty \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i (empty \<and>\<^sub>i bs \<not>\<^sub>i f)" by auto
 have 8: "\<turnstile> empty \<and>\<^sub>i  bs \<not>\<^sub>i f \<equiv>\<^sub>i empty \<and>\<^sub>i(empty \<or>\<^sub>i bi \<not>\<^sub>i f;skip)" by (simp add:bs_d_def)
 have 9: "\<turnstile> empty \<and>\<^sub>i(empty \<or>\<^sub>i bi \<not>\<^sub>i f;skip) \<equiv>\<^sub>i empty" by auto
 have 10: "\<turnstile> empty \<and>\<^sub>i  bs \<not>\<^sub>i f \<equiv>\<^sub>i empty " using "8" "9" by auto
 have 11: "\<turnstile> (\<rhd>empty \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i bs \<not>\<^sub>i empty) \<equiv>\<^sub>i
          empty \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i empty)" using "7" "10" "5" by auto
 have 12: "\<turnstile> empty \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i empty) \<equiv>\<^sub>i empty" by auto
 from 1 11 12 show ?thesis by auto
qed

lemma ChopEmptyAndEmpty:
 "\<turnstile> f;g \<and>\<^sub>i empty \<equiv>\<^sub>i f \<and>\<^sub>i g \<and>\<^sub>i empty"
apply simp_all 
by (metis interval_prefix_intlen interval_suffix_zero le_zero_eq)

lemma FstChopEmptyEqvFstChopFstEmpty:
 "\<turnstile> \<rhd>f;g \<and>\<^sub>i empty \<equiv>\<^sub>i \<rhd>f;\<rhd>g \<and>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd>f;g \<and>\<^sub>i empty \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i g \<and>\<^sub>i empty" using ChopEmptyAndEmpty by blast
 have 2: "\<turnstile> g \<and>\<^sub>i empty \<equiv>\<^sub>i \<rhd>g \<and>\<^sub>i empty" using FstAndEmptyEqvAndEmpty using itl_prop(30) by blast
 hence 3: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<and>\<^sub>i empty \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i \<rhd>g \<and>\<^sub>i empty" by auto
 have 4: "\<turnstile> \<rhd>f;\<rhd>g \<and>\<^sub>i empty \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i \<rhd>g \<and>\<^sub>i empty" using ChopEmptyAndEmpty by blast
 from 1 3 4 show ?thesis by auto
qed

lemma ChopSkipImpMore:
 "\<turnstile> f;skip \<supset>\<^sub>i more"
proof -
 have 1: "\<turnstile> \<not>\<^sub>i(f;skip \<and>\<^sub>i empty)" by auto
 hence 2: "\<turnstile> \<not>\<^sub>i(f;skip) \<or>\<^sub>i more" by auto
 from 2 show ?thesis by auto
qed

lemma MoreEqvMoreChopTrue:
 "\<turnstile> more \<equiv>\<^sub>i more;true\<^sub>i"
proof -
 have 1: "\<turnstile> more \<equiv>\<^sub>i skip;true\<^sub>i" 
      using MoreEqvSkipChopTrue by blast
 have 2: "\<turnstile> true\<^sub>i \<equiv>\<^sub>i true\<^sub>i;true\<^sub>i" 
      by auto
 hence 3: "\<turnstile> skip;true\<^sub>i \<equiv>\<^sub>i skip;(true\<^sub>i;true\<^sub>i)" 
      using RightChopEqvChop by blast
 have 4: "\<turnstile> skip;(true\<^sub>i;true\<^sub>i) \<equiv>\<^sub>i (skip;true\<^sub>i);true\<^sub>i" 
      using ChopAssoc by blast
 have 5: "\<turnstile> (skip;true\<^sub>i);true\<^sub>i \<equiv>\<^sub>i more;true\<^sub>i" 
      using MoreEqvSkipChopTrue by (simp add: more_d_def next_d_def)
 from 1 3 4 5 show ?thesis using prop03 by blast
qed

lemma BiEmptyEqvEmpty:
 "\<turnstile> bi empty \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> bi empty \<equiv>\<^sub>i \<not>\<^sub>i (di \<not>\<^sub>i empty)" by (simp add: bi_d_def)
 have 2: "\<turnstile> \<not>\<^sub>i (di \<not>\<^sub>i empty) \<equiv>\<^sub>i \<not>\<^sub>i (\<not>\<^sub>i empty;true\<^sub>i)" by (simp add: di_d_def)
 have 3: "\<turnstile> \<not>\<^sub>i (\<not>\<^sub>i empty;true\<^sub>i) \<equiv>\<^sub>i \<not>\<^sub>i (more;true\<^sub>i)" by auto
 have 4: "\<turnstile> more;true\<^sub>i \<equiv>\<^sub>i  more" using MoreEqvMoreChopTrue by auto
 hence 5: "\<turnstile> \<not>\<^sub>i(more;true\<^sub>i) \<equiv>\<^sub>i  \<not>\<^sub>i more" using prop01 by blast
 from 1 2 3 5 show ?thesis by auto
qed

lemma FstMoreEqvSkip:
 "\<turnstile> \<rhd> more \<equiv>\<^sub>i skip "
proof -
 have 1: "\<turnstile> \<rhd> more \<equiv>\<^sub>i more \<and>\<^sub>i bs \<not>\<^sub>i more " by (simp add: first_d_def)
 have 2: "\<turnstile> more \<and>\<^sub>i bs \<not>\<^sub>i more \<equiv>\<^sub>i more \<and>\<^sub>i (empty \<or>\<^sub>i bi \<not>\<^sub>i more;skip) " by (simp add:bs_d_def) 
 have 3: "\<turnstile> more \<and>\<^sub>i (empty \<or>\<^sub>i bi \<not>\<^sub>i more;skip) \<equiv>\<^sub>i more \<and>\<^sub>i bi \<not>\<^sub>i more;skip" by auto
 have 4: "\<turnstile> more \<and>\<^sub>i ((bi \<not>\<^sub>i more);skip) \<equiv>\<^sub>i ((bi \<not>\<^sub>i more);skip)" using ChopSkipImpMore by auto
 have 5: "\<turnstile> ((bi \<not>\<^sub>i more);skip) \<equiv>\<^sub>i bi empty;skip" by auto
 have 6: "\<turnstile> bi empty \<equiv>\<^sub>i empty" using BiEmptyEqvEmpty by auto
 hence 7: "\<turnstile> bi empty;skip \<equiv>\<^sub>i empty;skip" using LeftChopEqvChop by blast
 have 8: "\<turnstile> empty;skip \<equiv>\<^sub>i skip " using EmptyChop by blast
 from 1 2 3 4 5 7 8 show ?thesis by (metis prop03)
qed

lemma FstEqvBsNotAndDi:
 "\<turnstile> \<rhd>f \<equiv>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i di f"
proof -
 have 1: "\<turnstile> bs \<not>\<^sub>i f  \<equiv>\<^sub>i \<not>\<^sub>i(ds f) " by (simp add: ds_d_def)
 hence 2: "\<turnstile> bs \<not>\<^sub>i f  \<and>\<^sub>i di f \<equiv>\<^sub>i \<not>\<^sub>i(ds f) \<and>\<^sub>i di f" by auto
 have 3: "\<turnstile>  di f \<equiv>\<^sub>i (ds f \<or>\<^sub>i f)" using OrDsEqvDi by auto
 hence 4 : "\<turnstile> \<not>\<^sub>i(ds f)  \<and>\<^sub>i di f \<equiv>\<^sub>i \<not>\<^sub>i(ds f)  \<and>\<^sub>i (ds f \<or>\<^sub>i f)" by auto
 have 5 : "\<turnstile> \<not>\<^sub>i(ds f)  \<and>\<^sub>i (ds f \<or>\<^sub>i f) \<equiv>\<^sub>i \<not>\<^sub>i(ds f)  \<and>\<^sub>i f" by auto
 have 6 : "\<turnstile> \<not>\<^sub>i(ds f)  \<and>\<^sub>i f \<equiv>\<^sub>i f \<and>\<^sub>i  bs \<not>\<^sub>i f" using 1 by auto
 from 2 4 5 6 show ?thesis by (simp add: first_d_def)
qed

lemma FstOrDiEqvDi:
 "\<turnstile> \<rhd>f \<or>\<^sub>i di f \<equiv>\<^sub>i di f"
proof -
 have 1: "\<turnstile> \<rhd>f \<or>\<^sub>i di f \<equiv>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>if) \<or>\<^sub>i di f" by (simp add: first_d_def)
 have 2: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>if) \<or>\<^sub>i di f \<equiv>\<^sub>i (f \<or>\<^sub>i di f) \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i di f)" by auto
 have 3: "\<turnstile> (f \<or>\<^sub>i di f) \<equiv>\<^sub>i di f" by auto
 hence 4: "\<turnstile> (f \<or>\<^sub>i di f) \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i di f) \<equiv>\<^sub>i di f \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i di f)" by auto
 have 5: "\<turnstile> di f \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i di f) \<equiv>\<^sub>i di f" by auto
 from 1 2 4 5 show ?thesis by auto
qed

lemma FstAndDiEqvFst:
  "\<turnstile> \<rhd>f \<and>\<^sub>i di f \<equiv>\<^sub>i \<rhd>f"
proof -
 have 1: "\<turnstile> \<rhd>f \<and>\<^sub>i di f \<equiv>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>if \<and>\<^sub>i di f" by (simp add: first_d_def)
 have 2: "\<turnstile> f \<and>\<^sub>i di f \<equiv>\<^sub>i f" by auto
 hence 3: "\<turnstile> f \<and>\<^sub>i bs \<not>\<^sub>if \<and>\<^sub>i di f \<equiv>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>if" by auto
 from 1 3 show ?thesis by (simp add: first_d_def)
qed

lemma EmptyChopSkipInduct:
 assumes "\<turnstile> empty \<supset>\<^sub>i f"
         "\<turnstile> prev f \<supset>\<^sub>i f"
 shows   "\<turnstile> f"
proof -
 have 1: "\<turnstile> empty  \<supset>\<^sub>i f" using assms(1) by auto
 have 2: "\<turnstile> prev f \<supset>\<^sub>i f " using assms(2) by blast
 have 3: "\<turnstile> (empty \<or>\<^sub>i prev f) \<supset>\<^sub>i f" using "1" "2"  prop30 by blast
 have 4: "\<turnstile> wprev f \<equiv>\<^sub>i (empty \<or>\<^sub>i prev f)" by auto
 hence 5: "\<turnstile> wprev f \<supset>\<^sub>i f" using 3 using itl_prop(31) prop02 by blast  
 hence 6: "\<turnstile> \<not>\<^sub>if \<supset>\<^sub>i \<not>\<^sub>i (wprev f) " using prop27 by blast
 hence 7: "\<turnstile> \<not>\<^sub>if \<supset>\<^sub>i prev (\<not>\<^sub>i f) " by auto
 hence 8: "\<turnstile> \<not>\<^sub>i \<not>\<^sub>i f " by (rule PrevLoop)
 from 8 show ?thesis by auto
qed

(* changed proof *)
lemma DiAndEmptyEqvAndEmpty:
 "\<turnstile> di f \<and>\<^sub>i empty \<equiv>\<^sub>i f  \<and>\<^sub>i empty"
proof -
 have 1 : "\<turnstile> di f \<equiv>\<^sub>i (f \<or>\<^sub>i di f;skip) " using DiEqvOrDiChopSkipB by blast
 hence 2: "\<turnstile>  di f \<and>\<^sub>i empty \<equiv>\<^sub>i (f \<or>\<^sub>i di f;skip) \<and>\<^sub>i empty" using prop06 by blast
 have 3 :  "\<turnstile> (f \<or>\<^sub>i di f;skip) \<and>\<^sub>i empty \<equiv>\<^sub>i (f \<and>\<^sub>i empty) \<or>\<^sub>i (di f;skip \<and>\<^sub>i empty) " by auto
 have 4: "\<turnstile>  \<not>\<^sub>i(di f;skip \<and>\<^sub>i empty)" by auto
 hence 5 : "\<turnstile> (f \<and>\<^sub>i empty) \<or>\<^sub>i (di f;skip \<and>\<^sub>i empty) \<equiv>\<^sub>i (f \<and>\<^sub>i empty)" by auto
 from 2 3 5 show ?thesis by auto
qed

lemma MoreImpImpChopSkipEqv:
 "\<turnstile> more \<supset>\<^sub>i ( (f\<supset>\<^sub>ig);skip \<equiv>\<^sub>i ((f;skip)\<supset>\<^sub>i(g;skip)) )  "
proof -
 have 1: "\<turnstile> more \<and>\<^sub>i (f\<supset>\<^sub>ig);skip \<equiv>\<^sub>i more \<and>\<^sub>i  (\<not>\<^sub>i f \<or>\<^sub>i g);skip" 
      by auto
 have 2: "\<turnstile> (\<not>\<^sub>i f \<or>\<^sub>i g);skip \<equiv>\<^sub>i \<not>\<^sub>i f;skip \<or>\<^sub>i g;skip" 
      using OrChopEqv by auto
 hence 3: "\<turnstile> more \<and>\<^sub>i  (\<not>\<^sub>i f \<or>\<^sub>i g);skip \<equiv>\<^sub>i more \<and>\<^sub>i(\<not>\<^sub>i f;skip \<or>\<^sub>i g;skip)" 
      by auto
 have 4: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i f;skip) \<equiv>\<^sub>i empty \<or>\<^sub>i (f;skip) " 
      using NotNotChopSkip by blast
 hence 5: "\<turnstile> (\<not>\<^sub>i f;skip) \<equiv>\<^sub>i \<not>\<^sub>i(empty \<or>\<^sub>i (f;skip))" 
      using itl_prop(30) itl_prop(33) itl_prop(4) prop03 by blast
 have 6: "\<turnstile> \<not>\<^sub>i(empty \<or>\<^sub>i (f;skip)) \<equiv>\<^sub>i (more \<and>\<^sub>i \<not>\<^sub>i(f;skip))" 
      by auto
 have 7: "\<turnstile> (\<not>\<^sub>i f;skip \<or>\<^sub>i g;skip) \<equiv>\<^sub>i ( (more \<and>\<^sub>i \<not>\<^sub>i(f;skip)) \<or>\<^sub>i g;skip)" 
      using "5" "6" by auto
 hence 8: "\<turnstile> more \<and>\<^sub>i(\<not>\<^sub>i f;skip \<or>\<^sub>i g;skip) \<equiv>\<^sub>i more \<and>\<^sub>i ( (more \<and>\<^sub>i \<not>\<^sub>i(f;skip)) \<or>\<^sub>i g;skip) " 
      by auto
 have 9: "\<turnstile> more \<and>\<^sub>i ( (more \<and>\<^sub>i \<not>\<^sub>i(f;skip)) \<or>\<^sub>i g;skip) \<equiv>\<^sub>i more \<and>\<^sub>i (\<not>\<^sub>i(f;skip) \<or>\<^sub>i g;skip)" 
      by auto
 have 10: "\<turnstile> more \<and>\<^sub>i (\<not>\<^sub>i(f;skip) \<or>\<^sub>i g;skip) \<equiv>\<^sub>i more \<and>\<^sub>i ((f;skip)\<supset>\<^sub>i(g;skip))  " 
      by auto
 have 11: "\<turnstile> more \<and>\<^sub>i (f\<supset>\<^sub>ig);skip \<equiv>\<^sub>i more \<and>\<^sub>i ((f;skip)\<supset>\<^sub>i(g;skip))" 
      using "1" "2" "3" "8" "9" "10" by auto
 from 11 show ?thesis using prop31 using MP itl_prop(31) by blast
qed

lemma MoreImpImpPrevEqv:
 "\<turnstile> more \<supset>\<^sub>i ( prev(f\<supset>\<^sub>ig) \<equiv>\<^sub>i (prev f\<supset>\<^sub>i prev g) )"
using MoreImpImpChopSkipEqv by auto

lemma DiEqvDiFst:
 "\<turnstile> di f \<equiv>\<^sub>i di (\<rhd> f)"
proof -
 have 1: "\<turnstile>  di (\<rhd> f) \<equiv>\<^sub>i di (f \<and>\<^sub>i bs \<not>\<^sub>if) " 
      by (simp add: first_d_def)
 have 2: "\<turnstile> di (f \<and>\<^sub>i bs \<not>\<^sub>if) \<supset>\<^sub>i di f \<and>\<^sub>i di (bs \<not>\<^sub>i f)" 
      using DiAndImpAnd by auto
 hence 3: "\<turnstile> di (f \<and>\<^sub>i bs \<not>\<^sub>if) \<supset>\<^sub>i di f" 
      by auto
 have 4: "\<turnstile> di (\<rhd> f) \<supset>\<^sub>i di f" using "1" "3" 
      by auto
 have 5: "\<turnstile> di f \<and>\<^sub>i empty \<equiv>\<^sub>i f \<and>\<^sub>i empty" 
      using DiAndEmptyEqvAndEmpty by blast
 have 6: "\<turnstile> \<rhd> f \<and>\<^sub>i empty \<equiv>\<^sub>i f \<and>\<^sub>i empty " 
      using FstAndEmptyEqvAndEmpty by auto
 have 7: "\<turnstile> di f \<and>\<^sub>i empty \<supset>\<^sub>i \<rhd>f" 
      using "5" "6" by auto
 have 8: "\<turnstile> \<rhd>f \<supset>\<^sub>i di (\<rhd> f)" 
      using DiIntro by auto
 have 9: "\<turnstile>  di f \<and>\<^sub>i empty \<supset>\<^sub>i di (\<rhd> f)" 
      using "7" "8" using prop02 by blast
 hence 10: "\<turnstile> empty \<supset>\<^sub>i (di f \<supset>\<^sub>i di (\<rhd> f)) " 
      by auto
 have 11: "\<turnstile> prev ( di f \<supset>\<^sub>i di (\<rhd> f))\<supset>\<^sub>i more " 
      by auto 
 have 12: "\<turnstile> more \<supset>\<^sub>i( prev ( di f \<supset>\<^sub>i di (\<rhd> f)) \<equiv>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f)))) " 
      using MoreImpImpPrevEqv by auto
 have 13: "\<turnstile> more \<and>\<^sub>i prev ( di f \<supset>\<^sub>i di (\<rhd> f)) \<equiv>\<^sub>i more \<and>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f)))" 
      using "12" prop31 by auto
 have 14: "\<turnstile> prev ( di f \<supset>\<^sub>i di (\<rhd> f)) \<equiv>\<^sub>i more \<and>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f)))" 
      using "11" by auto
 have 15: "\<turnstile> di f \<equiv>\<^sub>i f \<or>\<^sub>i ds f" 
      using OrDsEqvDi by auto
 have 16: "\<turnstile> di f \<equiv>\<^sub>i di f \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f))  " 
      by auto
 have 17: "\<turnstile> di f \<and>\<^sub>i (bs \<not>\<^sub>i f \<or>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f)) \<equiv>\<^sub>i (di f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (di f \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f)) " 
      by auto
 have 18: "\<turnstile> (di f \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i (f \<or>\<^sub>i ds f) \<and>\<^sub>i bs \<not>\<^sub>i f" 
      using "15" by auto
 have 19: "\<turnstile> (f \<or>\<^sub>i ds f) \<and>\<^sub>i bs \<not>\<^sub>i f \<equiv>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (ds f \<and>\<^sub>i bs \<not>\<^sub>i f) " 
      by auto
 have 20: "\<turnstile> \<not>\<^sub>i(ds f \<and>\<^sub>i bs \<not>\<^sub>i f)" 
      by (simp add: ds_d_def)
 have 21: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (ds f \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>i f)" 
      using "20" by auto
 have 22: "\<turnstile> (di f \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>i f)" 
      using "18" "19" "21" by auto 
 have 23: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i \<rhd>f" 
      by (simp add: first_d_def)
 have 24: "\<turnstile> (\<rhd>f) \<supset>\<^sub>i di (\<rhd>f)" 
      using DiIntro by auto
 have 25: "\<turnstile> (f \<and>\<^sub>i bs \<not>\<^sub>i f) \<supset>\<^sub>i di (\<rhd>f)" 
      using "23" "24" by auto
 have 26: "\<turnstile> (di f \<and>\<^sub>i bs \<not>\<^sub>i f) \<supset>\<^sub>i di (\<rhd>f)" 
      using "25" "22" by auto
 hence 27: "\<turnstile> (di f \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f)))) \<supset>\<^sub>i di (\<rhd>f) " 
      by auto
 have 28: "\<turnstile> di f \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f) \<equiv>\<^sub>i di f \<and>\<^sub>i ds f " 
      by (simp add: ds_d_def)
 hence 29: "\<turnstile> di f \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f) \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f)))  \<equiv>\<^sub>i 
              di f \<and>\<^sub>i ds f \<and>\<^sub>i  (prev (di f \<supset>\<^sub>i di (\<rhd> f)))" 
      by auto
 have 30: "\<turnstile>  ds f \<equiv>\<^sub>i prev(di f)" 
      using DsDi by (metis prev_d_def)
 hence 31: "\<turnstile> di f \<and>\<^sub>i ds f \<and>\<^sub>i  (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<equiv>\<^sub>i  
              di f \<and>\<^sub>i prev(di f) \<and>\<^sub>i  (prev (di f \<supset>\<^sub>i di (\<rhd> f)))" 
      by auto
 have 32: "\<turnstile> prev ( di f \<supset>\<^sub>i di (\<rhd> f)) \<supset>\<^sub>i  (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f)))" 
      using "14" by auto
 hence 33: "\<turnstile> di f \<and>\<^sub>i prev(di f) \<and>\<^sub>i prev ( di f \<supset>\<^sub>i di (\<rhd> f)) \<supset>\<^sub>i 
              di f \<and>\<^sub>i prev(di f) \<and>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f)))" 
       by auto
 have 34: "\<turnstile> di f \<and>\<^sub>i prev(di f) \<and>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f))) \<supset>\<^sub>i prev(di (\<rhd>f)) " 
      by auto
 have 35: "\<turnstile> prev(di (\<rhd>f))\<equiv>\<^sub>i (di (\<rhd>f));skip" 
      by (simp add: prev_d_def)
 have 36: "\<turnstile> (di (\<rhd>f));skip \<supset>\<^sub>i di(di (\<rhd>f))" 
      using ChopImpDi by auto
 have 37: "\<turnstile> di(di (\<rhd>f)) \<equiv>\<^sub>i di (\<rhd>f)" 
      using DiEqvDiDi using itl_prop(30) by blast
 have 38: "\<turnstile> di f \<and>\<^sub>i prev(di f) \<and>\<^sub>i (prev(di f) \<supset>\<^sub>i prev(di (\<rhd>f))) \<supset>\<^sub>i di (\<rhd>f)" 
      using "37" "36" "35"  "34" itl_prop(31) prop02 by blast
 have 39: "\<turnstile> di f \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i f) \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<supset>\<^sub>i di (\<rhd>f)" 
      using "29" "31" "33" "38" by (meson  itl_prop(31) prop02)
 hence 40: "\<turnstile> \<not>\<^sub>i(bs \<not>\<^sub>i f) \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<supset>\<^sub>i (di f \<supset>\<^sub>i di (\<rhd>f))" 
      using prop32 by blast
 have 41: "\<turnstile> bs \<not>\<^sub>i f \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<supset>\<^sub>i(di f  \<supset>\<^sub>i di (\<rhd>f))" 
      using "27" prop32 by blast
 have 42: "\<turnstile> (\<not>\<^sub>i(bs \<not>\<^sub>i f) \<or>\<^sub>i bs \<not>\<^sub>i f) \<and>\<^sub>i (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<supset>\<^sub>i(di f  \<supset>\<^sub>i di (\<rhd>f)) " 
      using "40" "41" prop33 by blast
 have 43: "\<turnstile> (\<not>\<^sub>i(bs \<not>\<^sub>i f) \<or>\<^sub>i bs \<not>\<^sub>i f)" 
      by auto
 have 44: "\<turnstile> (prev (di f \<supset>\<^sub>i di (\<rhd> f))) \<supset>\<^sub>i(di f  \<supset>\<^sub>i di (\<rhd>f))" 
      using "42" "43" prop34 by blast
 have 45: "\<turnstile> di f  \<supset>\<^sub>i di (\<rhd>f) " 
      using "10" "44" EmptyChopSkipInduct by blast
 from 4 45 show ?thesis by auto
qed

lemma FstDiEqvFst:
 "\<turnstile> \<rhd>(di f) \<equiv>\<^sub>i \<rhd>f"
proof -
 have 1: "\<turnstile> \<rhd>(di f) \<equiv>\<^sub>i di f \<and>\<^sub>i bs \<not>\<^sub>i (di f)" by (simp add: first_d_def)
 have 2: "\<turnstile>  \<not>\<^sub>i (di f) \<equiv>\<^sub>i bi \<not>\<^sub>i f  " by auto
 hence 3: "\<turnstile> bs \<not>\<^sub>i (di f) \<equiv>\<^sub>i bs (bi \<not>\<^sub>i f)" using BsEqvRule by blast
 have 4: "\<turnstile> bs (bi \<not>\<^sub>i f) \<equiv>\<^sub>i bs ( \<not>\<^sub>i f)" using BsEqvBsBi using itl_prop(30) by blast
 hence 5: "\<turnstile> di f \<and>\<^sub>i bs \<not>\<^sub>i (di f) \<equiv>\<^sub>i di f \<and>\<^sub>i bs ( \<not>\<^sub>i f)" using "3" by auto
 have 6 : "\<turnstile> di f \<equiv>\<^sub>i f \<or>\<^sub>i ds f" using OrDsEqvDi using itl_prop(30) by blast
 hence 7: "\<turnstile> di f \<and>\<^sub>i bs ( \<not>\<^sub>i f) \<equiv>\<^sub>i (f \<or>\<^sub>i ds f) \<and>\<^sub>i bs ( \<not>\<^sub>i f) " by auto
 have 8: "\<turnstile> (f \<or>\<^sub>i ds f) \<and>\<^sub>i bs ( \<not>\<^sub>i f) \<equiv>\<^sub>i (f \<and>\<^sub>i bs ( \<not>\<^sub>i f)) \<or>\<^sub>i (ds f \<and>\<^sub>i bs ( \<not>\<^sub>i f))" by auto
 have 9: "\<turnstile> \<not>\<^sub>i(ds f \<and>\<^sub>i bs ( \<not>\<^sub>i f)) " by (simp add: ds_d_def)
 have 10: "\<turnstile> (f \<and>\<^sub>i bs ( \<not>\<^sub>i f)) \<equiv>\<^sub>i \<rhd>f" by (simp add: first_d_def)
 from 1 5 7 8 9 10 show ?thesis by auto
qed

lemma DiAndFstOrEqvFstOrDiAnd:
 "\<turnstile> di f \<and>\<^sub>i (\<rhd>f \<or>\<^sub>i g) \<equiv>\<^sub>i \<rhd>f \<or>\<^sub>i (di f \<and>\<^sub>i g)"
proof -
 have 1: "\<turnstile> di f \<and>\<^sub>i (\<rhd>f \<or>\<^sub>i g) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i di f) \<or>\<^sub>i (di f \<and>\<^sub>i g)" by auto
 have 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i di f) \<equiv>\<^sub>i \<rhd>f"  using FstAndDiEqvFst by blast
 from 1 2 show ?thesis by auto
qed

lemma DiOrFstAndEqvDi:
 "\<turnstile> di f \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i di f"
proof -
 have 1: "\<turnstile> di f \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i (\<rhd>f \<or>\<^sub>i di f) \<and>\<^sub>i (di f \<or>\<^sub>i g) " by auto
 have 2: "\<turnstile> (\<rhd>f \<or>\<^sub>i di f) \<equiv>\<^sub>i di f" using FstOrDiEqvDi by blast
 from 1 2 show ?thesis by auto
qed

lemma FstDiAndDiEqv:
 "\<turnstile> \<rhd>(di f \<and>\<^sub>i di g) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i di g) \<or>\<^sub>i (\<rhd>g \<and>\<^sub>i di f)"
proof -
 have 1: "\<turnstile> \<rhd>(di f \<and>\<^sub>i di g) \<equiv>\<^sub>i di f \<and>\<^sub>i di g \<and>\<^sub>i bs \<not>\<^sub>i(di f \<and>\<^sub>i di g)" by (simp add: first_d_def)
 have 2: "\<turnstile> \<not>\<^sub>i(di f \<and>\<^sub>i di g) \<equiv>\<^sub>i bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>ig " by auto
 hence 3: "\<turnstile> bs \<not>\<^sub>i(di f \<and>\<^sub>i di g) \<equiv>\<^sub>i bs(bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>ig)" using BsEqvRule by blast
 hence 4: "\<turnstile>  di f \<and>\<^sub>i di g \<and>\<^sub>i bs \<not>\<^sub>i(di f \<and>\<^sub>i di g) \<equiv>\<^sub>i
              di f \<and>\<^sub>i di g \<and>\<^sub>i bs(bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>ig) " by auto
 have 5: "\<turnstile> bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig \<equiv>\<^sub>i bs(bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>ig)" using BsOrBsEqvBsBiOrBi by blast
 hence 6: "\<turnstile> di f \<and>\<^sub>i di g \<and>\<^sub>i bs(bi \<not>\<^sub>if \<or>\<^sub>i bi \<not>\<^sub>ig) \<equiv>\<^sub>i
             di f \<and>\<^sub>i di g \<and>\<^sub>i(bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig)" by auto
 have 7: "\<turnstile> di f \<and>\<^sub>i di g \<and>\<^sub>i(bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig) \<equiv>\<^sub>i 
            (  bs \<not>\<^sub>if \<and>\<^sub>i di f \<and>\<^sub>i di g) \<or>\<^sub>i (di f \<and>\<^sub>i bs \<not>\<^sub>ig \<and>\<^sub>i di g  )" by auto
 have 8: "\<turnstile> \<rhd>f \<equiv>\<^sub>i  bs \<not>\<^sub>if  \<and>\<^sub>i di f" using FstEqvBsNotAndDi by blast
 hence 9: "\<turnstile> bs \<not>\<^sub>if \<and>\<^sub>i di f \<and>\<^sub>i di g \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i di g" by auto
 have 10: "\<turnstile> \<rhd>g \<equiv>\<^sub>i  bs \<not>\<^sub>ig  \<and>\<^sub>i di g" using FstEqvBsNotAndDi by blast
 hence 11: "\<turnstile> di f \<and>\<^sub>i bs \<not>\<^sub>ig \<and>\<^sub>i di g \<equiv>\<^sub>i di f \<and>\<^sub>i \<rhd>g" by auto
 have 12: "\<turnstile> di f \<and>\<^sub>i di g \<and>\<^sub>i(bs \<not>\<^sub>if \<or>\<^sub>i bs \<not>\<^sub>ig) \<equiv>\<^sub>i 
             (\<rhd>f \<and>\<^sub>i di g) \<or>\<^sub>i (di f \<and>\<^sub>i \<rhd>g)" using "7" "9" "11" by auto
 from 1 4 6 12 show ?thesis by auto
qed

lemma BiNotFstEqvBiNot: 
 "\<turnstile> bi \<not>\<^sub>i (\<rhd>f) \<equiv>\<^sub>i bi \<not>\<^sub>i f"
proof -
 have 1: "\<turnstile> di f \<equiv>\<^sub>i di (\<rhd>f)" using DiEqvDiFst by blast
 hence 2: "\<turnstile> \<not>\<^sub>i(di f) \<equiv>\<^sub>i\<not>\<^sub>i( di (\<rhd>f))" by auto
 from 1 2 show ?thesis using NotDiEqvBiNot itl_prop(30) prop03 by blast
qed

lemma BsNotFstEqvBsNot:
 "\<turnstile> bs \<not>\<^sub>i (\<rhd>f) \<equiv>\<^sub>i bs \<not>\<^sub>i f"
proof -
 have 1: "\<turnstile> bs \<not>\<^sub>i (\<rhd>f) \<equiv>\<^sub>i empty \<or>\<^sub>i bi \<not>\<^sub>i (\<rhd>f);skip " by (simp add: bs_d_def)
 have 2: "\<turnstile> bi \<not>\<^sub>i (\<rhd>f) \<equiv>\<^sub>i bi \<not>\<^sub>i f " using BiNotFstEqvBiNot by blast
 hence 3: "\<turnstile> bi \<not>\<^sub>i (\<rhd>f);skip \<equiv>\<^sub>i bi \<not>\<^sub>i f;skip" using LeftChopEqvChop by blast
 hence 4: "\<turnstile> empty \<or>\<^sub>i bi \<not>\<^sub>i (\<rhd>f);skip \<equiv>\<^sub>i empty \<or>\<^sub>i bi \<not>\<^sub>i f;skip" by auto
 from 1 4 show ?thesis by (simp add: bs_d_def)
qed

lemma BsFalseEqvEmpty:
 "\<turnstile> bs false\<^sub>i \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> bs false\<^sub>i \<equiv>\<^sub>i empty \<or>\<^sub>i bi false\<^sub>i;skip " by (simp add: bs_d_def)
 have 2: "\<turnstile> \<not>\<^sub>i(bi false\<^sub>i;skip)" by auto
 from 1 2 show ?thesis by auto
qed

lemma FstState:
 "\<turnstile> \<rhd> (init w) \<equiv>\<^sub>i empty \<and>\<^sub>i init w"
proof -
 have 1: "\<turnstile> \<rhd> (init w) \<equiv>\<^sub>i init w \<and>\<^sub>i bs \<not>\<^sub>i(init w)" by (simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i init w" by auto
 have 3: "\<turnstile> init w \<supset>\<^sub>i bs (init w)" using StateImpBs by auto
 have 4: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i bs (init w)" using "2" "3" by auto
 have 5: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i bs \<not>\<^sub>i(init w)" using "1" by auto
 have 6: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i bs (init w) \<and>\<^sub>i bs \<not>\<^sub>i(init w) " using "4" "5" by auto
 have 7: "\<turnstile> bs (init w) \<and>\<^sub>i bs \<not>\<^sub>i(init w) \<equiv>\<^sub>i bs((init w) \<and>\<^sub>i \<not>\<^sub>i(init w)) " using BsAndEqv by blast
 have 8: "\<turnstile> (init w) \<and>\<^sub>i \<not>\<^sub>i(init w) \<equiv>\<^sub>i  false\<^sub>i"  by auto
 hence 9: "\<turnstile> bs((init w) \<and>\<^sub>i \<not>\<^sub>i(init w)) \<equiv>\<^sub>i  bs false\<^sub>i" using BsEqvRule by blast
 have 10: "\<turnstile> bs false\<^sub>i \<equiv>\<^sub>i empty" using BsFalseEqvEmpty by auto
 have 11: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i empty" using "10" "9" "7" "6" by auto
 have 12: "\<turnstile> \<rhd> (init w) \<supset>\<^sub>i empty \<and>\<^sub>i init w" using "11" "2" by auto
 have 13: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i empty" by auto
 hence 14: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i empty \<or>\<^sub>i bi \<not>\<^sub>i(init w);skip" by auto
 hence 15: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i bs \<not>\<^sub>i(init w)" by (simp add: bs_d_def)
 have 16: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i init w  " by auto
 have 17: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i init w \<and>\<^sub>i bs \<not>\<^sub>i(init w)" using "16" "15" by auto
 hence 18: "\<turnstile> empty \<and>\<^sub>i init w \<supset>\<^sub>i \<rhd>(init w)" by (simp add: first_d_def)
 from 12 18 show ?thesis using itl_prop(31) by blast 
qed

lemma FstStateAndBsNotEmpty:
 "\<turnstile> \<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i empty \<equiv>\<^sub>i \<rhd> (init w)"
proof -
 have 1: "\<turnstile> \<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i empty \<equiv>\<^sub>i \<rhd> (init w) \<and>\<^sub>i bs more " 
      using BsEqvRule NotEmptyEqvMore prop05 by blast
 have 2: "\<turnstile> \<rhd> (init w) \<and>\<^sub>i bs more \<equiv>\<^sub>i \<rhd> (init w) \<and>\<^sub>i empty" 
      using BsMoreEqvEmpty prop05 by blast
 have 3: "\<turnstile> \<rhd> (init w)  \<equiv>\<^sub>i  empty \<and>\<^sub>i (init w)  " 
      using FstState by blast
 hence 4: "\<turnstile> \<rhd> (init w)  \<and>\<^sub>i empty \<equiv>\<^sub>i  empty \<and>\<^sub>i (init w)  \<and>\<^sub>i empty " 
      by auto
 have 5: "\<turnstile> empty \<and>\<^sub>i (init w)  \<and>\<^sub>i empty \<equiv>\<^sub>i empty \<and>\<^sub>i (init w)  " 
      by auto
 have 6: "\<turnstile> empty \<and>\<^sub>i (init w) \<equiv>\<^sub>i \<rhd>(init w)" 
      using FstState using itl_prop(30) by blast
 from 1 2 4 5 6 show ?thesis by auto
qed

lemma FstStateImpFstStateOr:
 "\<turnstile> \<rhd>(init w) \<supset>\<^sub>i \<rhd>(init w \<or>\<^sub>i f)"
proof -
 have 1: "\<turnstile> \<rhd>(init w) \<equiv>\<^sub>i empty \<and>\<^sub>i init w" 
      using FstState by blast
 have 2: "\<turnstile>  empty \<and>\<^sub>i init w \<equiv>\<^sub>i  empty \<and>\<^sub>i (empty \<or>\<^sub>i bi \<not>\<^sub>i f;skip) \<and>\<^sub>i init w" 
      by auto
 have 3: "\<turnstile>  empty \<and>\<^sub>i (empty \<or>\<^sub>i bi \<not>\<^sub>i f;skip) \<and>\<^sub>i init w \<equiv>\<^sub>i 
             empty \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i init w" 
      by (simp add: bs_d_def)
 have 4: "\<turnstile>  empty \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i init w \<equiv>\<^sub>i empty \<and>\<^sub>i init w \<and>\<^sub>i bs \<not>\<^sub>i f " 
      by auto
 have 5: "\<turnstile> empty \<and>\<^sub>i init w  \<equiv>\<^sub>i \<rhd> (init w) " 
      using FstState  itl_prop(30) by blast
 hence 6: "\<turnstile> empty \<and>\<^sub>i init w \<and>\<^sub>i bs \<not>\<^sub>i f \<equiv>\<^sub>i \<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i f" 
      by auto
 have 7: "\<turnstile> \<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i f \<supset>\<^sub>i (\<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i bs \<not>\<^sub>i(init w))" 
      by auto
 have 8: "\<turnstile> \<rhd>(init w \<or>\<^sub>i f) \<equiv>\<^sub>i (\<rhd> (init w) \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (\<rhd>f \<and>\<^sub>i bs \<not>\<^sub>i(init w))" 
      using FstWithOrEqv by blast
 from 1 2 3 4 5 6 7 8 show ?thesis by auto
qed

lemma FstLenSame:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  di (\<rhd>f \<and>\<^sub>i len(i)) \<and>\<^sub>i di (\<rhd>f \<and>\<^sub>i len(j))) \<longrightarrow> (i=j))"
using FstLenSamesem DiLenFstsem by (metis and_defs)

lemma FstAndLenSame:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  di (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i)) \<and>\<^sub>i di (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j))) \<longrightarrow> (i=j))"
using  DiLenFstAndsem by (metis and_defs linorder_neqE_nat not_defs)

lemma FstLenSameChop:
  "(\<forall> \<sigma>. (\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1 \<and>\<^sub>i  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) \<longrightarrow> (i=j))"
proof
  fix \<sigma>
  show "(\<sigma> \<Turnstile>  (\<rhd>f  \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1 \<and>\<^sub>i  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) \<longrightarrow> (i=j) "
  proof 
  assume 0: "(\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1 \<and>\<^sub>i  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2)"
   have 1: "(\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1) " using "0" by auto
   have 2: "(\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1) \<longrightarrow>
            (\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));true\<^sub>i) " by auto
   have 3: "(\<sigma> \<Turnstile>  di(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i))) " using "1" "2" by auto
   have 4: "(\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) " using "0" by auto
   have 5: "(\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) \<longrightarrow>
            (\<sigma> \<Turnstile>  (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));true\<^sub>i) " by auto
   have 6: "(\<sigma> \<Turnstile>  di(\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j))) " using "4" "5" by auto
   have 7: "(\<sigma> \<Turnstile>  di(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i)) \<and>\<^sub>i di(\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j))) " using "3" "6" by auto
   thus "(i=j)" using FstAndLenSame by blast
   qed
qed

lemma DiImpExistsOneDiLenAndFst:
 "(\<forall>\<sigma>. (\<sigma>\<Turnstile> di f) \<longrightarrow> (\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and>\<^sub>i len(k)))))"
proof
 fix \<sigma>
 show "(\<sigma>\<Turnstile> di f) \<longrightarrow> (\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and>\<^sub>i len(k)))) "
 proof 
  assume 0: "(\<sigma>\<Turnstile> di f) "
  have 1: "(\<sigma> \<Turnstile> di (\<rhd> f)) " using "0" DiEqvDiFst valid_def by auto
  have 2: "(\<sigma> \<Turnstile> \<rhd> f) = ( (\<sigma>\<Turnstile>\<rhd> f) \<and> (\<exists>k. (\<sigma> \<Turnstile> len(k))))  " using AndExistsLen valid_def by auto
  have 3: "((\<sigma>\<Turnstile>\<rhd> f) \<and> (\<exists>k. (\<sigma> \<Turnstile> len(k)))) =
           (\<exists>k.  (\<sigma>\<Turnstile>\<rhd> f) \<and>  (\<sigma> \<Turnstile> len(k))) " by auto
  have 4: "(\<sigma> \<Turnstile> di(\<rhd> f)) = (\<exists>k. (\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(k))))" using "2" "3" DiEqvDi by auto
  have 5: "(\<exists>k. (\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(k)))) " using "1" by auto
  then obtain i where 6: "(\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(i))) " by blast
  from 5 obtain j where 7: "(\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(j))) " by blast
  have 8: "(\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(i))) \<and> (\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(j))) " using "6" "7" by auto
  hence 9: "(\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(i)) \<and>\<^sub>i di(\<rhd>f \<and>\<^sub>i len(j)))  " by simp
  hence 10: "i=j " using FstLenSame by blast
  have 11: "\<And>j. (\<sigma> \<Turnstile> di(\<rhd>f \<and>\<^sub>i len(j))) \<longrightarrow> (j=i) " using "9" "10" using FstLenSame and_defs by blast
  thus "(\<exists>! k. (\<sigma> \<Turnstile> di(  \<rhd>f \<and>\<^sub>i len(k) )))" using "11" "5" by blast  
  qed
qed

lemma LFstAndDist_help:
 "(\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i len(k));h1  \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));h2) =
  (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) )"
using LFixedAndDistr using itl_eq by blast  

lemma LFstAndDist_help_1:
 "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i len(k));h1  \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));h2)) =
            (\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) )) "
proof 
  assume 0: "\<exists>k. \<sigma> \<Turnstile>  ((\<rhd> f \<and>\<^sub>i g1) \<and>\<^sub>i len k) ; h1  \<and>\<^sub>i  ((\<rhd> f \<and>\<^sub>i g2) \<and>\<^sub>i len k) ; h2"
  obtain k where 1: " \<sigma> \<Turnstile>  ((\<rhd> f \<and>\<^sub>i g1) \<and>\<^sub>i len k) ; h1  \<and>\<^sub>i  ((\<rhd> f \<and>\<^sub>i g2) \<and>\<^sub>i len k) ; h2" 
  using "0" by auto
  hence 2: "(\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2)) " 
  using  LFstAndDist_help by blast
  show "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) )) "  
  using "2" by auto
  next
  assume 3: "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) )) "
  obtain k where 4: "(\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) ) " 
  using "3" by auto
  hence 5: "(\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i len(k));h1  \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));h2) " 
  using  LFstAndDist_help by blast
  show "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i len(k));h1  \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));h2)) " 
  using "5" by auto
qed

lemma LFstAndDistrsem:
 "(\<forall> \<sigma>. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2)))"
proof 
 fix \<sigma> 
 show "(\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2))" 
 proof -
  have 1: "(\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1);h1) = (\<exists> i. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1) ) " 
  using AndExistsLenChop by auto
  have 2: "(\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g2);h2) = (\<exists> j. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) ) " 
  using AndExistsLenChop by auto
  have 3: "(\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h2) =
          ( (\<exists> i j. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1  \<and>\<^sub>i
                             (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) )
          ) " 
  using "1" "2" by auto
  have 4: "( (\<exists> i j. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(i));h1  \<and>\<^sub>i
                             (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(j));h2) )
          ) =
          ( (\<exists> k. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i len(k));h1  \<and>\<^sub>i
                             (\<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(k));h2) )
          )  " 
  using FstLenSameChop by blast
  have 5: "(\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i len(k));h1  \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));h2)) =
            (\<exists> k. (\<sigma> \<Turnstile>((\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2) \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) ))  " 
  using LFstAndDist_help_1 by blast 
  have 6 : " (\<exists> k. (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i \<rhd>f \<and>\<^sub>i g2 \<and>\<^sub>i len(k));(h1 \<and>\<^sub>i h2) )) =
             (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i \<rhd>f \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2)) "   
  using AndExistsLenChop by auto
  have 7 : "(\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i \<rhd>f \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2)) =
            (\<sigma> \<Turnstile>(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2)) " 
  by auto
  from 3 4 5 6 7 show ?thesis by auto
  qed
qed

lemma LFstAndDistr:
 "\<turnstile> (\<rhd>f \<and>\<^sub>i g1);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h1 \<and>\<^sub>i h2) "
using  LFstAndDistrsem by simp

lemma LFstAndDistrA:
 "\<turnstile> (\<rhd>f \<and>\<^sub>i g1);h \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);h "
proof -
 have 1: "\<turnstile> (\<rhd>f \<and>\<^sub>i g1);h \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);h \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h \<and>\<^sub>i h) " using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(h \<and>\<^sub>i h) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);h " by auto
 from 1 2 show ?thesis by auto
qed

lemma LFstAndDistrB:
 "\<turnstile> (\<rhd>f \<and>\<^sub>i g);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g);(h1 \<and>\<^sub>i h2) " 
proof -
 have 1: "\<turnstile> (\<rhd>f \<and>\<^sub>i g);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g \<and>\<^sub>i g);(h1 \<and>\<^sub>i h2) " using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i g \<and>\<^sub>i g);(h1 \<and>\<^sub>i h2) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g);(h1 \<and>\<^sub>i h2) " by auto
 from 1 2 show ?thesis by auto
qed

lemma LFstAndDistrC:
  "\<turnstile> (\<rhd>f );h1 \<and>\<^sub>i (\<rhd>f );h2 \<equiv>\<^sub>i (\<rhd>f);(h1 \<and>\<^sub>i h2)"
proof -
 have 1: "\<turnstile> (\<rhd>f \<and>\<^sub>i true\<^sub>i);h1 \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i true\<^sub>i);h2 \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i true\<^sub>i \<and>\<^sub>i true\<^sub>i);(h1 \<and>\<^sub>i h2) " 
      using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i true\<^sub>i);h1 \<equiv>\<^sub>i (\<rhd>f );h1 " 
      by auto
 have 3: "\<turnstile> (\<rhd>f \<and>\<^sub>i true\<^sub>i);h2 \<equiv>\<^sub>i (\<rhd>f );h2 " 
      by auto
 have 4: "\<turnstile> (\<rhd>f \<and>\<^sub>i true\<^sub>i \<and>\<^sub>i true\<^sub>i);(h1 \<and>\<^sub>i h2) \<equiv>\<^sub>i (\<rhd>f);(h1 \<and>\<^sub>i h2)  " 
      by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma LFstAndDistrD:
 "\<turnstile> di(\<rhd>f \<and>\<^sub>i g1) \<and>\<^sub>i di(\<rhd>f \<and>\<^sub>i g2) \<equiv>\<^sub>i di(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2) "
proof -
 have 1: "\<turnstile> (\<rhd>f \<and>\<^sub>i g1);true\<^sub>i \<and>\<^sub>i (\<rhd>f \<and>\<^sub>i g2);true\<^sub>i \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(true\<^sub>i \<and>\<^sub>i true\<^sub>i) " 
      using LFstAndDistr by blast
 have 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i g1);true\<^sub>i \<equiv>\<^sub>i di(\<rhd>f \<and>\<^sub>i g1) " 
      by (simp add: di_d_def)
 have 3: "\<turnstile> (\<rhd>f \<and>\<^sub>i g2);true\<^sub>i \<equiv>\<^sub>i di(\<rhd>f \<and>\<^sub>i g2) " 
      by (simp add: di_d_def)
 have 4: "\<turnstile> (\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2);(true\<^sub>i \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i di(\<rhd>f \<and>\<^sub>i g1 \<and>\<^sub>i g2) "  
      by (simp add: di_d_def)
 from 1 2 3 4 show ?thesis by auto
qed


lemma LstAndDistr:
 "\<turnstile> h1;(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i h2;(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) "
proof -
 have 1: "\<turnstile> (\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r);(h1\<^sup>r) \<and>\<^sub>i (\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));(h2\<^sup>r) \<equiv>\<^sub>i 
            (\<rhd>(f\<^sup>r) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r)) " using LFstAndDistr by blast
 hence 2: "\<turnstile> ((\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r);(h1\<^sup>r) \<and>\<^sub>i (\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));(h2\<^sup>r))\<^sup>r \<equiv>\<^sub>i 
            ((\<rhd>(f\<^sup>r) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r)))\<^sup>r" using "1" REqvRule by blast
 have 3: "\<turnstile>  ((\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r);(h1\<^sup>r) \<and>\<^sub>i (\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));(h2\<^sup>r))\<^sup>r \<equiv>\<^sub>i
            ((\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r);(h1\<^sup>r))\<^sup>r \<and>\<^sub>i ((\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));(h2\<^sup>r))\<^sup>r
         " using RAnd by blast
 have 4: "\<turnstile> ((\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r);(h1\<^sup>r))\<^sup>r \<and>\<^sub>i ((\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));(h2\<^sup>r))\<^sup>r \<equiv>\<^sub>i
             (h1\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r)\<^sup>r \<and>\<^sub>i (h2\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r))\<^sup>r
         " by simp
 have 5: "\<turnstile> (h1\<^sup>r)\<^sup>r \<equiv>\<^sub>i h1" using EqvReverseReverse itl_prop(30) by blast
 have 6: "\<turnstile> (h2\<^sup>r)\<^sup>r \<equiv>\<^sub>i h2" using EqvReverseReverse itl_prop(30) by blast
 have 7: "\<turnstile> (g1\<^sup>r)\<^sup>r \<equiv>\<^sub>i g1" using EqvReverseReverse itl_prop(30) by blast
 have 8: "\<turnstile> (g2\<^sup>r)\<^sup>r \<equiv>\<^sub>i g2" using EqvReverseReverse itl_prop(30) by blast
 have 9: "\<turnstile> (f\<^sup>r)\<^sup>r  \<equiv>\<^sub>i f"using EqvReverseReverse itl_prop(30) by blast
 have 10: "\<turnstile> (\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r)\<^sup>r \<equiv>\<^sub>i (\<rhd>(f\<^sup>r))\<^sup>r  \<and>\<^sub>i (g1\<^sup>r)\<^sup>r" using RAnd by blast
 have 11: "\<turnstile> (\<rhd>(f\<^sup>r) \<and>\<^sub>i g2\<^sup>r)\<^sup>r \<equiv>\<^sub>i (\<rhd>(f\<^sup>r))\<^sup>r  \<and>\<^sub>i (g2\<^sup>r)\<^sup>r" using RAnd by blast
 have 12: "\<turnstile> (\<rhd>(f\<^sup>r))\<^sup>r  \<equiv>\<^sub>i \<lhd> (f)  " using RRFirstEqvLast by blast 
 have 13: "\<turnstile> (\<rhd>(f\<^sup>r))\<^sup>r  \<and>\<^sub>i (g1\<^sup>r)\<^sup>r \<equiv>\<^sub>i \<lhd> f \<and>\<^sub>i g1" using  "12" "7" by auto
 have 14: "\<turnstile> (\<rhd>(f\<^sup>r))\<^sup>r  \<and>\<^sub>i (g2\<^sup>r)\<^sup>r \<equiv>\<^sub>i \<lhd> f \<and>\<^sub>i g2" using  "12" "8" by auto
 have 15: "\<turnstile> (h1\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and>\<^sub>i g1\<^sup>r)\<^sup>r \<and>\<^sub>i (h2\<^sup>r)\<^sup>r;(\<rhd>(f\<^sup>r) \<and>\<^sub>i (g2\<^sup>r))\<^sup>r \<equiv>\<^sub>i
              h1;(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i h2;(\<lhd> f \<and>\<^sub>i g2)" using "14" "13" "10" "11" "5" "6" by auto
 have 16: "\<turnstile> ((\<rhd>(f\<^sup>r) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2\<^sup>r));((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r)))\<^sup>r \<equiv>\<^sub>i 
             ((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r))\<^sup>r;((\<rhd> (f\<^sup>r)) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2\<^sup>r))\<^sup>r" by simp
 have 17: "\<turnstile> ((\<rhd> (f\<^sup>r)) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2)\<^sup>r)\<^sup>r \<equiv>\<^sub>i ((\<rhd> (f\<^sup>r))\<^sup>r \<and>\<^sub>i (g1\<^sup>r)\<^sup>r \<and>\<^sub>i (g2\<^sup>r)\<^sup>r)" using RAnd by auto
 have 18: "\<turnstile> ((\<rhd> (f\<^sup>r))\<^sup>r \<and>\<^sub>i (g1\<^sup>r)\<^sup>r \<and>\<^sub>i (g2\<^sup>r)\<^sup>r) \<equiv>\<^sub>i \<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2" using "12" "7" "8" by auto
 have 19: "\<turnstile> ((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r))\<^sup>r \<equiv>\<^sub>i h1 \<and>\<^sub>i h2 " using RRAnd by auto
 have 20: "\<turnstile> ((h1\<^sup>r) \<and>\<^sub>i (h2\<^sup>r))\<^sup>r;((\<rhd> (f\<^sup>r)) \<and>\<^sub>i (g1\<^sup>r) \<and>\<^sub>i (g2\<^sup>r))\<^sup>r \<equiv>\<^sub>i
             (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2)" using "19" "17" "18" by auto
 from 20 15  1 2 3 4 show ?thesis by simp
qed

lemma LstAndDistrA:
  "\<turnstile> h;(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i h;(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i h;(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) "
proof -
 have 1: "\<turnstile> h;(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i h;(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i (h \<and>\<^sub>i h);(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h \<and>\<^sub>i h);(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) \<equiv>\<^sub>i h;(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) " 
      by auto
 from 1 2 show ?thesis by auto
qed

lemma LstAndDistrB:
 "\<turnstile>  h1;(\<lhd> f \<and>\<^sub>i g) \<and>\<^sub>i h2;(\<lhd> f \<and>\<^sub>i g) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g) "
proof -
 have 1: "\<turnstile> h1;(\<lhd> f \<and>\<^sub>i g) \<and>\<^sub>i h2;(\<lhd> f \<and>\<^sub>i g) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g \<and>\<^sub>i g) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g \<and>\<^sub>i g) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i g) " 
      by auto
 from 1 2 show ?thesis by auto
qed

lemma LstAndDistrC:
 "\<turnstile>  h1;(\<lhd> f ) \<and>\<^sub>i h2;(\<lhd> f ) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f ) "
proof -
 have 1: "\<turnstile> h1;(\<lhd> f \<and>\<^sub>i true\<^sub>i) \<and>\<^sub>i h2;(\<lhd> f \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i true\<^sub>i \<and>\<^sub>i true\<^sub>i  )  " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (h1 \<and>\<^sub>i h2);(\<lhd> f \<and>\<^sub>i true\<^sub>i \<and>\<^sub>i true\<^sub>i  ) \<equiv>\<^sub>i (h1 \<and>\<^sub>i h2);(\<lhd> f ) " 
      by auto
 have 3: "\<turnstile> h1;(\<lhd> f \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i h1;(\<lhd> f ) " 
      by auto
 have 4: "\<turnstile> h2;(\<lhd> f \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i h2;(\<lhd> f ) " 
      by auto
 from 1 2 3 4 show ?thesis by auto
qed

lemma LstAndDistrD:
 "\<turnstile>  \<diamond>(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i \<diamond>(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i \<diamond>(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) "
proof -
 have 1: "\<turnstile> true\<^sub>i;(\<lhd> f \<and>\<^sub>i g1) \<and>\<^sub>i true\<^sub>i;(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i (true\<^sub>i \<and>\<^sub>i true\<^sub>i );(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) " 
      using LstAndDistr by blast
 have 2: "\<turnstile> (true\<^sub>i \<and>\<^sub>i true\<^sub>i );(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) \<equiv>\<^sub>i \<diamond>(\<lhd> f \<and>\<^sub>i g1 \<and>\<^sub>i g2) " 
      by (simp add: sometimes_d_def)
 have 3: "\<turnstile> true\<^sub>i;(\<lhd> f \<and>\<^sub>i g1) \<equiv>\<^sub>i \<diamond>(\<lhd> f \<and>\<^sub>i g1) " 
      by (simp add: sometimes_d_def)
 have 4: "\<turnstile> true\<^sub>i;(\<lhd> f \<and>\<^sub>i g2) \<equiv>\<^sub>i \<diamond>(\<lhd> f \<and>\<^sub>i g2) " 
      by (simp add: sometimes_d_def)
 from 1 2 3 4 show ?thesis by auto
qed

lemma NotFstChop:
 "\<turnstile> \<not>\<^sub>i(\<rhd>f ;g) \<equiv>\<^sub>i \<not>\<^sub>i(di (\<rhd>f)) \<or>\<^sub>i (\<rhd>f;\<not>\<^sub>ig)"
proof -
 have  1: "\<turnstile> g \<supset>\<^sub>i true\<^sub>i " by auto
 hence 2: "\<turnstile> \<rhd>f;g \<supset>\<^sub>i \<rhd>f;true\<^sub>i " using RightChopImpChop by blast
 hence 3: "\<turnstile> \<rhd>f;g \<supset>\<^sub>i di(\<rhd>f)" by (simp add:di_d_def)
 hence 4: "\<turnstile> \<not>\<^sub>i(di(\<rhd>f)) \<supset>\<^sub>i \<not>\<^sub>i(\<rhd>f;g) " by auto
 have  5: "\<turnstile> (\<rhd>f; \<not>\<^sub>ig \<supset>\<^sub>i \<not>\<^sub>i(\<rhd>f;g)) \<equiv>\<^sub>i ((\<rhd>f; \<not>\<^sub>ig) \<and>\<^sub>i (\<rhd>f;g) \<supset>\<^sub>i false\<^sub>i ) " by auto
 have  6: "\<turnstile> (\<rhd>f; \<not>\<^sub>ig) \<and>\<^sub>i (\<rhd>f;g) \<equiv>\<^sub>i \<rhd>f;(\<not>\<^sub>ig \<and>\<^sub>i g) " using LFstAndDistrC by blast
 have  7: "\<turnstile> \<not>\<^sub>i(\<rhd>f;(\<not>\<^sub>ig \<and>\<^sub>i g)) " by auto
 have  8: "\<turnstile> \<rhd>f; \<not>\<^sub>ig \<supset>\<^sub>i \<not>\<^sub>i(\<rhd>f;g) " using "5" "6" "7" by auto
 have  9: "\<turnstile> \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f; \<not>\<^sub>ig) \<supset>\<^sub>i \<not>\<^sub>i(\<rhd>f;g) " using "4" "8" by auto
 have 10: "\<turnstile> di(\<rhd>f) \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f))" by auto
 hence 11: "\<turnstile> (\<rhd>f;true\<^sub>i) \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) " by (simp add: di_d_def)
 hence 12: "\<turnstile> (\<rhd>f;(g \<or>\<^sub>i \<not>\<^sub>ig)) \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f))" by  auto
 have 13: "\<turnstile> (\<rhd>f;(g \<or>\<^sub>i \<not>\<^sub>ig)) \<equiv>\<^sub>i (\<rhd>f;g) \<or>\<^sub>i (\<rhd>f;\<not>\<^sub>ig)  "  using ChopOrEqv by auto
 have 14: "\<turnstile> ((\<rhd>f;g) \<or>\<^sub>i (\<rhd>f;\<not>\<^sub>ig)) \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f))  " using "12" "13" by auto
 hence 15: "\<turnstile> \<not>\<^sub>i(\<rhd>f;g) \<supset>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f; \<not>\<^sub>ig)"  by auto
 from 9 15 show ?thesis using itl_prop(31) by blast
qed

lemma BsNotFstChop:
 "\<turnstile> bs(\<not>\<^sub>i(\<rhd>f;g)) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig)"
proof -
 have 1: "\<turnstile> bs(\<not>\<^sub>i(\<rhd>f;g))\<equiv>\<^sub>i empty \<or>\<^sub>i bi\<not>\<^sub>i(\<rhd>f;g);skip " 
      by (simp add:bs_d_def)
 have 2: "\<turnstile> empty \<or>\<^sub>i bi\<not>\<^sub>i(\<rhd>f;g);skip \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f;g));skip" 
      by auto
 have 3: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f;g));skip \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i((\<rhd>f;g);true\<^sub>i);skip" 
      by auto
 have 4: "\<turnstile>  \<not>\<^sub>i((\<rhd>f;g);true\<^sub>i);skip \<equiv>\<^sub>i  \<not>\<^sub>i(\<rhd>f;(g;true\<^sub>i));skip" 
      using ChopAssocB using LeftChopEqvChop itl_prop(33) by blast
 hence 5: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i((\<rhd>f;g);true\<^sub>i);skip \<equiv>\<^sub>i  empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(g;true\<^sub>i));skip" 
      by auto
 have 6: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(g;true\<^sub>i));skip \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;di(g));skip" 
      by (simp add: di_d_def)
 have 7: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;di(g));skip \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(\<not>\<^sub>i(\<rhd>f;di(g));skip))  " 
      by auto
 have 8: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(\<not>\<^sub>i(\<rhd>f;di(g));skip)) \<equiv>\<^sub>i  \<not>\<^sub>i(empty \<or>\<^sub>i (\<rhd>f;di(g));skip ) " 
      using NotNotChopSkip using prop01 by blast
 hence 9: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<not>\<^sub>i(\<not>\<^sub>i(\<rhd>f;di(g));skip)) \<equiv>\<^sub>i  empty \<or>\<^sub>i \<not>\<^sub>i(empty \<or>\<^sub>i (\<rhd>f;di(g));skip )" 
      by auto
 have 10: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(empty \<or>\<^sub>i (\<rhd>f;di(g));skip ) \<equiv>\<^sub>i empty \<or>\<^sub>i (more \<and>\<^sub>i \<not>\<^sub>i((\<rhd>f;di(g));skip ))" 
      by auto
 have 11: "\<turnstile> empty \<or>\<^sub>i (more \<and>\<^sub>i \<not>\<^sub>i((\<rhd>f;di(g));skip )) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i((\<rhd>f;di(g));skip )" 
      by auto
 have 12: "\<turnstile>  empty \<or>\<^sub>i \<not>\<^sub>i((\<rhd>f;di(g));skip ) \<equiv>\<^sub>i  empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(di(g);skip) )" 
      using ChopAssocB "11" itl_prop(30) prop01 prop03 prop28 by blast
 have 13: "\<turnstile>  \<not>\<^sub>i(\<rhd>f;(di(g);skip) ) \<equiv>\<^sub>i  \<not>\<^sub>i(\<rhd>f;(ds(g)) ) " 
      using DsDi using RightChopEqvChop itl_prop(30) itl_prop(33) by blast
 hence 14: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(di(g);skip) ) \<equiv>\<^sub>i  empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(ds(g)) ) " 
      by auto
 have 15: "\<turnstile> empty \<or>\<^sub>i \<not>\<^sub>i(\<rhd>f;(ds(g)) ) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(di (\<rhd>f)) \<or>\<^sub>i (\<rhd>f;\<not>\<^sub>i(ds g))" 
      using NotFstChop by auto
 have 16: "\<turnstile>  (\<rhd>f;\<not>\<^sub>i(ds g))\<equiv>\<^sub>i   (\<rhd>f;(bs \<not>\<^sub>ig))" 
      using NotDsEqvBsNot RightChopEqvChop by blast
 hence 17: "\<turnstile> (empty \<or>\<^sub>i \<not>\<^sub>i(di (\<rhd>f))) \<or>\<^sub>i (\<rhd>f;\<not>\<^sub>i(ds g))\<equiv>\<^sub>i (empty \<or>\<^sub>i \<not>\<^sub>i(di (\<rhd>f))) \<or>\<^sub>i  (\<rhd>f;(bs \<not>\<^sub>ig))" 
      by auto
 from 1 2 3 5 6 7 9 10 11 12 14 15 17 show ?thesis by auto
qed

lemma FstFstChopEqvFstChopFst:
 "\<turnstile> \<rhd>(\<rhd>f;g) \<equiv>\<^sub>i \<rhd>f;\<rhd>g "
proof -
 have 1: "\<turnstile> \<rhd>(\<rhd>f;g) \<equiv>\<^sub>i (\<rhd>f;g) \<and>\<^sub>i bs \<not>\<^sub>i(\<rhd>f;g) " 
      by (simp add: first_d_def)
 have 2: "\<turnstile> bs \<not>\<^sub>i(\<rhd>f;g) \<equiv>\<^sub>i empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig) " 
      using BsNotFstChop by auto
 hence 3: "\<turnstile> (\<rhd>f;g) \<and>\<^sub>i bs \<not>\<^sub>i(\<rhd>f;g) \<equiv>\<^sub>i (\<rhd>f;g) \<and>\<^sub>i (empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig))" 
      by auto
 have 4: "\<turnstile> (\<rhd>f;g) \<and>\<^sub>i (empty \<or>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)) \<or>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig)) \<equiv>\<^sub>i
            ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i \<not>\<^sub>i(di(\<rhd>f))) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig)) " 
      by auto
 have 5: "\<turnstile> \<not>\<^sub>i((\<rhd>f;g) \<and>\<^sub>i \<not>\<^sub>i(di(\<rhd>f)))" 
      by auto
 hence 6: "\<turnstile> ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i \<not>\<^sub>i(di(\<rhd>f))) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig)) \<equiv>\<^sub>i
             ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i (\<rhd>f;bs\<not>\<^sub>ig))" 
      by auto
 have 7: "\<turnstile>  ((\<rhd>f;g) \<and>\<^sub>i (\<rhd>f;(bs\<not>\<^sub>ig))) \<equiv>\<^sub>i ((\<rhd>f;(g \<and>\<^sub>i (bs\<not>\<^sub>ig))))" 
      using LFstAndDistrC by blast
 hence 8: "\<turnstile> ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i (\<rhd>f;(bs\<not>\<^sub>ig))) \<equiv>\<^sub>i 
             ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;(g \<and>\<^sub>i (bs\<not>\<^sub>ig)))) " 
      by auto
 have 9: "\<turnstile> ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i ((\<rhd>f;(g \<and>\<^sub>i (bs\<not>\<^sub>ig)))) \<equiv>\<^sub>i ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i \<rhd>f;\<rhd>g" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> ((\<rhd>f;g) \<and>\<^sub>i empty) \<equiv>\<^sub>i ((\<rhd>f;\<rhd>g) \<and>\<^sub>i empty)" 
      using FstChopEmptyEqvFstChopFstEmpty by blast
 hence 11: "\<turnstile> ((\<rhd>f;g) \<and>\<^sub>i empty) \<or>\<^sub>i \<rhd>f;\<rhd>g \<equiv>\<^sub>i ((\<rhd>f;\<rhd>g) \<and>\<^sub>i empty) \<or>\<^sub>i \<rhd>f;\<rhd>g " 
      by auto
 have 12: "\<turnstile> ((\<rhd>f;\<rhd>g) \<and>\<^sub>i empty) \<or>\<^sub>i \<rhd>f;\<rhd>g \<equiv>\<^sub>i \<rhd>f;\<rhd>g" 
      by auto
 from 1 3 4 6 8 9 11 12 show ?thesis by auto
qed

lemma FstFixFst:
 "\<turnstile>\<rhd>(\<rhd>f) \<equiv>\<^sub>i \<rhd>f"
proof -
 have 1: "\<turnstile> \<rhd>f \<equiv>\<^sub>i (\<rhd>f);empty " using ChopEmpty using itl_prop(30) by blast
 hence 2: "\<turnstile>\<rhd>( \<rhd>f ) \<equiv>\<^sub>i \<rhd>((\<rhd>f);empty) " using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>((\<rhd>f);empty) \<equiv>\<^sub>i \<rhd>f;\<rhd>empty" using FstFstChopEqvFstChopFst by auto
 have 4: "\<turnstile> \<rhd>f;\<rhd>empty \<equiv>\<^sub>i \<rhd>f;empty" using FstEmpty by auto
 have 5: "\<turnstile> \<rhd>f;empty \<equiv>\<^sub>i \<rhd>f" using ChopEmpty by blast
 from 2 3 4 5 show ?thesis by auto
qed

lemma FstCSEqvEmpty:
 "\<turnstile> \<rhd>(f\<^sup>\<star>) \<equiv>\<^sub>i empty"
proof -
 have 1: "\<turnstile> \<rhd>(f\<^sup>\<star>) \<equiv>\<^sub>i \<rhd>(empty \<or>\<^sub>i ((f \<and>\<^sub>i more);f\<^sup>\<star>))" using ChopstarEqv FstEqvRule by blast
 from 1 show ?thesis using FstEmptyOrEqvEmpty by auto
qed

lemma FstIterFixFst:
 "\<turnstile> power (\<rhd> f) n \<equiv>\<^sub>i \<rhd>(power (\<rhd> f) n)"
proof 
 (induct n)
 case 0
 then show ?case 
 proof -
  have 1: "\<turnstile> power (\<rhd> f) 0 \<equiv>\<^sub>i empty" by auto
  have 2: "\<turnstile> empty \<equiv>\<^sub>i \<rhd> empty" using FstEmpty by auto
  have 3: "\<turnstile> \<rhd> empty \<equiv>\<^sub>i \<rhd>(power (\<rhd>f) 0)" by auto
  from 1 2 3 show ?thesis by auto
 qed
 next
  case (Suc n)
  then show ?case 
  proof -
   have 4: "\<turnstile> (power (\<rhd> f) (Suc n)) \<equiv>\<^sub>i (\<rhd> f) ;(power (\<rhd>f) n) " 
   using pow_Suc by simp
   have 5: "\<turnstile> (\<rhd> f) ;(power (\<rhd>f) n) \<equiv>\<^sub>i (\<rhd> f) ; \<rhd> (power (\<rhd> f) n) " 
   using RightChopEqvChop Suc.hyps by blast
   have 6: "\<turnstile> (\<rhd> f) ; \<rhd> (power (\<rhd> f) n) \<equiv>\<^sub>i \<rhd>(\<rhd>f;(power (\<rhd> f) n))" 
   using FstFstChopEqvFstChopFst by auto
   have 7: "\<turnstile> \<rhd>(\<rhd>f;(power (\<rhd> f) n)) \<equiv>\<^sub>i \<rhd>(power (\<rhd> f) (Suc n))" 
   using pow_Suc by simp
   from 4 5 6 7 show ?thesis by auto
  qed
qed 

lemma DsImpNotFst:
 "\<turnstile> ds f \<supset>\<^sub>i (\<not>\<^sub>i(\<rhd>f))"
proof -
 have 1: "\<turnstile> ds f \<and>\<^sub>i \<rhd>f \<equiv>\<^sub>i ds f \<and>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>i f)" by (simp add: first_d_def)
 have 2: "\<turnstile> ds f \<and>\<^sub>i (f \<and>\<^sub>i bs \<not>\<^sub>i f) \<equiv>\<^sub>i ds f \<and>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>i(ds f) " using NotDsEqvBsNot by auto
 from 1 2 show ?thesis by auto
qed

lemma LFixedAndDistrB1:
 "\<turnstile> len(k);f \<and>\<^sub>i len(k);g \<equiv>\<^sub>i len(k);(f \<and>\<^sub>i g) "
proof -
 have 1: "\<turnstile> len(k);f \<equiv>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));f " 
      by auto
 have 2: "\<turnstile> len(k);g \<equiv>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));g " 
      by auto
 have 3: "\<turnstile> len(k);f \<and>\<^sub>i len(k);g \<equiv>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));f \<and>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));g" 
      using "1" "2" by auto
 have 4: "\<turnstile> (true\<^sub>i \<and>\<^sub>i len(k));f \<and>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));g \<equiv>\<^sub>i (true\<^sub>i \<and>\<^sub>i len(k));(f \<and>\<^sub>i g)" 
      using LFixedAndDistrB by blast
 have 5: "\<turnstile> (true\<^sub>i \<and>\<^sub>i len(k));(f \<and>\<^sub>i g) \<equiv>\<^sub>i (len(k));(f \<and>\<^sub>i g)" 
      by auto
 from 1 2 3 4 5 show ?thesis by auto
qed

lemma FstLenAndEqvLenAnd:
 "\<turnstile> \<rhd>(len(k) \<and>\<^sub>i f) \<equiv>\<^sub>i len(k) \<and>\<^sub>i f"
proof -
 have 1: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i ds (len(k))  " 
      using DsAndImpElimL by auto
 hence 2: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (di(len(k)));skip " 
      using DsDi itl_prop(31) prop02 by blast
 hence 3: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i ((len(k);true\<^sub>i));skip" 
      by (simp add: di_d_def)
 hence 4: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (len(k);(true\<^sub>i;skip))" 
      using ChopAssocB itl_prop(31) prop02 by blast
 hence 5: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (len(k);(skip;true\<^sub>i))" 
      using SkipTrueEqvTrueSkip TrueChopSkipEqvSkipChopTrue RightChopEqvChop itl_prop(31) 
            prop02 by blast
 hence 6: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (len(k);(skip;true\<^sub>i)) \<and>\<^sub>i len(k) " 
      by auto
 hence 7: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (len(k);(skip;true\<^sub>i)) \<and>\<^sub>i len(k);empty" 
      using ChopEmpty by (metis itl_prop(31) itl_prop(32) prop02)
 hence 8: "\<turnstile> len(k) \<and>\<^sub>i f \<and>\<^sub>i ds(len(k)\<and>\<^sub>i f) \<supset>\<^sub>i (len(k);((skip;true\<^sub>i) \<and>\<^sub>i empty)) " 
      using LFixedAndDistrB1 using itl_prop(31) prop02 by blast 
 have 9: "\<turnstile> \<not>\<^sub>i(len(k);((skip;true\<^sub>i) \<and>\<^sub>i empty))" 
      by auto
 have 10: "\<turnstile> len(k) \<and>\<^sub>i f \<supset>\<^sub>i \<not>\<^sub>i(ds(len(k)\<and>\<^sub>i f)) " 
      using "8" "9" by auto
 hence 11: "\<turnstile> len(k) \<and>\<^sub>i f \<supset>\<^sub>i bs \<not>\<^sub>i(len(k)\<and>\<^sub>i f)" 
      using NotDsEqvBsNot by auto
 hence 12: "\<turnstile> len(k) \<and>\<^sub>i f \<supset>\<^sub>i len(k) \<and>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i(len(k)\<and>\<^sub>i f)  " 
      by auto
 hence 13: "\<turnstile> len(k) \<and>\<^sub>i f \<supset>\<^sub>i \<rhd>(len(k) \<and>\<^sub>i f)" 
      by (simp add: first_d_def)
 have 14: "\<turnstile> \<rhd>(len(k) \<and>\<^sub>i f) \<supset>\<^sub>i len(k) \<and>\<^sub>i f" 
      by (simp add: first_d_def)
 from 13 14 show ?thesis using itl_prop(31) by blast
qed

lemma FstAndElimL:
 "\<turnstile> \<rhd>f \<supset>\<^sub>i f"
by (simp add: first_d_def)

lemma FstImpNotDiChopSkip:
 "\<turnstile> \<rhd>f \<supset>\<^sub>i \<not>\<^sub>i(di f;skip)"
proof -
 have 1: "\<turnstile> \<rhd>f \<supset>\<^sub>i bs \<not>\<^sub>i f" by (simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd>f \<supset>\<^sub>i \<not>\<^sub>i(ds f)" using NotDsEqvBsNot by auto
 have 3: "\<turnstile> ds f \<equiv>\<^sub>i  di f ; skip" using DsDi by blast
 hence 4: "\<turnstile> \<not>\<^sub>i(ds f) \<equiv>\<^sub>i \<not>\<^sub>i(di f;skip) " by auto
 from 2 4 show ?thesis by auto
qed

lemma FstImpNotDiChopSkipB:
 "\<turnstile> \<rhd>f \<supset>\<^sub>i \<not>\<^sub>i(di (f;skip))"
proof -
 have 1: "\<turnstile> \<rhd>f \<supset>\<^sub>i bs \<not>\<^sub>i f" 
      by (simp add: first_d_def)
 hence 2: "\<turnstile> \<rhd>f \<supset>\<^sub>i \<not>\<^sub>i(ds f)" 
      using NotDsEqvBsNot by auto
 have 3: "\<turnstile> ds f \<equiv>\<^sub>i  di f ; skip" 
      using DsDi by blast
 have 4: "\<turnstile> di f ; skip \<equiv>\<^sub>i (f;true\<^sub>i);skip" 
      by (simp add: di_d_def)
 have 5: "\<turnstile>(f;true\<^sub>i);skip \<equiv>\<^sub>i f;(true\<^sub>i;skip) " 
      using ChopAssocB by blast
 have 6: "\<turnstile> f;(true\<^sub>i;skip) \<equiv>\<^sub>i f;(skip;true\<^sub>i)" 
      using SkipTrueEqvTrueSkip using TrueChopSkipEqvSkipChopTrue RightChopEqvChop by blast
 have 7: "\<turnstile> f;(skip;true\<^sub>i) \<equiv>\<^sub>i (f;skip);true\<^sub>i" 
      using ChopAssoc by blast
 have 8: "\<turnstile> (f;skip);true\<^sub>i \<equiv>\<^sub>i di(f;skip)" 
      by (simp add: di_d_def)
 have 9: "\<turnstile>  \<not>\<^sub>i(ds f) \<equiv>\<^sub>i \<not>\<^sub>i(di(f;skip)) " 
      using "3" "4" "5" "6" "7" "8" by auto
 from 2 9 show ?thesis by auto
qed

lemma FstImpDiEqv:
 "\<turnstile> \<rhd>f \<supset>\<^sub>i (di f \<equiv>\<^sub>i f)"
proof -
 have 1: "\<turnstile> \<rhd>f  \<supset>\<^sub>i \<not>\<^sub>i(di f;skip)" using FstImpNotDiChopSkip by blast
 have 2: "\<turnstile> di f \<supset>\<^sub>i f \<or>\<^sub>i (di f;skip) " using DiEqvOrDiChopSkipB itl_prop(31) by blast
 have 3: "\<turnstile> \<rhd>f  \<and>\<^sub>i di f \<supset>\<^sub>i (f \<or>\<^sub>i (di f;skip)) \<and>\<^sub>i \<not>\<^sub>i(di f;skip) " using "1" "2" by auto
 have 4: "\<turnstile> (f \<or>\<^sub>i (di f;skip)) \<and>\<^sub>i \<not>\<^sub>i(di f;skip) \<equiv>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>i(di f;skip)" by auto
 have 5: "\<turnstile>  \<rhd>f  \<and>\<^sub>i di f \<supset>\<^sub>i f \<and>\<^sub>i \<not>\<^sub>i(di f;skip) " using "3" "4" using itl_prop(31) prop02 by blast
 hence 6: "\<turnstile>  \<rhd>f  \<and>\<^sub>i di f \<supset>\<^sub>i f " using itl_prop(32) by blast
 hence 7: "\<turnstile> \<rhd>f \<supset>\<^sub>i (di f \<supset>\<^sub>i f)" using FstAndElimL prop13 prop26 prop32 by blast
 have 8: "\<turnstile> f  \<supset>\<^sub>i di f " using DiIntro by auto
 hence 9: "\<turnstile> \<rhd>f \<supset>\<^sub>i (f  \<supset>\<^sub>i (di f))" by auto
 from 7 9 show ?thesis by auto
qed

lemma FstAndDiFstAndEqvFstAnd:
 "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i \<rhd>f \<and>\<^sub>i g "
proof -
 have 1: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i \<rhd> f " 
      by auto
 have 2: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i di(\<rhd>f \<and>\<^sub>i g)" 
      by auto
 have 3: "\<turnstile> di(\<rhd>f \<and>\<^sub>i g) \<equiv>\<^sub>i (\<rhd>f \<and>\<^sub>i g) \<or>\<^sub>i di((\<rhd>f \<and>\<^sub>i g);skip)" 
      using DiEqvOrDiChopSkipA by blast
 have 4: "\<turnstile> di((\<rhd>f \<and>\<^sub>i g);skip) \<equiv>\<^sub>i ((\<rhd>f \<and>\<^sub>i g);skip);true\<^sub>i " 
      by (simp add: di_d_def)
 have 5: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i (\<rhd>f \<and>\<^sub>i g) \<or>\<^sub>i ((\<rhd>f \<and>\<^sub>i g);skip);true\<^sub>i " 
      using "2" "3" "4" by auto 
 have 6: "\<turnstile> \<rhd>f \<and>\<^sub>i g \<supset>\<^sub>i f" 
      using FstAndElimL by auto
 hence 7: "\<turnstile> ((\<rhd>f \<and>\<^sub>i g);skip);true\<^sub>i  \<supset>\<^sub>i (f;skip);true\<^sub>i" 
      by auto
 hence 8: "\<turnstile> ((\<rhd>f \<and>\<^sub>i g);skip);true\<^sub>i  \<supset>\<^sub>i di(f;skip)" 
      by auto
 have  9: "\<turnstile> \<rhd>f \<supset>\<^sub>i \<not>\<^sub>i(di(f;skip))" 
      using FstImpNotDiChopSkipB by  blast
 have 10: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i ((\<rhd>f \<and>\<^sub>i g) \<or>\<^sub>i di(f;skip)) " 
      using "5" "8"  prop35 by blast
 have 11: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i \<not>\<^sub>i(di(f;skip)) \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g) \<or>\<^sub>i di(f;skip))" 
      using "9" "10" "1" itl_prop(32) prop02 by blast
 have 12: "\<turnstile> \<not>\<^sub>i(di(f;skip)) \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g) \<or>\<^sub>i di(f;skip)) \<equiv>\<^sub>i \<not>\<^sub>i(di(f;skip)) \<and>\<^sub>i ((\<rhd>f \<and>\<^sub>i g)) " 
      by auto
 have 13: "\<turnstile> \<rhd>f \<and>\<^sub>i  di(\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i (\<rhd>f \<and>\<^sub>i g)"  
      using "11" "12" by auto
 have 14: "\<turnstile> (\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i \<rhd>f" 
      by auto
 hence 15: "\<turnstile> (\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i di(\<rhd>f \<and>\<^sub>i g ) " 
      using DiIntro by auto
 have 16: "\<turnstile> (\<rhd>f \<and>\<^sub>i g) \<supset>\<^sub>i \<rhd>f \<and>\<^sub>i di(\<rhd>f \<and>\<^sub>i g)"  
      using "14" "15" by auto
 from 13 16 show ?thesis using itl_prop(31) by blast
qed

lemma FstAndDiImpBsNotAndDi:
 "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<supset>\<^sub>i (bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) "
proof -
 have 1: "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i ds(di f \<and>\<^sub>i g)  " 
      by (simp add: ds_d_def)
 hence 2: "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i ds(di f)  " 
      using DsAndImp by auto
 hence 3: "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i di(di f);skip" 
      using DsDi using itl_prop(31) prop02 by blast
 hence 4: "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i di f;skip" 
      using DiEqvDiDi using LeftChopImpChop itl_prop(31) prop02 by blast
 hence 5: "\<turnstile> (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i ds f" 
      using DsDi using itl_prop(31) prop02 by blast
 hence 6: "\<turnstile>  (\<rhd>f \<and>\<^sub>i di g) \<and>\<^sub>i \<not>\<^sub>i(bs \<not>\<^sub>i(di f \<and>\<^sub>i g)) \<supset>\<^sub>i\<not>\<^sub>i (\<rhd> f) " 
      using DsImpNotFst using itl_prop(31) prop02 by blast
 from 6 show ?thesis by auto
qed
 
lemma FstFstOrEqvFstOrL:
 "\<turnstile> \<rhd>(\<rhd>f \<or>\<^sub>i g) \<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i g)"
proof -
 have 1: "\<turnstile> \<rhd>(f \<or>\<^sub>i g) \<equiv>\<^sub>i (f \<or>\<^sub>i g) \<and>\<^sub>i bs\<not>\<^sub>i(f \<or>\<^sub>i g)" 
      by (simp add: first_d_def)
 have 2: "\<turnstile>  \<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i  (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>i g)" 
      by auto
 hence 3: "\<turnstile>  bs\<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i  bs (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>i g)" 
      using BsEqvRule by blast
 have 4: "\<turnstile> bs (\<not>\<^sub>if \<and>\<^sub>i \<not>\<^sub>i g) \<equiv>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g" 
      using BsAndEqv itl_prop(30) by blast
 hence 5: "\<turnstile> (f \<or>\<^sub>i g) \<and>\<^sub>i bs\<not>\<^sub>i(f \<or>\<^sub>i g) \<equiv>\<^sub>i  (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g " 
      using "4" "3" by simp
 have 6: "\<turnstile> (f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
            ((f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f)) \<and>\<^sub>i bs \<not>\<^sub>i g" 
      by auto
 have 7: "\<turnstile> ((f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f)) \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
            (\<rhd>f \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f)) \<and>\<^sub>i bs \<not>\<^sub>i g " 
      by (simp add: first_d_def)
 have 8: "\<turnstile> (\<rhd>f \<or>\<^sub>i (g \<and>\<^sub>i bs \<not>\<^sub>i f)) \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
            ((\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i (\<rhd>f \<or>\<^sub>i bs\<not>\<^sub>if)) \<and>\<^sub>i bs \<not>\<^sub>i g " 
      by auto
 have 9: "\<turnstile> ((\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i (\<rhd>f \<or>\<^sub>i bs\<not>\<^sub>if)) \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
            ((\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i ((f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i bs\<not>\<^sub>if)) \<and>\<^sub>i bs \<not>\<^sub>i g" 
      by (simp add: first_d_def)
 have 10: "\<turnstile> ((\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i ((f \<and>\<^sub>i bs \<not>\<^sub>i f) \<or>\<^sub>i bs\<not>\<^sub>if)) \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
             (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g" 
      by auto
 have 11: "\<turnstile> (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i f \<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i 
              (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs(\<not>\<^sub>i(\<rhd>f))\<and>\<^sub>i bs \<not>\<^sub>i g  " 
      using BsNotFstEqvBsNot using itl_prop(30) prop05 prop06 by blast
 have 12: "\<turnstile> (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs(\<not>\<^sub>i(\<rhd>f))\<and>\<^sub>i bs \<not>\<^sub>i g \<equiv>\<^sub>i
             (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs (\<not>\<^sub>i(\<rhd>f) \<and>\<^sub>i \<not>\<^sub>i g)" 
      using BsAndEqv using prop05 by blast
 have 13: "\<turnstile>  (\<not>\<^sub>i(\<rhd>f) \<and>\<^sub>i \<not>\<^sub>i g) \<equiv>\<^sub>i \<not>\<^sub>i(\<rhd>f \<or>\<^sub>i g)" 
      by auto
 hence 14: "\<turnstile> bs (\<not>\<^sub>i(\<rhd>f) \<and>\<^sub>i \<not>\<^sub>i g) \<equiv>\<^sub>i bs \<not>\<^sub>i(\<rhd>f \<or>\<^sub>i g)" 
      using BsEqvRule by blast
 hence 15: "\<turnstile> (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs (\<not>\<^sub>i(\<rhd>f) \<and>\<^sub>i \<not>\<^sub>i g) \<equiv>\<^sub>i (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i(\<rhd>f \<or>\<^sub>i g)" 
      by auto
 have 16: " \<turnstile> (\<rhd>f \<or>\<^sub>i g) \<and>\<^sub>i bs \<not>\<^sub>i(\<rhd>f \<or>\<^sub>i g) \<equiv>\<^sub>i \<rhd>(\<rhd>f \<or>\<^sub>i g)" 
      by (simp add: first_d_def)
 from 16 15 12 11 10 9 8 7 6 5 1 show ?thesis by auto
qed

lemma FstFstOrEqvFstOrR:
 "\<turnstile> \<rhd>(f \<or>\<^sub>i \<rhd>g)\<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i g)"
proof -
 have 1: "\<turnstile> (f \<or>\<^sub>i \<rhd>g)\<equiv>\<^sub>i (\<rhd>g \<or>\<^sub>i f)" by auto
 hence 2: "\<turnstile> \<rhd> (f \<or>\<^sub>i \<rhd>g)\<equiv>\<^sub>i \<rhd>(\<rhd>g \<or>\<^sub>i f) " using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>(\<rhd>g \<or>\<^sub>i f) \<equiv>\<^sub>i \<rhd>(g \<or>\<^sub>i f)" using FstFstOrEqvFstOrL by blast
 have 4: "\<turnstile> (g \<or>\<^sub>i f) \<equiv>\<^sub>i (f \<or>\<^sub>i g)" by auto
 hence 5: "\<turnstile> \<rhd>(g \<or>\<^sub>i f) \<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i g)"  using FstEqvRule by blast
 from 2 3 5 show ?thesis by auto
qed

lemma FstFstOrEqvFstOr:
 "\<turnstile> \<rhd>(\<rhd>f \<or>\<^sub>i \<rhd>g) \<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i g)"
proof -
 have 1: "\<turnstile>  \<rhd>(\<rhd>f \<or>\<^sub>i \<rhd>g) \<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i \<rhd>g)" using FstFstOrEqvFstOrL by blast
 have 2: "\<turnstile> \<rhd>(f \<or>\<^sub>i \<rhd>g) \<equiv>\<^sub>i \<rhd>(f \<or>\<^sub>i g)" using FstFstOrEqvFstOrR by blast
 from 1 2 show ?thesis by auto
qed

lemma FstLenEqvLen:
 "\<turnstile> \<rhd>( len(k)) \<equiv>\<^sub>i len(k) "
proof -
 have 1: "\<turnstile> \<rhd>( len(k) \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i len(k) \<and>\<^sub>i true\<^sub>i" using FstLenAndEqvLenAnd by blast
 have 2: "\<turnstile> len(k) \<and>\<^sub>i true\<^sub>i \<equiv>\<^sub>i len(k)" by auto
 hence 3: "\<turnstile> \<rhd>( len(k) \<and>\<^sub>i true\<^sub>i) \<equiv>\<^sub>i \<rhd>(len(k)) " using FstEqvRule by blast
 from 1 2 3 show ?thesis by auto
qed

lemma FstSkip:
 "\<turnstile> \<rhd> skip \<equiv>\<^sub>i skip"
proof -
 have 1: "\<turnstile> skip \<equiv>\<^sub>i len(1)  " using LenOneEqvSkip using itl_prop(30) by blast
 hence 2: "\<turnstile> \<rhd>skip \<equiv>\<^sub>i \<rhd> (len(1))" using FstEqvRule by blast
 have 3: "\<turnstile> \<rhd>( len(1)) \<equiv>\<^sub>i len(1)" using FstLenEqvLen by blast
 from 1 2 3 show ?thesis using LenOneEqvSkip prop03 by blast
qed

lemma NotChopNotSkip:
 "\<turnstile> \<not>\<^sub>i(f;skip) \<equiv>\<^sub>i empty \<or>\<^sub>i ((\<not>\<^sub>i f);skip)"
proof -
 have 1: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(\<not>\<^sub>if);skip) \<equiv>\<^sub>i empty \<or>\<^sub>i ((\<not>\<^sub>i f);skip)" using NotNotChopSkip by blast
 have 2: "\<turnstile> \<not>\<^sub>i(\<not>\<^sub>i(\<not>\<^sub>if);skip) \<equiv>\<^sub>i \<not>\<^sub>i(f;skip)" by auto
 from 1 2 show ?thesis by auto
qed

lemma BiBoxNotEqvNotTrueChopChopTrue:
 "\<turnstile> bi(\<box> \<not>\<^sub>i f) \<equiv>\<^sub>i \<not>\<^sub>i((true\<^sub>i;f);true\<^sub>i ) "
by auto

lemma BoxMoreStateEqvBsFinState:
 "\<turnstile> \<box>(more \<supset>\<^sub>i \<not>\<^sub>i (init w)) \<equiv>\<^sub>i bs(\<not>\<^sub>i(fin(init w))) "
proof -
 have 1: "\<turnstile> \<box>(more \<supset>\<^sub>i \<not>\<^sub>i (init w)) \<equiv>\<^sub>i \<not>\<^sub>i(\<diamond>(\<not>\<^sub>i(more \<supset>\<^sub>i \<not>\<^sub>i (init w)))) " 
      by auto
 have 2: "\<turnstile> \<not>\<^sub>i(\<diamond>(\<not>\<^sub>i(more \<supset>\<^sub>i \<not>\<^sub>i (init w)))) \<equiv>\<^sub>i \<not>\<^sub>i(true\<^sub>i;(init w \<and>\<^sub>i  more))" 
      by auto
 have 3: "\<turnstile> more \<equiv>\<^sub>i true\<^sub>i; skip" 
      using MoreEqvSkipChopTrue SkipTrueEqvTrueSkip prop03 by blast
 have 4: "\<turnstile> init w \<and>\<^sub>i  more \<equiv>\<^sub>i init w \<and>\<^sub>i (true\<^sub>i; skip) " 
      using "3" by auto
 have 5: "\<turnstile> init w \<and>\<^sub>i (true\<^sub>i; skip)   \<equiv>\<^sub>i ((init w \<and>\<^sub>i empty);(true\<^sub>i;skip))" 
      using StateAndEmptyChop itl_prop(30) by blast
 have 6: "\<turnstile> init w \<and>\<^sub>i  more \<equiv>\<^sub>i ((init w \<and>\<^sub>i empty);(true\<^sub>i;skip)) " 
      using "4" "5" by auto
 have 7: "\<turnstile>  (true\<^sub>i;(init w \<and>\<^sub>i  more)) \<equiv>\<^sub>i (true\<^sub>i;((init w \<and>\<^sub>i empty);(true\<^sub>i;skip)) )" 
      using "6" RightChopEqvChop by blast
 have 8: "\<turnstile> (true\<^sub>i;((init w \<and>\<^sub>i empty);(true\<^sub>i;skip)) ) \<equiv>\<^sub>i
            (((true\<^sub>i;(init w \<and>\<^sub>i empty));(true\<^sub>i;skip)) ) " 
      using ChopAssoc by blast
 have 9: "\<turnstile> (((true\<^sub>i;(init w \<and>\<^sub>i empty));(true\<^sub>i;skip)) ) \<equiv>\<^sub>i 
            ((((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip) )" 
      using ChopAssoc by blast
 have 10: "\<turnstile> (true\<^sub>i;(init w \<and>\<^sub>i  more)) \<equiv>\<^sub>i 
             ((((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip) )" 
      using "7" "8" "9" by auto
 hence 11: "\<turnstile> \<not>\<^sub>i(true\<^sub>i;(init w \<and>\<^sub>i  more)) \<equiv>\<^sub>i 
              \<not>\<^sub>i((((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip) )" 
      by auto
 have 12: "\<turnstile> \<not>\<^sub>i((((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip) ) \<equiv>\<^sub>i
             empty \<or>\<^sub>i (\<not>\<^sub>i((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip)" 
      using NotChopNotSkip by blast
 have 13: "\<turnstile> (\<not>\<^sub>i((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i)) \<equiv>\<^sub>i bi(\<box> \<not>\<^sub>i(init w \<and>\<^sub>i empty))" 
      using BiBoxNotEqvNotTrueChopChopTrue  itl_prop(30) by blast 
 hence 14: "\<turnstile> (\<not>\<^sub>i((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i));skip \<equiv>\<^sub>i 
              (bi(\<box> \<not>\<^sub>i(init w \<and>\<^sub>i empty)));skip" 
      using RightChopEqvChop by auto
 hence 15: "\<turnstile> empty \<or>\<^sub>i (\<not>\<^sub>i((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i));skip \<equiv>\<^sub>i 
              empty \<or>\<^sub>i(bi(\<box> \<not>\<^sub>i(init w \<and>\<^sub>i empty)));skip" 
      by auto
 have 16: "\<turnstile> \<not>\<^sub>i((((true\<^sub>i;(init w \<and>\<^sub>i empty));true\<^sub>i);skip) ) \<equiv>\<^sub>i
              empty \<or>\<^sub>i (bi(\<box> \<not>\<^sub>i(init w \<and>\<^sub>i empty));skip) " 
      using "12" "15" by auto
 have 17: "\<turnstile> empty \<or>\<^sub>i (bi(\<box> \<not>\<^sub>i(init w \<and>\<^sub>i empty));skip) \<equiv>\<^sub>i
             empty \<or>\<^sub>i (bi(\<box> (\<not>\<^sub>i(init w) \<or>\<^sub>i \<not>\<^sub>iempty));skip)" 
      by auto
 have 18: "\<turnstile> \<box> (\<not>\<^sub>i(init w) \<or>\<^sub>i \<not>\<^sub>iempty) \<equiv>\<^sub>i \<box> (\<not>\<^sub>iempty \<or>\<^sub>i \<not>\<^sub>i(init w)  )" 
      by auto
 have 19: "\<turnstile> \<box> (\<not>\<^sub>iempty \<or>\<^sub>i \<not>\<^sub>i(init w)  ) \<equiv>\<^sub>i \<box>(empty \<supset>\<^sub>i \<not>\<^sub>i(init w) )" 
      by auto
 have 20: "\<turnstile> \<box>(empty \<supset>\<^sub>i \<not>\<^sub>i(init w) ) \<equiv>\<^sub>i fin (\<not>\<^sub>i(init w) )" 
      by (simp add: fin_d_def)
 have 21: "\<turnstile> fin (\<not>\<^sub>i(init w) ) \<equiv>\<^sub>i \<not>\<^sub>i(fin (init w)) " 
      using FinEqvFin FinNotStateEqvNotFinState Initprop(2) prop03 by blast
 have 22: "\<turnstile> bi(\<box> (\<not>\<^sub>i(init w) \<or>\<^sub>i \<not>\<^sub>iempty)) \<equiv>\<^sub>i bi (\<not>\<^sub>i(fin (init w)))" 
      using "18" "19" "20" "21" BiEqvBi using prop03 by blast
 hence 23: "\<turnstile> (bi(\<box> (\<not>\<^sub>i(init w) \<or>\<^sub>i \<not>\<^sub>iempty)));skip \<equiv>\<^sub>i (bi (\<not>\<^sub>i(fin (init w))));skip " 
      using RightChopEqvChop by auto
 hence 24: "\<turnstile> empty \<or>\<^sub>i (bi(\<box> (\<not>\<^sub>i(init w) \<or>\<^sub>i \<not>\<^sub>iempty)));skip \<equiv>\<^sub>i 
              empty \<or>\<^sub>i (bi (\<not>\<^sub>i(fin (init w))));skip" 
      by auto
 hence 25: "\<turnstile> empty \<or>\<^sub>i (bi (\<not>\<^sub>i(fin (init w))));skip \<equiv>\<^sub>i bs(\<not>\<^sub>i(fin (init w))) " 
      by (simp add:bs_d_def)
 from 1 2 11 16 17 24  25 show ?thesis by auto
qed

lemma HaltStateEqvFstFinState:
 "\<turnstile> halt (init w) \<equiv>\<^sub>i \<rhd> (fin (init w))"
proof -
 have 1: "\<turnstile> halt (init w) \<equiv>\<^sub>i \<box>(empty \<equiv>\<^sub>i (init w))" by (simp add: halt_d_def)
 have 2: "\<turnstile> \<box>(empty \<equiv>\<^sub>i (init w)) \<equiv>\<^sub>i \<box>((empty \<supset>\<^sub>i (init w)) \<and>\<^sub>i ((init w) \<supset>\<^sub>i empty))" by auto
 have 3: "\<turnstile> \<box>((empty \<supset>\<^sub>i (init w)) \<and>\<^sub>i ((init w) \<supset>\<^sub>i empty)) \<equiv>\<^sub>i 
            \<box>((empty \<supset>\<^sub>i (init w))) \<and>\<^sub>i \<box>((init w) \<supset>\<^sub>i empty)" by auto
 have 4: "\<turnstile> ((init w) \<supset>\<^sub>i empty) \<equiv>\<^sub>i (more \<supset>\<^sub>i \<not>\<^sub>i(init w))" by auto
 hence 5: "\<turnstile> \<box> ((init w) \<supset>\<^sub>i empty) \<equiv>\<^sub>i \<box> (more \<supset>\<^sub>i \<not>\<^sub>i(init w))" using BoxEqvBox by blast
 have 6: "\<turnstile> \<box> (more \<supset>\<^sub>i \<not>\<^sub>i(init w)) \<equiv>\<^sub>i bs(\<not>\<^sub>i(fin(init w)))" using BoxMoreStateEqvBsFinState by blast
 have 7: "\<turnstile> \<box>((empty \<supset>\<^sub>i (init w))) \<equiv>\<^sub>i fin(init w)" by (simp add: fin_d_def)
 have 8: "\<turnstile>  \<box>((empty \<supset>\<^sub>i (init w))) \<and>\<^sub>i \<box>((init w) \<supset>\<^sub>i empty) \<equiv>\<^sub>i
             fin(init w) \<and>\<^sub>i bs(\<not>\<^sub>i(fin(init w)))" using "5" "6" "7" by auto
 from 1 2 3 8 show ?thesis by (simp add:first_d_def)
qed

lemma FstLenEqvLenFst:
 "\<turnstile> \<rhd>(len k ; f) \<equiv>\<^sub>i len k ; \<rhd> f"
proof -
  have 1: "\<turnstile> len k ; f \<equiv>\<^sub>i \<rhd>(len k) ; f " using FstLenEqvLen LeftChopEqvChop by auto
  have 2: "\<turnstile> \<rhd>(len k ; f) \<equiv>\<^sub>i \<rhd> (\<rhd>(len k) ; f) " using 1 FstEqvRule by blast
  have 3: "\<turnstile> \<rhd> (\<rhd>(len k) ; f) \<equiv>\<^sub>i \<rhd>(len k) ; \<rhd>f" using FstFstChopEqvFstChopFst by blast
  have 4: "\<turnstile> \<rhd>(len k) ; \<rhd>f \<equiv>\<^sub>i len k ; \<rhd>f" using FstLenEqvLen LeftChopEqvChop by auto
  from 2 3 4 show ?thesis by auto
qed

lemma FstNextEqvNextFst:
 "\<turnstile> \<rhd>(\<circle> f) \<equiv>\<^sub>i \<circle>( \<rhd> f)"
proof -
  have 1: "\<turnstile> \<rhd>(\<circle> f) \<equiv>\<^sub>i \<rhd>(skip ; f) " using FstEqvRule by (simp add: next_d_def)
  have 2: "\<turnstile> skip ; f \<equiv>\<^sub>i \<rhd>skip ; f " using FstSkip by auto
  have 3: "\<turnstile> \<rhd>(skip ; f) \<equiv>\<^sub>i \<rhd> (\<rhd>skip ; f) " using 2 FstEqvRule LeftChopEqvChop by blast
  have 4: "\<turnstile> \<rhd> (\<rhd>skip ; f) \<equiv>\<^sub>i \<rhd>skip ; \<rhd>f " using 3 FstFstChopEqvFstChopFst by blast
  have 5: "\<turnstile> \<rhd>skip ; \<rhd>f \<equiv>\<^sub>i skip ; \<rhd>f " using 4 FstSkip LeftChopEqvChop by blast
  have 6: "\<turnstile> skip ; \<rhd>f  \<equiv>\<^sub>i  \<circle>( \<rhd> f)" by (simp add: next_d_def)
  from 1 2 3 4 5 6 show ?thesis by auto
qed

lemma FstDiamondStateEqvHalt:
 "\<turnstile> \<rhd> (\<diamond> (init w)) \<equiv>\<^sub>i halt (init w)"
proof -
  have 1: "\<turnstile> \<diamond> (init w) \<equiv>\<^sub>i \<diamond> ((init w) \<and>\<^sub>i true\<^sub>i)" by simp
  have 2: "\<turnstile> fin (init w) ; true\<^sub>i \<equiv>\<^sub>i \<diamond> ((init w) \<and>\<^sub>i true\<^sub>i)" using 1 FinChopEqvDiamond by blast
  have 3: "\<turnstile> fin (init w) ; true\<^sub>i \<equiv>\<^sub>i di (fin (init w))" using di_d_def by simp
  have 4: "\<turnstile> (\<diamond> (init w)) \<equiv>\<^sub>i (di (fin (init w)))" using 1 2 3 by auto
  have 5: "\<turnstile> \<rhd> (\<diamond> (init w)) \<equiv>\<^sub>i \<rhd> (di (fin (init w)))" using 4 FstEqvRule by blast
  hence 6: "\<turnstile> \<rhd> (\<diamond> (init w)) \<equiv>\<^sub>i \<rhd> (fin (init w))" using FstDiEqvFst by auto
  hence 7: "\<turnstile> \<rhd> (\<diamond> (init w)) \<equiv>\<^sub>i halt (init w)" using HaltStateEqvFstFinState by auto
  from 7 show ?thesis by simp
qed

lemma FstBoxStateEqvStateAndEmpty:
 "\<turnstile> \<rhd> (\<box> (init w)) \<equiv>\<^sub>i (init w) \<and>\<^sub>i empty"
proof -
  have 1: "\<turnstile> (init w) \<and>\<^sub>i (\<box> (init w))\<^sup>\<star> \<equiv>\<^sub>i \<box> (init w)" 
  using BoxCSEqvBox by blast
  have 2: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i (init w) \<and>\<^sub>i (\<box> (init w))\<^sup>\<star>" 
  using 1 by simp
  hence 3: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i (init w) \<and>\<^sub>i (\<box> (init w))\<^sup>\<star>" 
  by blast
  have 4: "\<turnstile> ((init w) \<and>\<^sub>i empty) ; (\<box> (init w))\<^sup>\<star> \<equiv>\<^sub>i (init w) \<and>\<^sub>i (\<box> (init w))\<^sup>\<star>" 
  using StateAndEmptyChop by blast
  have 5: "\<turnstile> (init w) \<and>\<^sub>i (\<box> (init w))\<^sup>\<star>  \<equiv>\<^sub>i  ((init w) \<and>\<^sub>i empty) ; (\<box> (init w))\<^sup>\<star>" 
  using 4 by simp
  have 6: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i ((init w) \<and>\<^sub>i empty) ; (\<box> (init w))\<^sup>\<star> " 
  using 3 5 prop03  by blast
  have 7: "\<turnstile> ((init w) \<and>\<^sub>i empty) ; (\<box> (init w))\<^sup>\<star> \<equiv>\<^sub>i \<rhd> (init w) ; (\<box> (init w))\<^sup>\<star> " 
  using FstState by auto
  have 8: "\<turnstile> \<box> (init w) \<equiv>\<^sub>i \<rhd> (init w) ; (\<box> (init w))\<^sup>\<star> " 
  using 6 7 prop03 by blast
  have 9: "\<turnstile> \<rhd> (\<box> (init w)) \<equiv>\<^sub>i \<rhd> (\<rhd> (init w) ; (\<box> (init w))\<^sup>\<star>) " 
  using 8 FstEqvRule by blast
  have 10: "\<turnstile> \<rhd> (\<rhd> (init w) ; (\<box> (init w))\<^sup>\<star>) \<equiv>\<^sub>i \<rhd> (init w) ;  \<rhd> ((\<box> (init w))\<^sup>\<star>) " 
  using FstFstChopEqvFstChopFst by blast
  have 11: "\<turnstile> \<rhd> (init w) ;  \<rhd> ((\<box> (init w))\<^sup>\<star>) \<equiv>\<^sub>i \<rhd> (init w) ; empty" 
  using RightChopEqvChop FstCSEqvEmpty by blast
  have 12: "\<turnstile> \<rhd> (init w) ; empty \<equiv>\<^sub>i \<rhd> (init w)" 
  using RightChopEqvChop ChopEmpty by blast
  have 13: "\<turnstile> \<rhd> (init w) \<equiv>\<^sub>i (init w) \<and>\<^sub>i empty" 
  using FstState by auto
  from 9 10 11 12 13 show ?thesis by auto
qed

end
